/************************************************************
Copyright 1992 by Digital Equipment Corporation, Maynard, Massachusetts,

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of Digital not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

********************************************************/

/*
This program is a basic block instruction scheduler for the
Digital "Alpha"  21064 microprocessor.

It's purpose is to rewrite a basic block of alpha instructions
in such a way as to minimize stalls and maximize the number
of multiple issue slots kept full.

This scheduler would be useful, for example, in a compiler
back-end or optimizer.

There is a "main" tacked on the end that should produce the
following output:

%a.out
Basic block before code scheduling
|01234567890123456789| Cycle   Offset Number Instruction
|I                   |     0 00000000      0 LDD      F0,0(R16)
|.I                  |     1 00000004      1 LDD      F8,0(R0)
|  I                 |     2 00000008      2 LDD      F1,8(R16)
|  .I                |     3 0000000C      3 LDD      F9,8(R0)
|    I               |     4 00000010      4 MULG     F0,F5,F0
|    ......I         |    10 00000014      5 ADDG     F8,F0,F8
|           I        |    11 00000018      6 MULG     F1,F5,F1
|           ......I  |    17 0000001C      7 ADDG     F9,F1,F9
|                  I |    18 00000020      8 LDA      R16,10(R16)
|                  .I|    19 00000024      9 CMPLT    R16,R10,R11
|I                   |    20 00000028     10 LDA      R0,10(R0)
|..I                 |    22 0000002C     11 STD      F8,0(R0)
|   I                |    23 00000030     12 STD      F9,8(R0)
|   I                |    23 00000034     13 BNE      R11,0

Basic block after code scheduling
|01234567890123456789| Cycle   Offset Number Instruction
|I                   |     0 00000000      0 LDD      F0,0(R16)
|.I                  |     1 00000004      2 LDD      F1,8(R16)
|  I                 |     2 00000008      1 LDD      F8,0(R0)
|  I                 |     2 0000000C      8 LDA      R16,10(R16)
|   I                |     3 00000010      4 MULG     F0,F5,F0
|   I                |     3 00000014      3 LDD      F9,8(R0)
|    I               |     4 00000018      6 MULG     F1,F5,F1
|    I               |     4 0000001C     10 LDA      R0,10(R0)
|     I              |     5 00000020      9 CMPLT    R16,R10,R11
|     ....I          |     9 00000024      5 ADDG     F8,F0,F8
|          I         |    10 00000028      7 ADDG     F9,F1,F9
|          ...I      |    13 0000002C     11 STD      F8,0(R0)
|              I     |    14 00000030     12 STD      F9,8(R0)
|              I     |    14 00000034     13 BNE      R11,0

*/


/* This is version 1.0 dated 12 Aug 1992 */

#include <assert.h>
#include <sys/types.h>

/*
** Breakup alpha instruction into component fields.
*/

typedef union alpha_instruction {
	unsigned word;
	unsigned char byte[4];

	/* common fields to all instructions */
	struct {
		unsigned : 26;
		unsigned opcode : 6;
	} common;

	/* memory format */
	struct {
		signed memory_displacement : 16;
		unsigned rb : 5;
		unsigned ra : 5;
		unsigned opcode : 6;
	} m_format;

	/* branch format */
	struct {
		signed branch_displacement : 21;
		unsigned ra : 5;
		unsigned opcode : 6;
	} b_format;

	/* operate format */
	struct {
		unsigned rc : 5;
		unsigned function : 7;
		unsigned form : 1;
		unsigned : 3;
		unsigned rb : 5;
		unsigned ra : 5;
		unsigned opcode : 6;
	} o_format;

	/* literal (operate) format */
	struct {
		unsigned rc : 5;
		unsigned function : 7;
		unsigned form : 1;
		unsigned literal : 8;
		unsigned ra : 5;
		unsigned opcode : 6;
	} l_format;

	/* floating point format */
	struct {
		unsigned fc: 5;
		unsigned function : 11;
		unsigned fb : 5;
		unsigned fa : 5;
		unsigned opcode : 6;
	} f_format;

	/* pal format */
	struct {
		unsigned function : 26;
		unsigned opcode : 6;
	} pal_format;

	/* jsr (memory) format */
	struct {
		signed hint : 14;
		unsigned function : 2;
		unsigned rb : 5;
		unsigned ra : 5;
		unsigned opcode : 6;
	} j_format;
} alpha_inst;

/* Opcode definitions */

#define EVX_OPC_LDL 40
#define EVX_OPC_LDQ 41
#define EVX_OPC_LDLL 42
#define EVX_OPC_LDQL 43
#define EVX_OPC_LDQU 11
#define EVX_OPC_LDA 8
#define EVX_OPC_STL 44
#define EVX_OPC_STQ 45
#define EVX_OPC_STLC 46
#define EVX_OPC_STQC 47
#define EVX_OPC_STQU 15
#define EVX_OPC_LDAH 9
#define EVX_OPC_LDF 32
#define EVX_OPC_LDD 33
#define EVX_OPC_LDS 34
#define EVX_OPC_LDT 35
#define EVX_OPC_STF 36
#define EVX_OPC_STD 37
#define EVX_OPC_STS 38
#define EVX_OPC_STT 39
#define EVX_OPC_SYNC 24
#define EVX_SYNC_MB 16384
#define EVX_SYNC_FETCH 32768
#define EVX_SYNC_RS 61440
#define EVX_SYNC_TRAPB 0
#define EVX_SYNC_FETCH_M 40960
#define EVX_SYNC_RCC 49152
#define EVX_SYNC_RC 57344
#define EVX_OPC_JSR 26
/* Opcodes for the Branch format instructions                               */
#define EVX_OPC_BR 48
#define EVX_OPC_BSR 52
#define EVX_OPC_BLBC 56
#define EVX_OPC_BLBS 60
#define EVX_OPC_FBEQ 49
#define EVX_OPC_FBNE 53
#define EVX_OPC_BEQ 57
#define EVX_OPC_BNE 61
#define EVX_OPC_FBLT 50
#define EVX_OPC_FBGE 54
#define EVX_OPC_BLT 58
#define EVX_OPC_BGE 62
#define EVX_OPC_FBLE 51
#define EVX_OPC_FBGT 55
#define EVX_OPC_BLE 59
#define EVX_OPC_BGT 63
/* Opcodes for the Operate format instructions                              */
#define EVX_OPC_ADDI 16
#define EVX_OPC_LOGI 17
#define EVX_OPC_SHFI 18
#define EVX_OPC_MULI 19
#define EVX_ADDI_ADDL 0
#define EVX_ADDI_ADDLV 64
#define EVX_ADDI_ADDQ 32
#define EVX_ADDI_ADDQV 96
#define EVX_ADDI_S4ADDL 2
#define EVX_ADDI_S4ADDQ 34
#define EVX_ADDI_S8ADDL 18
#define EVX_ADDI_S8ADDQ 50
#define EVX_ADDI_S4SUBL 11
#define EVX_ADDI_S4SUBQ 43
#define EVX_ADDI_S8SUBL 27
#define EVX_ADDI_S8SUBQ 59
#define EVX_ADDI_SUBL 9
#define EVX_ADDI_SUBLV 73
#define EVX_ADDI_SUBQ 41
#define EVX_ADDI_SUBQV 105
#define EVX_ADDI_CMPEQ 45
#define EVX_ADDI_CMPLT 77
#define EVX_ADDI_CMPLE 109
#define EVX_ADDI_CMPULT 29
#define EVX_ADDI_CMPULE 61
#define EVX_ADDI_CMPBGE 15
#define EVX_LOGI_AND 0
#define EVX_LOGI_BIC 8
#define EVX_LOGI_CMOVEQ 36
#define EVX_LOGI_CMOVNE 38
#define EVX_LOGI_CMOVLBS 20
#define EVX_LOGI_BIS 32
#define EVX_LOGI_ORNOT 40
#define EVX_LOGI_CMOVLT 68
#define EVX_LOGI_CMOVGE 70
#define EVX_LOGI_CMOVLBC 22
#define EVX_LOGI_XOR 64
#define EVX_LOGI_EQV 72
#define EVX_LOGI_CMOVLE 100
#define EVX_LOGI_CMOVGT 102
#define EVX_SHFI_SLL 57
#define EVX_SHFI_SRA 60
#define EVX_SHFI_SRL 52
#define EVX_SHFI_EXTBL 6
#define EVX_SHFI_EXTWL 22
#define EVX_SHFI_EXTLL 38
#define EVX_SHFI_EXTQL 54
#define EVX_SHFI_EXTWH 90
#define EVX_SHFI_EXTLH 106
#define EVX_SHFI_EXTQH 122
#define EVX_SHFI_INSBL 11
#define EVX_SHFI_INSWL 27
#define EVX_SHFI_INSLL 43
#define EVX_SHFI_INSQL 59
#define EVX_SHFI_INSWH 87
#define EVX_SHFI_INSLH 103
#define EVX_SHFI_INSQH 119
#define EVX_SHFI_MSKBL 2
#define EVX_SHFI_MSKWL 18
#define EVX_SHFI_MSKLL 34
#define EVX_SHFI_MSKQL 50
#define EVX_SHFI_MSKWH 82
#define EVX_SHFI_MSKLH 98
#define EVX_SHFI_MSKQH 114
#define EVX_SHFI_ZAP 48
#define EVX_SHFI_ZAPNOT 49
#define EVX_MULI_MULL 0
#define EVX_MULI_MULQV 96
#define EVX_MULI_MULLV 64
#define EVX_MULI_UMULH 48
#define EVX_MULI_MULQ 32
/* Opcodes for the Floating point Operate format instructions               */
#define EVX_OPC_FPDEC 21
#define EVX_OPC_FPIEEE 22
#define EVX_OPC_FPAUX 23
/* FUNCTION FIELDS for the Floating point instructions                      */
#define EVX_FPAUX_CPYS 32
#define EVX_FPAUX_CVTLQ 16
#define EVX_FPAUX_FCMOVEQ 42
#define EVX_FPAUX_FCMOVNE 43
#define EVX_FPAUX_WR_FPCTL 36
#define EVX_FPAUX_RD_FPCTL 37
#define EVX_FPAUX_CPYSN 33
#define EVX_FPAUX_CVTQL 48
#define EVX_FPAUX_FCMOVLT 44
#define EVX_FPAUX_FCMOVGE 45
#define EVX_FPAUX_CPYSE 34
#define EVX_FPAUX_CVTQLV 304
#define EVX_FPAUX_CVTQLSV 1328
#define EVX_FPAUX_FCMOVLE 46
#define EVX_FPAUX_FCMOVGT 47
#define EVX_FPAUX_CPYSEE 3
#define EVX_FPIEEE_CMPSEQ 133
#define EVX_FPIEEE_CMPSLT 134
#define EVX_FPIEEE_CMPSLE 135
#define EVX_FPIEEE_CMPSUN 132
#define EVX_FPIEEE_CMPTEQ 165
#define EVX_FPIEEE_CMPTLT 166
#define EVX_FPIEEE_CMPTLE 167
#define EVX_FPIEEE_CMPTUN 164
#define EVX_FPIEEE_CMPSEQSU 1413
#define EVX_FPIEEE_CMPSLTSU 1414
#define EVX_FPIEEE_CMPSLESU 1415
#define EVX_FPIEEE_CMPSUNSU 1412
#define EVX_FPIEEE_CMPTEQSU 1445
#define EVX_FPIEEE_CMPTLTSU 1446
#define EVX_FPIEEE_CMPTLESU 1447
#define EVX_FPIEEE_CMPTUNSU 1444
#define EVX_FPIEEE_CVTQS 188
#define EVX_FPIEEE_CVTQSC 60
#define EVX_FPIEEE_CVTQSM 124
#define EVX_FPIEEE_CVTQSP 252
#define EVX_FPIEEE_CVTQSSUI 1980
#define EVX_FPIEEE_CVTQSSUIC 1852
#define EVX_FPIEEE_CVTQSSUIM 1916
#define EVX_FPIEEE_CVTQSSUIP 2044
#define EVX_FPIEEE_CVTQT 190
#define EVX_FPIEEE_CVTQTC 62
#define EVX_FPIEEE_CVTQTM 126
#define EVX_FPIEEE_CVTQTP 254
#define EVX_FPIEEE_CVTQTSUI 1982
#define EVX_FPIEEE_CVTQTSUIC 1854
#define EVX_FPIEEE_CVTQTSUIM 1918
#define EVX_FPIEEE_CVTQTSUIP 2046
#define EVX_FPIEEE_CVTST 142
#define EVX_FPIEEE_CVTSTC 14
#define EVX_FPIEEE_CVTSTM 78
#define EVX_FPIEEE_CVTSTP 206
#define EVX_FPIEEE_CVTSTSUI 1934
#define EVX_FPIEEE_CVTSTSUIC 1806
#define EVX_FPIEEE_CVTSTSUIM 1870
#define EVX_FPIEEE_CVTSTSUIP 1998
#define EVX_FPIEEE_CVTTS 172
#define EVX_FPIEEE_CVTTSC 44
#define EVX_FPIEEE_CVTTSM 108
#define EVX_FPIEEE_CVTTSP 236
#define EVX_FPIEEE_CVTTSU 428
#define EVX_FPIEEE_CVTTSUC 300
#define EVX_FPIEEE_CVTTSUM 364
#define EVX_FPIEEE_CVTTSUP 492
#define EVX_FPIEEE_CVTTSSU 1452
#define EVX_FPIEEE_CVTTSSUC 1324
#define EVX_FPIEEE_CVTTSSUM 1388
#define EVX_FPIEEE_CVTTSSUP 1516
#define EVX_FPIEEE_CVTTSSUI 1964
#define EVX_FPIEEE_CVTTSSUIC 1836
#define EVX_FPIEEE_CVTTSSUIM 1900
#define EVX_FPIEEE_CVTTSSUIP 2028
#define EVX_FPIEEE_DIVS 131
#define EVX_FPIEEE_DIVSC 3
#define EVX_FPIEEE_DIVSM 67
#define EVX_FPIEEE_DIVSP 195
#define EVX_FPIEEE_DIVSU 387
#define EVX_FPIEEE_DIVSUC 259
#define EVX_FPIEEE_DIVSUM 323
#define EVX_FPIEEE_DIVSUP 451
#define EVX_FPIEEE_DIVSSU 1411
#define EVX_FPIEEE_DIVSSUC 1283
#define EVX_FPIEEE_DIVSSUM 1347
#define EVX_FPIEEE_DIVSSUP 1475
#define EVX_FPIEEE_DIVSSUI 1923
#define EVX_FPIEEE_DIVSSUIC 1795
#define EVX_FPIEEE_DIVSSUIM 1859
#define EVX_FPIEEE_DIVSSUIP 1987
#define EVX_FPIEEE_DIVT 163
#define EVX_FPIEEE_DIVTC 35
#define EVX_FPIEEE_DIVTM 99
#define EVX_FPIEEE_DIVTP 227
#define EVX_FPIEEE_DIVTU 419
#define EVX_FPIEEE_DIVTUC 291
#define EVX_FPIEEE_DIVTUM 355
#define EVX_FPIEEE_DIVTUP 483
#define EVX_FPIEEE_DIVTSU 1443
#define EVX_FPIEEE_DIVTSUC 1315
#define EVX_FPIEEE_DIVTSUM 1379
#define EVX_FPIEEE_DIVTSUP 1507
#define EVX_FPIEEE_DIVTSUI 1955
#define EVX_FPIEEE_DIVTSUIC 1827
#define EVX_FPIEEE_DIVTSUIM 1891
#define EVX_FPIEEE_DIVTSUIP 2019
#define EVX_FPIEEE_MULS 130
#define EVX_FPIEEE_MULSC 2
#define EVX_FPIEEE_MULSM 66
#define EVX_FPIEEE_MULSP 194
#define EVX_FPIEEE_MULSU 386
#define EVX_FPIEEE_MULSUC 258
#define EVX_FPIEEE_MULSUM 322
#define EVX_FPIEEE_MULSUP 450
#define EVX_FPIEEE_MULSSU 1410
#define EVX_FPIEEE_MULSSUC 1282
#define EVX_FPIEEE_MULSSUM 1346
#define EVX_FPIEEE_MULSSUP 1474
#define EVX_FPIEEE_MULSSUI 1922
#define EVX_FPIEEE_MULSSUIC 1794
#define EVX_FPIEEE_MULSSUIM 1858
#define EVX_FPIEEE_MULSSUIP 1986
#define EVX_FPIEEE_MULT 162
#define EVX_FPIEEE_MULTC 34
#define EVX_FPIEEE_MULTM 98
#define EVX_FPIEEE_MULTP 226
#define EVX_FPIEEE_MULTU 418
#define EVX_FPIEEE_MULTUC 290
#define EVX_FPIEEE_MULTUM 354
#define EVX_FPIEEE_MULTUP 482
#define EVX_FPIEEE_MULTSU 1442
#define EVX_FPIEEE_MULTSUC 1314
#define EVX_FPIEEE_MULTSUM 1378
#define EVX_FPIEEE_MULTSUP 1506
#define EVX_FPIEEE_MULTSUI 1954
#define EVX_FPIEEE_MULTSUIC 1826
#define EVX_FPIEEE_MULTSUIM 1890
#define EVX_FPIEEE_MULTSUIP 2018
#define EVX_FPIEEE_SUBS 129
#define EVX_FPIEEE_SUBSC 1
#define EVX_FPIEEE_SUBSM 65
#define EVX_FPIEEE_SUBSP 193
#define EVX_FPIEEE_SUBSU 385
#define EVX_FPIEEE_SUBSUC 257
#define EVX_FPIEEE_SUBSUM 321
#define EVX_FPIEEE_SUBSUP 449
#define EVX_FPIEEE_SUBSSU 1409
#define EVX_FPIEEE_SUBSSUC 1281
#define EVX_FPIEEE_SUBSSUM 1345
#define EVX_FPIEEE_SUBSSUP 1473
#define EVX_FPIEEE_SUBSSUI 1921
#define EVX_FPIEEE_SUBSSUIC 1793
#define EVX_FPIEEE_SUBSSUIM 1857
#define EVX_FPIEEE_SUBSSUIP 1985
#define EVX_FPIEEE_SUBT 161
#define EVX_FPIEEE_SUBTC 33
#define EVX_FPIEEE_SUBTM 97
#define EVX_FPIEEE_SUBTP 225
#define EVX_FPIEEE_SUBTU 417
#define EVX_FPIEEE_SUBTUC 289
#define EVX_FPIEEE_SUBTUM 353
#define EVX_FPIEEE_SUBTUP 481
#define EVX_FPIEEE_SUBTSU 1441
#define EVX_FPIEEE_SUBTSUC 1313
#define EVX_FPIEEE_SUBTSUM 1377
#define EVX_FPIEEE_SUBTSUP 1505
#define EVX_FPIEEE_SUBTSUI 1953
#define EVX_FPIEEE_SUBTSUIC 1825
#define EVX_FPIEEE_SUBTSUIM 1889
#define EVX_FPIEEE_SUBTSUIP 2017
#define EVX_FPIEEE_ADDS 128
#define EVX_FPIEEE_ADDSC 0
#define EVX_FPIEEE_ADDSM 64
#define EVX_FPIEEE_ADDSP 192
#define EVX_FPIEEE_ADDSU 384
#define EVX_FPIEEE_ADDSUC 256
#define EVX_FPIEEE_ADDSUM 320
#define EVX_FPIEEE_ADDSUP 448
#define EVX_FPIEEE_ADDSSU 1408
#define EVX_FPIEEE_ADDSSUC 1280
#define EVX_FPIEEE_ADDSSUM 1344
#define EVX_FPIEEE_ADDSSUP 1472
#define EVX_FPIEEE_ADDSSUI 1920
#define EVX_FPIEEE_ADDSSUIC 1792
#define EVX_FPIEEE_ADDSSUIM 1856
#define EVX_FPIEEE_ADDSSUIP 1984
#define EVX_FPIEEE_ADDT 160
#define EVX_FPIEEE_ADDTC 32
#define EVX_FPIEEE_ADDTM 96
#define EVX_FPIEEE_ADDTP 224
#define EVX_FPIEEE_ADDTU 416
#define EVX_FPIEEE_ADDTUC 288
#define EVX_FPIEEE_ADDTUM 352
#define EVX_FPIEEE_ADDTUP 480
#define EVX_FPIEEE_ADDTSU 1440
#define EVX_FPIEEE_ADDTSUC 1312
#define EVX_FPIEEE_ADDTSUM 1376
#define EVX_FPIEEE_ADDTSUP 1504
#define EVX_FPIEEE_ADDTSUI 1952
#define EVX_FPIEEE_ADDTSUIC 1824
#define EVX_FPIEEE_ADDTSUIM 1888
#define EVX_FPIEEE_ADDTSUIP 2016
#define EVX_FPIEEE_CVTSQ 143
#define EVX_FPIEEE_CVTSQC 15
#define EVX_FPIEEE_CVTSQV 399
#define EVX_FPIEEE_CVTSQVC 271
#define EVX_FPIEEE_CVTSQSV 1423
#define EVX_FPIEEE_CVTSQSVC 1295
#define EVX_FPIEEE_CVTSQSVI 1935
#define EVX_FPIEEE_CVTSQSVIC 1807
#define EVX_FPIEEE_CVTTQ 175
#define EVX_FPIEEE_CVTTQC 47
#define EVX_FPIEEE_CVTTQV 431
#define EVX_FPIEEE_CVTTQVC 303
#define EVX_FPIEEE_CVTTQSV 1455
#define EVX_FPIEEE_CVTTQSVC 1327
#define EVX_FPIEEE_CVTTQSVI 1967
#define EVX_FPIEEE_CVTTQSVIC 1839
#define EVX_FPIEEE_CVTTQD 239
#define EVX_FPIEEE_CVTTQVD 495
#define EVX_FPIEEE_CVTTQSVD 1519
#define EVX_FPIEEE_CVTTQSVID 2031
#define EVX_FPIEEE_CVTTQM 111
#define EVX_FPIEEE_CVTTQVM 367
#define EVX_FPIEEE_CVTTQSVM 1391
#define EVX_FPIEEE_CVTTQSVIM 1903
/* FUNCTION FIELDS for the DEC Floating Point instructions                  */
#define EVX_FPDEC_ADDD 144
#define EVX_FPDEC_ADDDC 16
#define EVX_FPDEC_ADDDU 400
#define EVX_FPDEC_ADDDUC 272
#define EVX_FPDEC_ADDDS 1168
#define EVX_FPDEC_ADDDSC 1040
#define EVX_FPDEC_ADDDSU 1424
#define EVX_FPDEC_ADDDSUC 1296
#define EVX_FPDEC_ADDF 128
#define EVX_FPDEC_ADDFC 0
#define EVX_FPDEC_ADDFU 384
#define EVX_FPDEC_ADDFUC 256
#define EVX_FPDEC_ADDFS 1152
#define EVX_FPDEC_ADDFSC 1024
#define EVX_FPDEC_ADDFSU 1408
#define EVX_FPDEC_ADDFSUC 1280
#define EVX_FPDEC_ADDG 160
#define EVX_FPDEC_ADDGC 32
#define EVX_FPDEC_ADDGU 416
#define EVX_FPDEC_ADDGUC 288
#define EVX_FPDEC_ADDGS 1184
#define EVX_FPDEC_ADDGSC 1056
#define EVX_FPDEC_ADDGSU 1440
#define EVX_FPDEC_ADDGSUC 1312
#define EVX_FPDEC_SUBD 145
#define EVX_FPDEC_SUBDC 17
#define EVX_FPDEC_SUBDU 401
#define EVX_FPDEC_SUBDUC 273
#define EVX_FPDEC_SUBDS 1169
#define EVX_FPDEC_SUBDSC 1041
#define EVX_FPDEC_SUBDSU 1425
#define EVX_FPDEC_SUBDSUC 1297
#define EVX_FPDEC_SUBF 129
#define EVX_FPDEC_SUBFC 1
#define EVX_FPDEC_SUBFU 385
#define EVX_FPDEC_SUBFUC 257
#define EVX_FPDEC_SUBFS 1153
#define EVX_FPDEC_SUBFSC 1025
#define EVX_FPDEC_SUBFSU 1409
#define EVX_FPDEC_SUBFSUC 1281
#define EVX_FPDEC_SUBG 161
#define EVX_FPDEC_SUBGC 33
#define EVX_FPDEC_SUBGU 417
#define EVX_FPDEC_SUBGUC 289
#define EVX_FPDEC_SUBGS 1185
#define EVX_FPDEC_SUBGSC 1057
#define EVX_FPDEC_SUBGSU 1441
#define EVX_FPDEC_SUBGSUC 1313
#define EVX_FPDEC_MULD 146
#define EVX_FPDEC_MULDC 18
#define EVX_FPDEC_MULDU 402
#define EVX_FPDEC_MULDUC 274
#define EVX_FPDEC_MULDS 1170
#define EVX_FPDEC_MULDSC 1042
#define EVX_FPDEC_MULDSU 1426
#define EVX_FPDEC_MULDSUC 1298
#define EVX_FPDEC_MULF 130
#define EVX_FPDEC_MULFC 2
#define EVX_FPDEC_MULFU 386
#define EVX_FPDEC_MULFUC 258
#define EVX_FPDEC_MULFS 1154
#define EVX_FPDEC_MULFSC 1026
#define EVX_FPDEC_MULFSU 1410
#define EVX_FPDEC_MULFSUC 1282
#define EVX_FPDEC_MULG 162
#define EVX_FPDEC_MULGC 34
#define EVX_FPDEC_MULGU 418
#define EVX_FPDEC_MULGUC 290
#define EVX_FPDEC_MULGS 1186
#define EVX_FPDEC_MULGSC 1058
#define EVX_FPDEC_MULGSU 1442
#define EVX_FPDEC_MULGSUC 1314
#define EVX_FPDEC_DIVD 147
#define EVX_FPDEC_DIVDC 19
#define EVX_FPDEC_DIVDU 403
#define EVX_FPDEC_DIVDUC 275
#define EVX_FPDEC_DIVDS 1171
#define EVX_FPDEC_DIVDSC 1043
#define EVX_FPDEC_DIVDSU 1427
#define EVX_FPDEC_DIVDSUC 1299
#define EVX_FPDEC_DIVF 131
#define EVX_FPDEC_DIVFC 3
#define EVX_FPDEC_DIVFU 387
#define EVX_FPDEC_DIVFUC 259
#define EVX_FPDEC_DIVFS 1155
#define EVX_FPDEC_DIVFSC 1027
#define EVX_FPDEC_DIVFSU 1411
#define EVX_FPDEC_DIVFSUC 1283
#define EVX_FPDEC_DIVG 163
#define EVX_FPDEC_DIVGC 35
#define EVX_FPDEC_DIVGU 419
#define EVX_FPDEC_DIVGUC 291
#define EVX_FPDEC_DIVGS 1187
#define EVX_FPDEC_DIVGSC 1059
#define EVX_FPDEC_DIVGSU 1443
#define EVX_FPDEC_DIVGSUC 1315
#define EVX_FPDEC_CMPDEQ 149
#define EVX_FPDEC_CMPDLT 150
#define EVX_FPDEC_CMPDLE 151
#define EVX_FPDEC_CMPGEQ 165
#define EVX_FPDEC_CMPGLT 166
#define EVX_FPDEC_CMPGLE 167
#define EVX_FPDEC_CMPDEQS 1173
#define EVX_FPDEC_CMPDLTS 1174
#define EVX_FPDEC_CMPDLES 1175
#define EVX_FPDEC_CMPGEQS 1189
#define EVX_FPDEC_CMPGLTS 1190
#define EVX_FPDEC_CMPGLES 1191
#define EVX_FPDEC_CVTDF 156
#define EVX_FPDEC_CVTDFC 28
#define EVX_FPDEC_CVTDFS 1180
#define EVX_FPDEC_CVTDFSC 1052
#define EVX_FPDEC_CVTDG 158
#define EVX_FPDEC_CVTDGC 30
#define EVX_FPDEC_CVTDGU 414
#define EVX_FPDEC_CVTDGUC 286
#define EVX_FPDEC_CVTDGS 1182
#define EVX_FPDEC_CVTDGSC 1054
#define EVX_FPDEC_CVTDGSU 1438
#define EVX_FPDEC_CVTDGSUC 1310
#define EVX_FPDEC_CVTGD 173
#define EVX_FPDEC_CVTGDC 45
#define EVX_FPDEC_CVTGDU 429
#define EVX_FPDEC_CVTGDUC 301
#define EVX_FPDEC_CVTGDS 1197
#define EVX_FPDEC_CVTGDSC 1069
#define EVX_FPDEC_CVTGDSU 1453
#define EVX_FPDEC_CVTGDSUC 1325
#define EVX_FPDEC_CVTFG 142
#define EVX_FPDEC_CVTFGC 14
#define EVX_FPDEC_CVTFGS 1166
#define EVX_FPDEC_CVTFGSC 1038
#define EVX_FPDEC_CVTGF 172
#define EVX_FPDEC_CVTGFC 44
#define EVX_FPDEC_CVTGFU 428
#define EVX_FPDEC_CVTGFUC 300
#define EVX_FPDEC_CVTGFS 1196
#define EVX_FPDEC_CVTGFSC 1068
#define EVX_FPDEC_CVTGFSU 1452
#define EVX_FPDEC_CVTGFSUC 1324
#define EVX_FPDEC_CVTQD 189
#define EVX_FPDEC_CVTQDC 61
#define EVX_FPDEC_CVTQDS 1213
#define EVX_FPDEC_CVTQDSC 1085
#define EVX_FPDEC_CVTQF 188
#define EVX_FPDEC_CVTQFC 60
#define EVX_FPDEC_CVTQFS 1212
#define EVX_FPDEC_CVTQFSC 1084
#define EVX_FPDEC_CVTQG 190
#define EVX_FPDEC_CVTQGC 62
#define EVX_FPDEC_CVTQGS 1214
#define EVX_FPDEC_CVTQGSC 1086
#define EVX_FPDEC_CVTDQ 159
#define EVX_FPDEC_CVTDQC 31
#define EVX_FPDEC_CVTDQV 415
#define EVX_FPDEC_CVTDQVC 287
#define EVX_FPDEC_CVTDQS 1183
#define EVX_FPDEC_CVTDQSC 1055
#define EVX_FPDEC_CVTDQSV 1439
#define EVX_FPDEC_CVTDQSVC 1311
#define EVX_FPDEC_CVTGQ 175
#define EVX_FPDEC_CVTGQC 47
#define EVX_FPDEC_CVTGQV 431
#define EVX_FPDEC_CVTGQVC 303
#define EVX_FPDEC_CVTGQS 1199
#define EVX_FPDEC_CVTGQSC 1071
#define EVX_FPDEC_CVTGQSV 1455
#define EVX_FPDEC_CVTGQSVC 1327
/* Opcode/Functions for the PAL code instructions                           */
#define EVX_OPC_PAL 0
#define EVX_PAL_BPT 128
#define EVX_PAL_BUGCHK 129
#define EVX_PAL_CHME 130
#define EVX_PAL_CHMK 131
#define EVX_PAL_CHMS 132
#define EVX_PAL_CHMU 133
#define EVX_PAL_IMB 134
#define EVX_PAL_INSQHIL 135
#define EVX_PAL_INSQTIL 136
#define EVX_PAL_INSQHIQ 137
#define EVX_PAL_INSQTIQ 138
#define EVX_PAL_INSQUEL 139
#define EVX_PAL_INSQUEQ 140
#define EVX_PAL_INSQUELD 141
#define EVX_PAL_INSQUEQD 142
#define EVX_PAL_PROBER 143
#define EVX_PAL_PROBEW 144
#define EVX_PAL_RD_PS 145
#define EVX_PAL_REI 146
#define EVX_PAL_REMQHIL 147
#define EVX_PAL_REMQTIL 148
#define EVX_PAL_REMQHIQ 149
#define EVX_PAL_REMQTIQ 150
#define EVX_PAL_REMQUEL 151
#define EVX_PAL_REMQUEQ 152
#define EVX_PAL_REMQUELD 153
#define EVX_PAL_REMQUEQD 154
#define EVX_PAL_SWASTEN 155
#define EVX_PAL_WR_PS_SW 156
#define EVX_PAL_RSCC 157
#define EVX_PAL_READ_UNQ 158
#define EVX_PAL_WRITE_UNQ 159
#define EVX_PAL_CFLUSH 1
#define EVX_PAL_DRAINA 2
#define EVX_PAL_HALT 0
#define EVX_PAL_LDQP 3
#define EVX_PAL_STQP 4
#define EVX_PAL_SWPCTX 5
#define EVX_PAL_MFPR_ASN 6
#define EVX_PAL_MTPR_ASTEN 7
#define EVX_PAL_MTPR_ASTSR 8
#define EVX_PAL_MFPR_AT 9
#define EVX_PAL_MTPR_AT 10
#define EVX_PAL_MFPR_FEN 11
#define EVX_PAL_MTPR_FEN 12
#define EVX_PAL_MTPR_IPIR 13
#define EVX_PAL_MFPR_IPL 14
#define EVX_PAL_MTPR_IPL 15
#define EVX_PAL_MFPR_MCES 16
#define EVX_PAL_MTPR_MCES 17
#define EVX_PAL_MFPR_PCBB 18
#define EVX_PAL_MFPR_PRBR 19
#define EVX_PAL_MTPR_PRBR 20
#define EVX_PAL_MFPR_PTBR 21
#define EVX_PAL_MFPR_SCBB 22
#define EVX_PAL_MTPR_SCBB 23
#define EVX_PAL_MTPR_SIRR 24
#define EVX_PAL_MFPR_SISR 25
#define EVX_PAL_MFPR_TBCHK 26
#define EVX_PAL_MTPR_TBIA 27
#define EVX_PAL_MTPR_TBIAP 28
#define EVX_PAL_MTPR_TBIS 29
#define EVX_PAL_MFPR_ESP 30
#define EVX_PAL_MTPR_ESP 31
#define EVX_PAL_MFPR_SSP 32
#define EVX_PAL_MTPR_SSP 33
#define EVX_PAL_MFPR_USP 34
#define EVX_PAL_MTPR_USP 35
#define EVX_PAL_MTPR_TBISD 36
#define EVX_PAL_MTPR_TBISI 37
#define EVX_PAL_MFPR_ASTEN 38
#define EVX_PAL_MFPR_ASTSR 39
#define EVX_PAL_MFPR_WHAMI 63
#define EVX_PAL_RESERVED_GETSS 123
#define EVX_PAL_RESERVED_PUTSS 124
#define EVX_PAL_RESERVED_RLS1 125
#define EVX_PAL_RESERVED_RLS2 126
#define EVX_PAL_RESERVED_DVT1 127
#define EVX_PAL_NPHALT 251
#define EVX_PAL_RESERVED_DVT 254
#define EVX_PAL_RESERVED_SIM 255
/* Opcodes reserved for PALcode                                             */
#define EVX_OPC_PAL19 25
#define EVX_OPC_PAL1B 27
#define EVX_OPC_PAL1D 29
#define EVX_OPC_PAL1E 30
#define EVX_OPC_PAL1F 31
/* Opcodes reserved for DIGITAL                                             */
#define EVX_OPC_OPC01 1
#define EVX_OPC_OPC02 2
#define EVX_OPC_OPC03 3
#define EVX_OPC_OPC04 4
#define EVX_OPC_OPC05 5
#define EVX_OPC_OPC06 6
#define EVX_OPC_OPC07 7
#define EVX_OPC_OPC0A 10
#define EVX_OPC_OPC0C 12
#define EVX_OPC_OPC0D 13
#define EVX_OPC_OPC0E 14
#define EVX_OPC_OPC14 20
#define EVX_OPC_OPC1C 28
/* opcode and subfunction in correct location in long word                  */
/* suitable for OR'ing directly into the target instruction                 */
#define EVX_opc32_LDL -1610612736
#define EVX_opc32_LDQ -1543503872
#define EVX_opc32_LDLL -1476395008
#define EVX_opc32_LDQL -1409286144
#define EVX_opc32_LDQU 738197504
#define EVX_opc32_LDA 536870912
#define EVX_opc32_STL -1342177280
#define EVX_opc32_STQ -1275068416
#define EVX_opc32_STLC -1207959552
#define EVX_opc32_STQC -1140850688
#define EVX_opc32_STQU 1006632960
#define EVX_opc32_LDAH 603979776
#define EVX_opc32_LDF -2147483648
#define EVX_opc32_LDD -2080374784
#define EVX_opc32_LDS -2013265920
#define EVX_opc32_LDT -1946157056
#define EVX_opc32_STF -1879048192
#define EVX_opc32_STD -1811939328
#define EVX_opc32_STS -1744830464
#define EVX_opc32_STT -1677721600
#define EVX_opc32_MB 1612660736
#define EVX_opc32_TRAPB 1612644352
#define EVX_opc32_RCC 1612693504
#define EVX_opc32_FETCH 1610645504
#define EVX_opc32_FETCH_M 1610653696
#define EVX_opc32_RC 1612701696
#define EVX_opc32_RS 1612705792
#define EVX_opc32_JMP 1744830464
#define EVX_opc32_JSR 1744846848
#define EVX_opc32_RET 1744863232
#define EVX_opc32_JSR_COROUTINE 1744879616
#define EVX_opc32_BR -1073741824
#define EVX_opc32_BSR -805306368
#define EVX_opc32_BLBC -536870912
#define EVX_opc32_BLBS -268435456
#define EVX_opc32_FBEQ -1006632960
#define EVX_opc32_FBNE -738197504
#define EVX_opc32_BEQ -469762048
#define EVX_opc32_BNE -201326592
#define EVX_opc32_FBLT -939524096
#define EVX_opc32_FBGE -671088640
#define EVX_opc32_BLT -402653184
#define EVX_opc32_BGE -134217728
#define EVX_opc32_FBLE -872415232
#define EVX_opc32_FBGT -603979776
#define EVX_opc32_BLE -335544320
#define EVX_opc32_BGT -67108864
#define EVX_opc32_ADDL 1073741824
#define EVX_opc32_ADDLV 1073743872
#define EVX_opc32_ADDQ 1073742848
#define EVX_opc32_ADDQV 1073744896
#define EVX_opc32_S4ADDL 1073741888
#define EVX_opc32_S4ADDQ 1073742912
#define EVX_opc32_S8ADDL 1073742400
#define EVX_opc32_S8ADDQ 1073743424
#define EVX_opc32_S4SUBL 1073742176
#define EVX_opc32_S4SUBQ 1073743200
#define EVX_opc32_S8SUBL 1073742688
#define EVX_opc32_S8SUBQ 1073743712
#define EVX_opc32_SUBL 1073742112
#define EVX_opc32_SUBLV 1073744160
#define EVX_opc32_SUBQ 1073743136
#define EVX_opc32_SUBQV 1073745184
#define EVX_opc32_CMPEQ 1073743264
#define EVX_opc32_CMPLT 1073744288
#define EVX_opc32_CMPLE 1073745312
#define EVX_opc32_CMPULT 1073742752
#define EVX_opc32_CMPULE 1073743776
#define EVX_opc32_CMPBGE 1073742304
#define EVX_opc32_AND 1140850688
#define EVX_opc32_BIC 1140850944
#define EVX_opc32_CMOVEQ 1140851840
#define EVX_opc32_CMOVNE 1140851904
#define EVX_opc32_BIS 1140851712
#define EVX_opc32_ORNOT 1140851968
#define EVX_opc32_CMOVLT 1140852864
#define EVX_opc32_CMOVGE 1140852928
#define EVX_opc32_XOR 1140852736
#define EVX_opc32_EQV 1140852992
#define EVX_opc32_CMOVLE 1140853888
#define EVX_opc32_CMOVGT 1140853952
#define EVX_opc32_CMOVLBS 1140851328
#define EVX_opc32_CMOVLBC 1140851392
#define EVX_opc32_EXTBL 1207959744
#define EVX_opc32_EXTWL 1207960256
#define EVX_opc32_EXTLL 1207960768
#define EVX_opc32_EXTQL 1207961280
#define EVX_opc32_EXTWH 1207962432
#define EVX_opc32_EXTLH 1207962944
#define EVX_opc32_EXTQH 1207963456
#define EVX_opc32_INSBL 1207959904
#define EVX_opc32_INSWL 1207960416
#define EVX_opc32_INSLL 1207960928
#define EVX_opc32_INSQL 1207961440
#define EVX_opc32_INSWH 1207962336
#define EVX_opc32_INSLH 1207962848
#define EVX_opc32_INSQH 1207963360
#define EVX_opc32_MSKBL 1207959616
#define EVX_opc32_MSKWL 1207960128
#define EVX_opc32_MSKLL 1207960640
#define EVX_opc32_MSKQL 1207961152
#define EVX_opc32_MSKWH 1207962176
#define EVX_opc32_MSKLH 1207962688
#define EVX_opc32_MSKQH 1207963200
#define EVX_opc32_ZAP 1207961088
#define EVX_opc32_ZAPNOT 1207961120
#define EVX_opc32_SRA 1207961472
#define EVX_opc32_SRL 1207961216
#define EVX_opc32_SLL 1207961376
#define EVX_opc32_MULL 1275068416
#define EVX_opc32_MULQV 1275071488
#define EVX_opc32_MULLV 1275070464
#define EVX_opc32_UMULH 1275069952
#define EVX_opc32_MULQ 1275069440
#define EVX_opc32_CPYS 1543504896
#define EVX_opc32_CPYSN 1543504928
#define EVX_opc32_CPYSE 1543504960
#define EVX_opc32_CPYSEE 1543503968
#define EVX_opc32_WR_FPCTL 1543505024
#define EVX_opc32_RD_FPCTL 1543505056
#define EVX_opc32_CVTLQ 1543504384
#define EVX_opc32_CVTQL 1543505408
#define EVX_opc32_CVTQLV 1543513600
#define EVX_opc32_CVTQLSV 1543546368
#define EVX_opc32_FCMOVEQ 1543505216
#define EVX_opc32_FCMOVNE 1543505248
#define EVX_opc32_FCMOVLT 1543505280
#define EVX_opc32_FCMOVGE 1543505312
#define EVX_opc32_FCMOVLE 1543505344
#define EVX_opc32_FCMOVGT 1543505376
#define EVX_opc32_ADDS 1476399104
#define EVX_opc32_ADDSC 1476395008
#define EVX_opc32_ADDSM 1476397056
#define EVX_opc32_ADDSP 1476401152
#define EVX_opc32_ADDSU 1476407296
#define EVX_opc32_ADDSUC 1476403200
#define EVX_opc32_ADDSUM 1476405248
#define EVX_opc32_ADDSUP 1476409344
#define EVX_opc32_ADDSSU 1476440064
#define EVX_opc32_ADDSSUC 1476435968
#define EVX_opc32_ADDSSUM 1476438016
#define EVX_opc32_ADDSSUP 1476442112
#define EVX_opc32_ADDSSUI 1476456448
#define EVX_opc32_ADDSSUIC 1476452352
#define EVX_opc32_ADDSSUIM 1476454400
#define EVX_opc32_ADDSSUIP 1476458496
#define EVX_opc32_ADDT 1476400128
#define EVX_opc32_ADDTC 1476396032
#define EVX_opc32_ADDTM 1476398080
#define EVX_opc32_ADDTP 1476402176
#define EVX_opc32_ADDTU 1476408320
#define EVX_opc32_ADDTUC 1476404224
#define EVX_opc32_ADDTUM 1476406272
#define EVX_opc32_ADDTUP 1476410368
#define EVX_opc32_ADDTSU 1476441088
#define EVX_opc32_ADDTSUC 1476436992
#define EVX_opc32_ADDTSUM 1476439040
#define EVX_opc32_ADDTSUP 1476443136
#define EVX_opc32_ADDTSUI 1476457472
#define EVX_opc32_ADDTSUIC 1476453376
#define EVX_opc32_ADDTSUIM 1476455424
#define EVX_opc32_ADDTSUIP 1476459520
#define EVX_opc32_CMPSEQ 1476399264
#define EVX_opc32_CMPSEQSU 1476440224
#define EVX_opc32_CMPSLT 1476399296
#define EVX_opc32_CMPSLTSU 1476440256
#define EVX_opc32_CMPSLE 1476399328
#define EVX_opc32_CMPSLESU 1476440288
#define EVX_opc32_CMPSUN 1476399232
#define EVX_opc32_CMPSUNSU 1476440192
#define EVX_opc32_CMPTEQ 1476400288
#define EVX_opc32_CMPTEQSU 1476441248
#define EVX_opc32_CMPTLT 1476400320
#define EVX_opc32_CMPTLTSU 1476441280
#define EVX_opc32_CMPTLE 1476400352
#define EVX_opc32_CMPTLESU 1476441312
#define EVX_opc32_CMPTUN 1476400256
#define EVX_opc32_CMPTUNSU 1476441216
#define EVX_opc32_CVTQS 1476401024
#define EVX_opc32_CVTQSC 1476396928
#define EVX_opc32_CVTQSM 1476398976
#define EVX_opc32_CVTQSP 1476403072
#define EVX_opc32_CVTQSSUI 1476458368
#define EVX_opc32_CVTQSSUIC 1476454272
#define EVX_opc32_CVTQSSUIM 1476456320
#define EVX_opc32_CVTQSSUIP 1476460416
#define EVX_opc32_CVTQT 1476401088
#define EVX_opc32_CVTQTC 1476396992
#define EVX_opc32_CVTQTM 1476399040
#define EVX_opc32_CVTQTP 1476403136
#define EVX_opc32_CVTQTSUI 1476458432
#define EVX_opc32_CVTQTSUIC 1476454336
#define EVX_opc32_CVTQTSUIM 1476456384
#define EVX_opc32_CVTQTSUIP 1476460480
#define EVX_opc32_CVTST 1476399552
#define EVX_opc32_CVTSTC 1476395456
#define EVX_opc32_CVTSTM 1476397504
#define EVX_opc32_CVTSTP 1476401600
#define EVX_opc32_CVTSTSUI 1476456896
#define EVX_opc32_CVTSTSUIC 1476452800
#define EVX_opc32_CVTSTSUIM 1476454848
#define EVX_opc32_CVTSTSUIP 1476458944
#define EVX_opc32_CVTTS 1476400512
#define EVX_opc32_CVTTSC 1476396416
#define EVX_opc32_CVTTSM 1476398464
#define EVX_opc32_CVTTSP 1476402560
#define EVX_opc32_CVTTSU 1476408704
#define EVX_opc32_CVTTSUC 1476404608
#define EVX_opc32_CVTTSUM 1476406656
#define EVX_opc32_CVTTSUP 1476410752
#define EVX_opc32_CVTTSSU 1476441472
#define EVX_opc32_CVTTSSUC 1476437376
#define EVX_opc32_CVTTSSUM 1476439424
#define EVX_opc32_CVTTSSUP 1476443520
#define EVX_opc32_CVTTSSUI 1476457856
#define EVX_opc32_CVTTSSUIC 1476453760
#define EVX_opc32_CVTTSSUIM 1476455808
#define EVX_opc32_CVTTSSUIP 1476459904
#define EVX_opc32_DIVS 1476399200
#define EVX_opc32_DIVSC 1476395104
#define EVX_opc32_DIVSM 1476397152
#define EVX_opc32_DIVSP 1476401248
#define EVX_opc32_DIVSU 1476407392
#define EVX_opc32_DIVSUC 1476403296
#define EVX_opc32_DIVSUM 1476405344
#define EVX_opc32_DIVSUP 1476409440
#define EVX_opc32_DIVSSU 1476440160
#define EVX_opc32_DIVSSUC 1476436064
#define EVX_opc32_DIVSSUM 1476438112
#define EVX_opc32_DIVSSUP 1476442208
#define EVX_opc32_DIVSSUI 1476456544
#define EVX_opc32_DIVSSUIC 1476452448
#define EVX_opc32_DIVSSUIM 1476454496
#define EVX_opc32_DIVSSUIP 1476458592
#define EVX_opc32_DIVT 1476400224
#define EVX_opc32_DIVTC 1476396128
#define EVX_opc32_DIVTM 1476398176
#define EVX_opc32_DIVTP 1476402272
#define EVX_opc32_DIVTU 1476408416
#define EVX_opc32_DIVTUC 1476404320
#define EVX_opc32_DIVTUM 1476406368
#define EVX_opc32_DIVTUP 1476410464
#define EVX_opc32_DIVTSU 1476441184
#define EVX_opc32_DIVTSUC 1476437088
#define EVX_opc32_DIVTSUM 1476439136
#define EVX_opc32_DIVTSUP 1476443232
#define EVX_opc32_DIVTSUI 1476457568
#define EVX_opc32_DIVTSUIC 1476453472
#define EVX_opc32_DIVTSUIM 1476455520
#define EVX_opc32_DIVTSUIP 1476459616
#define EVX_opc32_MULS 1476399168
#define EVX_opc32_MULSC 1476395072
#define EVX_opc32_MULSM 1476397120
#define EVX_opc32_MULSP 1476401216
#define EVX_opc32_MULSU 1476407360
#define EVX_opc32_MULSUC 1476403264
#define EVX_opc32_MULSUM 1476405312
#define EVX_opc32_MULSUP 1476409408
#define EVX_opc32_MULSSU 1476440128
#define EVX_opc32_MULSSUC 1476436032
#define EVX_opc32_MULSSUM 1476438080
#define EVX_opc32_MULSSUP 1476442176
#define EVX_opc32_MULSSUI 1476456512
#define EVX_opc32_MULSSUIC 1476452416
#define EVX_opc32_MULSSUIM 1476454464
#define EVX_opc32_MULSSUIP 1476458560
#define EVX_opc32_MULT 1476400192
#define EVX_opc32_MULTC 1476396096
#define EVX_opc32_MULTM 1476398144
#define EVX_opc32_MULTP 1476402240
#define EVX_opc32_MULTU 1476408384
#define EVX_opc32_MULTUC 1476404288
#define EVX_opc32_MULTUM 1476406336
#define EVX_opc32_MULTUP 1476410432
#define EVX_opc32_MULTSU 1476441152
#define EVX_opc32_MULTSUC 1476437056
#define EVX_opc32_MULTSUM 1476439104
#define EVX_opc32_MULTSUP 1476443200
#define EVX_opc32_MULTSUI 1476457536
#define EVX_opc32_MULTSUIC 1476453440
#define EVX_opc32_MULTSUIM 1476455488
#define EVX_opc32_MULTSUIP 1476459584
#define EVX_opc32_SUBS 1476399136
#define EVX_opc32_SUBSC 1476395040
#define EVX_opc32_SUBSM 1476397088
#define EVX_opc32_SUBSP 1476401184
#define EVX_opc32_SUBSU 1476407328
#define EVX_opc32_SUBSUC 1476403232
#define EVX_opc32_SUBSUM 1476405280
#define EVX_opc32_SUBSUP 1476409376
#define EVX_opc32_SUBSSU 1476440096
#define EVX_opc32_SUBSSUC 1476436000
#define EVX_opc32_SUBSSUM 1476438048
#define EVX_opc32_SUBSSUP 1476442144
#define EVX_opc32_SUBSSUI 1476456480
#define EVX_opc32_SUBSSUIC 1476452384
#define EVX_opc32_SUBSSUIM 1476454432
#define EVX_opc32_SUBSSUIP 1476458528
#define EVX_opc32_SUBT 1476400160
#define EVX_opc32_SUBTC 1476396064
#define EVX_opc32_SUBTM 1476398112
#define EVX_opc32_SUBTP 1476402208
#define EVX_opc32_SUBTU 1476408352
#define EVX_opc32_SUBTUC 1476404256
#define EVX_opc32_SUBTUM 1476406304
#define EVX_opc32_SUBTUP 1476410400
#define EVX_opc32_SUBTSU 1476441120
#define EVX_opc32_SUBTSUC 1476437024
#define EVX_opc32_SUBTSUM 1476439072
#define EVX_opc32_SUBTSUP 1476443168
#define EVX_opc32_SUBTSUI 1476457504
#define EVX_opc32_SUBTSUIC 1476453408
#define EVX_opc32_SUBTSUIM 1476455456
#define EVX_opc32_SUBTSUIP 1476459552
#define EVX_opc32_CVTSQ 1476399584
#define EVX_opc32_CVTSQC 1476395488
#define EVX_opc32_CVTSQV 1476407776
#define EVX_opc32_CVTSQVC 1476403680
#define EVX_opc32_CVTSQSV 1476440544
#define EVX_opc32_CVTSQSVC 1476436448
#define EVX_opc32_CVTSQSVI 1476456928
#define EVX_opc32_CVTSQSVIC 1476452832
#define EVX_opc32_CVTTQ 1476400608
#define EVX_opc32_CVTTQC 1476396512
#define EVX_opc32_CVTTQV 1476408800
#define EVX_opc32_CVTTQVC 1476404704
#define EVX_opc32_CVTTQSV 1476441568
#define EVX_opc32_CVTTQSVC 1476437472
#define EVX_opc32_CVTTQSVI 1476457952
#define EVX_opc32_CVTTQSVIC 1476453856
#define EVX_opc32_CVTTQD 1476402656
#define EVX_opc32_CVTTQVD 1476410848
#define EVX_opc32_CVTTQSVD 1476443616
#define EVX_opc32_CVTTQSVID 1476460000
#define EVX_opc32_CVTTQM 1476398560
#define EVX_opc32_CVTTQVM 1476406752
#define EVX_opc32_CVTTQSVM 1476439520
#define EVX_opc32_CVTTQSVIM 1476455904
#define EVX_opc32_ADDD 1409290752
#define EVX_opc32_ADDDC 1409286656
#define EVX_opc32_ADDDU 1409298944
#define EVX_opc32_ADDDUC 1409294848
#define EVX_opc32_ADDDS 1409323520
#define EVX_opc32_ADDDSC 1409319424
#define EVX_opc32_ADDDSU 1409331712
#define EVX_opc32_ADDDSUC 1409327616
#define EVX_opc32_ADDF 1409290240
#define EVX_opc32_ADDFC 1409286144
#define EVX_opc32_ADDFU 1409298432
#define EVX_opc32_ADDFUC 1409294336
#define EVX_opc32_ADDFS 1409323008
#define EVX_opc32_ADDFSC 1409318912
#define EVX_opc32_ADDFSU 1409331200
#define EVX_opc32_ADDFSUC 1409327104
#define EVX_opc32_ADDG 1409291264
#define EVX_opc32_ADDGC 1409287168
#define EVX_opc32_ADDGU 1409299456
#define EVX_opc32_ADDGUC 1409295360
#define EVX_opc32_ADDGS 1409324032
#define EVX_opc32_ADDGSC 1409319936
#define EVX_opc32_ADDGSU 1409332224
#define EVX_opc32_ADDGSUC 1409328128
#define EVX_opc32_CMPDEQ 1409290912
#define EVX_opc32_CMPDLT 1409290944
#define EVX_opc32_CMPDLE 1409290976
#define EVX_opc32_CMPDEQS 1409323680
#define EVX_opc32_CMPDLTS 1409323712
#define EVX_opc32_CMPDLES 1409323744
#define EVX_opc32_CMPGEQ 1409291424
#define EVX_opc32_CMPGLT 1409291456
#define EVX_opc32_CMPGLE 1409291488
#define EVX_opc32_CMPGEQS 1409324192
#define EVX_opc32_CMPGLTS 1409324224
#define EVX_opc32_CMPGLES 1409324256
#define EVX_opc32_CVTDF 1409291136
#define EVX_opc32_CVTDFC 1409287040
#define EVX_opc32_CVTDFS 1409323904
#define EVX_opc32_CVTDFSC 1409319808
#define EVX_opc32_CVTDG 1409291200
#define EVX_opc32_CVTDGC 1409287104
#define EVX_opc32_CVTDGU 1409299392
#define EVX_opc32_CVTDGUC 1409295296
#define EVX_opc32_CVTDGS 1409323968
#define EVX_opc32_CVTDGSC 1409319872
#define EVX_opc32_CVTDGSU 1409332160
#define EVX_opc32_CVTDGSUC 1409328064
#define EVX_opc32_CVTGD 1409291680
#define EVX_opc32_CVTGDC 1409287584
#define EVX_opc32_CVTGDU 1409299872
#define EVX_opc32_CVTGDUC 1409295776
#define EVX_opc32_CVTGDS 1409324448
#define EVX_opc32_CVTGDSC 1409320352
#define EVX_opc32_CVTGDSU 1409332640
#define EVX_opc32_CVTGDSUC 1409328544
#define EVX_opc32_CVTFG 1409290688
#define EVX_opc32_CVTFGC 1409286592
#define EVX_opc32_CVTFGS 1409323456
#define EVX_opc32_CVTFGSC 1409319360
#define EVX_opc32_CVTGF 1409291648
#define EVX_opc32_CVTGFC 1409287552
#define EVX_opc32_CVTGFU 1409299840
#define EVX_opc32_CVTGFUC 1409295744
#define EVX_opc32_CVTGFS 1409324416
#define EVX_opc32_CVTGFSC 1409320320
#define EVX_opc32_CVTGFSU 1409332608
#define EVX_opc32_CVTGFSUC 1409328512
#define EVX_opc32_CVTQD 1409292192
#define EVX_opc32_CVTQDC 1409288096
#define EVX_opc32_CVTQDS 1409324960
#define EVX_opc32_CVTQDSC 1409320864
#define EVX_opc32_CVTQF 1409292160
#define EVX_opc32_CVTQFC 1409288064
#define EVX_opc32_CVTQFS 1409324928
#define EVX_opc32_CVTQFSC 1409320832
#define EVX_opc32_CVTQG 1409292224
#define EVX_opc32_CVTQGC 1409288128
#define EVX_opc32_CVTQGS 1409324992
#define EVX_opc32_CVTQGSC 1409320896
#define EVX_opc32_DIVD 1409290848
#define EVX_opc32_DIVDC 1409286752
#define EVX_opc32_DIVDU 1409299040
#define EVX_opc32_DIVDUC 1409294944
#define EVX_opc32_DIVDS 1409323616
#define EVX_opc32_DIVDSC 1409319520
#define EVX_opc32_DIVDSU 1409331808
#define EVX_opc32_DIVDSUC 1409327712
#define EVX_opc32_DIVF 1409290336
#define EVX_opc32_DIVFC 1409286240
#define EVX_opc32_DIVFU 1409298528
#define EVX_opc32_DIVFUC 1409294432
#define EVX_opc32_DIVFS 1409323104
#define EVX_opc32_DIVFSC 1409319008
#define EVX_opc32_DIVFSU 1409331296
#define EVX_opc32_DIVFSUC 1409327200
#define EVX_opc32_DIVG 1409291360
#define EVX_opc32_DIVGC 1409287264
#define EVX_opc32_DIVGU 1409299552
#define EVX_opc32_DIVGUC 1409295456
#define EVX_opc32_DIVGS 1409324128
#define EVX_opc32_DIVGSC 1409320032
#define EVX_opc32_DIVGSU 1409332320
#define EVX_opc32_DIVGSUC 1409328224
#define EVX_opc32_MULD 1409290816
#define EVX_opc32_MULDC 1409286720
#define EVX_opc32_MULDU 1409299008
#define EVX_opc32_MULDUC 1409294912
#define EVX_opc32_MULDS 1409323584
#define EVX_opc32_MULDSC 1409319488
#define EVX_opc32_MULDSU 1409331776
#define EVX_opc32_MULDSUC 1409327680
#define EVX_opc32_MULF 1409290304
#define EVX_opc32_MULFC 1409286208
#define EVX_opc32_MULFU 1409298496
#define EVX_opc32_MULFUC 1409294400
#define EVX_opc32_MULFS 1409323072
#define EVX_opc32_MULFSC 1409318976
#define EVX_opc32_MULFSU 1409331264
#define EVX_opc32_MULFSUC 1409327168
#define EVX_opc32_MULG 1409291328
#define EVX_opc32_MULGC 1409287232
#define EVX_opc32_MULGU 1409299520
#define EVX_opc32_MULGUC 1409295424
#define EVX_opc32_MULGS 1409324096
#define EVX_opc32_MULGSC 1409320000
#define EVX_opc32_MULGSU 1409332288
#define EVX_opc32_MULGSUC 1409328192
#define EVX_opc32_SUBD 1409290784
#define EVX_opc32_SUBDC 1409286688
#define EVX_opc32_SUBDU 1409298976
#define EVX_opc32_SUBDUC 1409294880
#define EVX_opc32_SUBDS 1409323552
#define EVX_opc32_SUBDSC 1409319456
#define EVX_opc32_SUBDSU 1409331744
#define EVX_opc32_SUBDSUC 1409327648
#define EVX_opc32_SUBF 1409290272
#define EVX_opc32_SUBFC 1409286176
#define EVX_opc32_SUBFU 1409298464
#define EVX_opc32_SUBFUC 1409294368
#define EVX_opc32_SUBFS 1409323040
#define EVX_opc32_SUBFSC 1409318944
#define EVX_opc32_SUBFSU 1409331232
#define EVX_opc32_SUBFSUC 1409327136
#define EVX_opc32_SUBG 1409291296
#define EVX_opc32_SUBGC 1409287200
#define EVX_opc32_SUBGU 1409299488
#define EVX_opc32_SUBGUC 1409295392
#define EVX_opc32_SUBGS 1409324064
#define EVX_opc32_SUBGSC 1409319968
#define EVX_opc32_SUBGSU 1409332256
#define EVX_opc32_SUBGSUC 1409328160
#define EVX_opc32_CVTDQ 1409291232
#define EVX_opc32_CVTDQC 1409287136
#define EVX_opc32_CVTDQV 1409299424
#define EVX_opc32_CVTDQVC 1409295328
#define EVX_opc32_CVTDQS 1409324000
#define EVX_opc32_CVTDQSC 1409319904
#define EVX_opc32_CVTDQSV 1409332192
#define EVX_opc32_CVTDQSVC 1409328096
#define EVX_opc32_CVTGQ 1409291744
#define EVX_opc32_CVTGQC 1409287648
#define EVX_opc32_CVTGQV 1409299936
#define EVX_opc32_CVTGQVC 1409295840
#define EVX_opc32_CVTGQS 1409324512
#define EVX_opc32_CVTGQSC 1409320416
#define EVX_opc32_CVTGQSV 1409332704
#define EVX_opc32_CVTGQSVC 1409328608
#define EVX_opc32_BPT 128
#define EVX_opc32_BUGCHK 129
#define EVX_opc32_CHME 130
#define EVX_opc32_CHMK 131
#define EVX_opc32_CHMS 132
#define EVX_opc32_CHMU 133
#define EVX_opc32_IMB 134
#define EVX_opc32_INSQHIL 135
#define EVX_opc32_INSQTIL 136
#define EVX_opc32_INSQHIQ 137
#define EVX_opc32_INSQTIQ 138
#define EVX_opc32_INSQUEL 139
#define EVX_opc32_INSQUEQ 140
#define EVX_opc32_INSQUELD 141
#define EVX_opc32_INSQUEQD 142
#define EVX_opc32_PROBER 143
#define EVX_opc32_PROBEW 144
#define EVX_opc32_RD_PS 145
#define EVX_opc32_REI 146
#define EVX_opc32_REMQHIL 147
#define EVX_opc32_REMQTIL 148
#define EVX_opc32_REMQHIQ 149
#define EVX_opc32_REMQTIQ 150
#define EVX_opc32_REMQUEL 151
#define EVX_opc32_REMQUEQ 152
#define EVX_opc32_REMQUELD 153
#define EVX_opc32_REMQUEQD 154
#define EVX_opc32_SWASTEN 155
#define EVX_opc32_WR_PS_SW 156
#define EVX_opc32_RSCC 157
#define EVX_opc32_READ_UNQ 158
#define EVX_opc32_WRITE_UNQ 159
#define EVX_opc32_CFLUSH 1
#define EVX_opc32_DRAINA 2
#define EVX_opc32_HALT 0
#define EVX_opc32_LDQP 3
#define EVX_opc32_STQP 4
#define EVX_opc32_SWPCTX 5
#define EVX_opc32_MFPR_ASN 6
#define EVX_opc32_MTPR_ASTEN 7
#define EVX_opc32_MTPR_ASTSR 8
#define EVX_opc32_MFPR_AT 9
#define EVX_opc32_MTPR_AT 10
#define EVX_opc32_MFPR_FEN 11
#define EVX_opc32_MTPR_FEN 12
#define EVX_opc32_MTPR_IPIR 13
#define EVX_opc32_MFPR_IPL 14
#define EVX_opc32_MTPR_IPL 15
#define EVX_opc32_MFPR_MCES 16
#define EVX_opc32_MTPR_MCES 17
#define EVX_opc32_MFPR_PCBB 18
#define EVX_opc32_MFPR_PRBR 19
#define EVX_opc32_MTPR_PRBR 20
#define EVX_opc32_MFPR_PTBR 21
#define EVX_opc32_MFPR_SCBB 22
#define EVX_opc32_MTPR_SCBB 23
#define EVX_opc32_MTPR_SIRR 24
#define EVX_opc32_MFPR_SISR 25
#define EVX_opc32_MFPR_TBCHK 26
#define EVX_opc32_MTPR_TBIA 27
#define EVX_opc32_MTPR_TBIAP 28
#define EVX_opc32_MTPR_TBIS 29
#define EVX_opc32_MFPR_ESP 30
#define EVX_opc32_MTPR_ESP 31
#define EVX_opc32_MFPR_SSP 32
#define EVX_opc32_MTPR_SSP 33
#define EVX_opc32_MFPR_USP 34
#define EVX_opc32_MTPR_USP 35
#define EVX_opc32_MTPR_TBISD 36
#define EVX_opc32_MTPR_TBISI 37
#define EVX_opc32_MFPR_ASTEN 38
#define EVX_opc32_MFPR_ASTSR 39
#define EVX_opc32_MFPR_WHAMI 63
#define EVX_opc32_GETSS 123
#define EVX_opc32_PUTSS 124
/* Opcodes reserved for PALcode                                             */
#define EVX_OPC32_PAL19 1677721600
#define EVX_OPC32_PAL1B 1811939328
#define EVX_OPC32_PAL1D 1946157056
#define EVX_OPC32_PAL1E 2013265920
#define EVX_OPC32_PAL1F 2080374784
/* Opcodes reserved for DIGITAL                                             */
#define EVX_OPC32_OPC01 67108864
#define EVX_OPC32_OPC02 134217728
#define EVX_OPC32_OPC03 201326592
#define EVX_OPC32_OPC04 268435456
#define EVX_OPC32_OPC05 335544320
#define EVX_OPC32_OPC06 402653184
#define EVX_OPC32_OPC07 469762048
#define EVX_OPC32_OPC0A 671088640
#define EVX_OPC32_OPC0C 805306368
#define EVX_OPC32_OPC0D 872415232
#define EVX_OPC32_OPC0E 939524096
#define EVX_OPC32_OPC14 1342177280
#define EVX_OPC32_OPC1C 1879048192
#define EVX_opc32_NPHALT 251

#define EVX_OPC_HW_MFPR 25
#define EVX_OPC_HW_LD 27
#define EVX_OPC_HW_MTPR 29
#define EVX_OPC_HW_REI 30
#define EVX_OPC_HW_ST 31
#define EVX_opc32_HW_REI 2013298688
#define EVX_opc32_HW_MFPR 1677721600
#define EVX_opc32_HW_MTPR 1946157056
#define EVX_opc32_HW_LD 1811939328
#define EVX_opc32_HW_ST 2080374784
#define EVX_opc32_MT 1946157056
#define EVX_opc32_MF 1677721600
#define EVX_opc32_LDL_P 1811972096
#define EVX_opc32_LDL_A 1811955712
#define EVX_opc32_LDL_W 1811947520
#define EVX_opc32_LDL_AW 1811963904
#define EVX_opc32_LDL_PL 1811988480
#define EVX_opc32_LDQ_P 1811976192
#define EVX_opc32_LDQ_A 1811959808
#define EVX_opc32_LDQ_W 1811951616
#define EVX_opc32_LDQ_AW 1811968000
#define EVX_opc32_LDQ_PL 1811992576
#define EVX_opc32_STL_P 2080407552
#define EVX_opc32_STL_A 2080391168
#define EVX_opc32_STL_PC 2080423936
#define EVX_opc32_STQ_P 2080411648
#define EVX_opc32_STQ_A 2080395264
#define EVX_opc32_STQ_PC 2080428032

/*////////////////////////////////////////////////////////////////////////////////////////
	Constants, enumerations and structure definitions.
////////////////////////////////////////////////////////////////////////////////////////*/
#define true  (1 == 1)
#define false (1 == 2)
#define MAX(a, b) ((a) >= (b) ? (a) : (b))
/*
//	The maximum number of nodes that the scheduler will schedule
//	at one time.
*/
#define MAX_NODES 300

/*
**	Non zero arc values (latencies) are scaled by this amount,
**	since ordinary ref/def constraints that are zero valued
**	get converted to one.
*/
#define SCALE_FACTOR 4

/*
//	Resource is being used or set.
*/
#define USES false
#define SETS true
typedef char boolean;

/*
//	The adjustment that needs to be made to the size and
//	offset fields for the unaligned access memory referencing
//	instructions. e.g. LDQU, STQU
*/
#define UNALIGNED_ACCESS_OFFSET_ADJUSTMENT -7
#define UNALIGNED_ACCESS_SIZE_ADJUSTMENT   7

#define FLAG_ENDBB 1		/*
				** If instruction must end a basic block,
				** because control flow moves to another
				** subroutine, via, BSR, JSR.
                                */

/*
//	The functional units that are available on EV-4 for utilization by
//	instructions,
*/
enum functional_units {
  fu_integer_pipeline,		/*	Integer pipeline		*/
  fu_floating_pipeline,		/*	Floating pipeline		*/
  fu_memory_pipeline,		/*	Load/Store pipeline		*/
  fu_branch_unit,		/*	Branch unit			*/
  fu_imultiplier,		/*	Integer multiplier		*/
  fu_fpdivider,			/*	Floating point divider		*/
  fu_null
  };

/*
//	Classify the resources being used/set as belonging to a
//	particular resource class. This information is needed to
//	index into the producer consumer table to compute the
//	inter instruction delays.
*/
enum resource_class {
  class_load,			/*	Integer load					*/
  class_fload,			/*	Floating point load				*/
  class_store_addr,		/*	Integer store base address			*/
  class_fstore_addr,		/*	Floating point store base address		*/
  class_store_data,		/*	Integer store data				*/
  class_fstore_data,		/*	Floating point store data			*/
  class_ibranch,		/*	Integer branch					*/
  class_fbranch,		/*	Floating point branch				*/
  class_subroutine,		/*	Integer subroutine call				*/
  class_iarithmetic,		/*	Integer arithmetic				*/
  class_imultiplyl,		/*	Integer longword multiply			*/
  class_imultiplyq,		/*	Integer quadword multiply			*/
  class_ilogical,		/*	Logical functions				*/
  class_ishift,			/*	Shift functions					*/
  class_icondmove,		/*	Conditional move				*/
  class_icompare,		/*	Integer compare					*/
  class_fpop,			/*	Other floating point operations			*/
  class_fdivs,			/*	Floating point single precision divide		*/
  class_fdivd,			/*	Floating point double precision divide		*/
  class_null
  };

#define is_in_ibus1 1		/*	Is in instruction bus 1				*/
#define is_in_ibus2 2		/*	Is in instruction bus 2				*/
#define is_in_abox  4		/*	Is in address box				*/
#define is_in_bbox  8		/*	Is in branch box				*/

static int class_flags [] = {
  is_in_ibus1 | is_in_ibus2 | is_in_abox,		/* class_load 			*/
  is_in_ibus1 | is_in_ibus2 | is_in_abox,		/* class_fload			*/
                is_in_ibus2 | is_in_abox,		/* class_store_addr		*/
  is_in_ibus1               | is_in_abox,		/* class_fstore_addr		*/
                is_in_ibus2 | is_in_abox,		/* class_store_data		*/
  is_in_ibus1               | is_in_abox,		/* class_fstore_data		*/
                is_in_ibus2              | is_in_bbox,  /* class_ibranch		*/
  is_in_ibus1                            | is_in_bbox,  /* class_fbranch		*/
                is_in_ibus2              | is_in_bbox,  /* class_subroutine		*/
  is_in_ibus1,						/* class_iarithmetic		*/
  is_in_ibus1,						/* class_imultiplyl		*/
  is_in_ibus1,						/* class_imultiplyq		*/
  is_in_ibus1,						/* class_ilogical		*/
  is_in_ibus1,						/* class_ishift			*/
  is_in_ibus1,						/* class_icondmove		*/
  is_in_ibus1,						/* class_icompare		*/
                is_in_ibus2,				/* class_fpop			*/
                is_in_ibus2,				/* class_fdivs			*/
                is_in_ibus2,				/* class_fdivd			*/
                is_in_ibus2				/* class_null pal,hw_x		*/
  };

/*
//	The set of non register and fake resources set and used
//	by an alpha instruction.
*/
enum other_alpha_resources {
		/*
		//	Arithmetic instructions that trap use this resource
		//	and the TRAPB instruction sets this resource.
		*/
  resource_trap = 64,
  		/*
		//	Memory instructions use the memory barrier resource
		//	and the MB instruction sets this resource.
		*/
  resource_mb,
  		/*
		//	All instructions use this resource, branching type
		//	instructions set this resource. The computation of
		//	this resource is based on the FLAG_ENDS_BASIC_BLOCK
		//	instruction block. other_defuse() does not explicitly
		//	return this resource as def/use.
		*/
  resource_pc,
		/*
		//	All instructions that access memory use this resource.
		*/
  resource_mem,
  		/*
		//	All memory instruction except LDQL and STQC use this
		//	resource and LDQL/STQC set this resource.
		*/
  resource_lock_flag,
  resource_end
  };

/*
//	Structure used to hold the resource use/def information for the instructions.
//	There will be listheads to linked list of these structures, with each list
//	corresponding to a resource.
*/
struct plain_res_s {
  struct plain_res_s	*next;		/* Pointer to next struct in list.		*/
  unsigned		node;		/* Id of instruction affecting this resource.	*/
  enum resource_class   class;		/* Class in which resource is being affected.	*/
  boolean		def;		/* Used or Set					*/
  };
typedef struct plain_res_s plain_resource;

/*
//	Structure used to hold the memory use/def information for the instructions.
//	There will be listheads to linked list of these structures, with each list
//	corresponding to a resource.
*/
struct memory_res_s {
  struct memory_res_s    *next;		/* Pointer to next struct in list.		*/
  unsigned		 node;		/* Id of instruction affecting this resource.	*/
  enum resource_class    class;		/* Class in which resource is being affected.	*/
  boolean		 def;		/* Used or Set					*/
  unsigned		 base;		/* Base register used in reference		*/
  int			 size;		/* Size of operand being referenced.		*/
  int			 offset;	/* Offset from base register			*/
  };
typedef struct memory_res_s memory_resource;

/*///////////////////////////////////////////////////////////////////////////////////////
	Module static variables
////////////////////////////////////////////////////////////////////////////////////////*/

/*
** List head/tail pointers to the resource utilization lists
*/
static  plain_resource		*plain_head [(unsigned) resource_end];
static  plain_resource  	*plain_tail [(unsigned) resource_end];
static  memory_resource 	*mem_head;
static  memory_resource 	*mem_tail;

/*
** Number of the current node being processed.
** Used as an implicit input by addres ().
*/
static	unsigned		current_node_id;

/*
** Global indicating the number of nodes being processed in the
** current basic block.
*/
static  int			number_of_nodes;
static unsigned			presented_instructions;

/*
** Buffer holding the array of instructions.
*/
static  alpha_inst		instructions [MAX_NODES];

/*
** Functional unit usage descriptor
*/
typedef struct {
  enum functional_units 	unit;
  unsigned			cycles;
  } fu_usage_t;

static int 		fu_count  [MAX_NODES];				/* Number of functional units used by instruction 	*/
static unsigned		exec_time [MAX_NODES];
static fu_usage_t 	fu_usage  [MAX_NODES][(int) fu_null];		/* Description of functional units used by instruction */

/*
** Buffer holding the ordering constraints between instruction
** along with the desired latencies.
*/
static  struct {
  boolean	valid;
  unsigned	latency;
  } constraints [MAX_NODES][MAX_NODES];

/*
** Keep track of when the given functional unit is
** available to issue another instruction.
*/
static unsigned fu_available [(int) fu_null];

/*
** Array to mark whether nodes have been issued or not.
** Also used to locate the ending nodes while assigning
** values to nodes.
*/
static boolean	node_processed [MAX_NODES];

/*
** Node values are place in this structure, prior to being
** passed to the qsort library routine to sort nodes in
** descending order of values.
*/
static struct node_value_s {
  unsigned	node;
  unsigned	value;
  } node_value [MAX_NODES];

/*
** Global to indicate the number of instructions that have been issued
** to this point in time.
*/
static int n_issued_nodes;

/*
** Structure containing the reorganized issue order of instructions,
** along with the cycle in which they issue.
*/
static struct node_issue_s {
  unsigned	node;
  unsigned      cycle;
  } node_issue [MAX_NODES];

/*
** Structure to keep track of the cycle in which unreorganized nodes
** were issued.
*/
static struct cycle_issue_s {
  unsigned	cycle;
  } cycle_issue [MAX_NODES];

/*
** Performance enhancement to keep track of whether the ordering
** constraints of a node have been met, and if so the earliest
** cycle in which the node can be issued.
*/
static struct cache_s {
  unsigned	cycle;
  boolean	valid;
  boolean	cons_ok;
  } cache [MAX_NODES];

/*
** Global needed do determine instruction compatiblity with last instruction,
** for dual issue.
*/
static unsigned last_node;

/*
//	This is the Producer/Consumer latency matrix for EV-4.
//	It represents the latencies between when an instruction
//	that produces an operand is issued to a functional unit
//	and the availability of the produced result for use by
//	another functional unit.
*/
static int PCLatencies [(unsigned) class_null+1][class_null+1] = {
   { 3, 3, 3, 3, 3, 3, 3, 0, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 0 },  /* class_iload	*/
   { 3, 3, 3, 3, 3, 3, 0, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 0 },  /* class_fload	*/
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_store_addr	*/
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_fstore_addr */
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_store_data	*/
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_fstore_data */
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_ibranch	*/
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_fbranch	*/
   { 3, 3, 3, 3, 3, 0, 3, 0, 3, 3, 3, 3, 3, 3, 3, 3, 0, 0, 0, 0 },  /* class_subroutine	*/
   { 2, 2, 2, 2, 0, 0, 1, 0, 2, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0 },  /* class_iarithmetic */
   {21,21,21,21,20, 0,21, 0,21,21,21,21,21,21,21,21, 0, 0, 0, 0 },  /* class_imultiplyl	*/
   {23,23,23,23,22, 0,23, 0,23,23,23,23,23,23,23,23, 0, 0, 0, 0 },  /* class_imultiplyq	*/
   { 2, 2, 2, 2, 0, 0, 1, 0, 2, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0 },  /* class_ilogical	*/
   { 2, 2, 2, 2, 0, 0, 2, 0, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0 },  /* class_ishift	*/
   { 2, 2, 2, 2, 0, 0, 2, 0, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0 },  /* class_icondmove	*/
   { 2, 2, 2, 2, 0, 0, 1, 0, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0 },  /* class_icompare	*/
   { 0, 0, 0, 0, 0, 4, 0, 6, 0, 0, 0, 0, 0, 0, 0, 0, 6, 6, 6, 0 },  /* class_fpop	*/
   { 0, 0, 0, 0, 0,32, 0,34, 0, 0, 0, 0, 0, 0, 0, 0,34,34,34, 0 },  /* class_fdivs	*/
   { 0, 0, 0, 0, 0,61, 0,63, 0, 0, 0, 0, 0, 0, 0, 0,63,63,63, 0 },  /* class_fdivd	*/
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }   /* class_null	*/
  };

/*
//	This is the Producer/Producer latency matrix for EV-4.
//	It is an artefact created to model write/write delays,
//	so that the scheduler assigns true costs to the latency
//	of a resource being written twice.
*/
static int PPLatencies [class_null+1][class_null+1] = {
   { 0, 0, 0, 0, 0, 0, 3, 0, 3, 3, 0, 0, 3, 3, 3, 3, 0, 0, 0, 0 },  /* class_iload	*/
   { 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 3, 3, 3, 0 },  /* class_fload	*/
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_store_addr	*/
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_fstore_addr */
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_store_data	*/
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_fstore_data */
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_ibranch	*/
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_fbranch	*/
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_subroutine	*/
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_iarithmetic */
   {19, 0, 0, 0, 0, 0,19, 0,19,19,19,19,19,19,19,19, 0, 0, 0, 0 },  /* class_imultiplyl	*/
   {21, 0, 0, 0, 0, 0,21, 0,21,21,21,21,21,21,21,21, 0, 0, 0, 0 },  /* class_imultiplyq	*/
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_ilogical	*/
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_ishift	*/
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_icondmove	*/
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_icompare	*/
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  /* class_fpop	*/
   { 0,30, 0, 0, 0, 0, 0,30, 0, 0, 0, 0, 0, 0, 0, 0,30,30,30, 0 },  /* class_fdivs	*/
   { 0,59, 0, 0, 0, 0, 0,59, 0, 0, 0, 0, 0, 0, 0, 0,59,59,59, 0 },  /* class_fdivd	*/
   { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }   /* class_null	*/
  };

/*///////////////////////////////////////////////////////////////////////////////////////
	Global Function Prototypes
////////////////////////////////////////////////////////////////////////////////////////*/
void     reorganize_init ();
int      schedule_bb_ev4 (alpha_inst *instp, unsigned number,
			  unsigned instruction_pc, boolean reorder);
void     reorganize_end ();
/*///////////////////////////////////////////////////////////////////////////////////////
	Local Function Prototypes and typedefs
////////////////////////////////////////////////////////////////////////////////////////*/
typedef void (*fu_func)(enum functional_units, unsigned);

static void	reorganize_reset();
static void     issue_node (unsigned node, unsigned cycle);
static boolean  are_functional_units_available (unsigned node, unsigned cycle);
static boolean  are_ordering_constraints_ok (unsigned node, unsigned *cycle);
static unsigned find_node_to_issue (unsigned cycle, boolean reorder);
static enum resource_class node_class (unsigned node);
static boolean  is_dual_issue_ok (unsigned node1, unsigned node2);
static unsigned find_node_to_dual_issue (unsigned cycle, unsigned last_node, boolean reorder);
static void     add_arc (unsigned node1, unsigned node2, unsigned latency);
static void     mark_ending_nodes ();
static int      sort_node_value (void *ent1, void *ent2);
static void     compute_aftercycles ();
static void     addres (boolean def, unsigned resource, enum resource_class class);
static void     resource_constraints (unsigned resource);
static boolean  memory_overlap_possible (memory_resource *ent1, memory_resource *ent2);
static boolean  memory_overlap_certain (memory_resource *ent1, memory_resource *ent2);
static void     updatefu (enum functional_units unit, unsigned cycles);
static void     memory_constraints ();
static unsigned instruction_flags (alpha_inst inst);
static void     resource_defuse (alpha_inst inst);
static void     memory_defuse (alpha_inst inst, memory_resource *mem);
static void     function_unit_use (alpha_inst inst, fu_func funcp);
static enum resource_class instruction_class (alpha_inst inst);
static void	evax_inst_decode(unsigned inst, unsigned va, char *string);
static void	print_ev4_pipeline_diagram (unsigned va);

/*///////////////////////////////////////////////////////////////////////////////////////
	Global Function definitions
////////////////////////////////////////////////////////////////////////////////////////*/
/*
**	Call this routine once per invokation of the program.
*/
void reorganize_init ()
{
}

/*
**	Call this routine once per program invokation to release
**	heap storage allocated by routines in this module.
*/
void reorganize_end ()
{
  reorganize_reset ();
}

#define NO_NODE_FOUND ((unsigned) -1)

/*
**	Reorganize the instructions in the given basic block in order to
**	minimize stalls and maximize issue/dual-issue slots.
**
**	Inputs: instp  - Pointer to the array of alpha instructions
**		number - Number of instructions in array.
**		ipc    - The (pc % 8) or the pc of the first instruction
**			 in the array.
**		reorder - Set true if instructions are to be reordered.
**			  If false, it is useful to see the execution time
**			  of the original (unreordered) instruction stream.
**	Outputs:
*/
int schedule_bb_ev4 (alpha_inst		*instp,
                     unsigned		number,
                     unsigned		ipc,
		     boolean		reorder)
{
  int i, j, node;

  unsigned cycle          = 0;
  unsigned instruction_pc = ipc;

  /*
  ** Clip the number of nodes to the maximum the scheduler can handle.
  */
  assert (number <= MAX_NODES);
  number_of_nodes = number;
  /*
  **	Copy instructions to buffer and work from that
  **	buffer.
  */
  memcpy(instructions, instp, sizeof(instp[0]) * number_of_nodes);

  /*
  **	Release allocated memory from last call to schedule if
  **	there was any.
  */
  reorganize_reset ();

  /*
  **	Work to be done once every basic block.
  **	1. Compute resource utilization
  **	2. Compute ordering constraints/inter node latencies
  **	3. Value the nodes and sort by value (descending order).
  **    4. Mark all nodes as unissued.
  */
  for (i = 0; i < number_of_nodes; i++) {
    current_node_id = i;
    resource_defuse (instp [i]);
    function_unit_use (instp [i], updatefu);
    }
  for (i = 0; i < (unsigned) resource_end; i++)
    resource_constraints (i);
  memory_constraints ();
  compute_aftercycles ();

  n_issued_nodes = 0;
  for (i = 0; i < number_of_nodes; i++)
    node_processed [i] = false;

  /*
  **	Issue nodes till there are no more to issue.
  */
  if (number_of_nodes) {
    node = find_node_to_issue (cycle, reorder);
    issue_node (node, cycle);
    instruction_pc += sizeof (alpha_inst);
    }

  while (n_issued_nodes < number_of_nodes) {
    if (instruction_pc % 8) {
      /*
      **	Second longword of aligned quadword, find an instruction
      **	to dual issue in the current cycle. If we succeed issue it
      **	else look for a single issue.
      */
      if ((node = find_node_to_dual_issue (cycle, last_node, reorder)) != NO_NODE_FOUND) {
	issue_node (node, cycle);
	instruction_pc += sizeof (alpha_inst);
	continue;
	}
      }
    /*
    **	At the first longword of an aligned quadword, or could
    **	not find a node to dual issue, look for a node to issue.
    */
    while ((node = find_node_to_issue (++cycle, reorder)) == NO_NODE_FOUND);
    issue_node (node, cycle);
    instruction_pc += sizeof (alpha_inst);
    }

  print_ev4_pipeline_diagram (ipc);
  return number_of_nodes;
}

/*///////////////////////////////////////////////////////////////////////////////////////
	Local Function definitions
////////////////////////////////////////////////////////////////////////////////////////*/

/*
**	Return a node that can be issued in the given
**	cycle that meets ordering constraints and will not
**	stall on issuing.
*/
static unsigned find_node_to_issue (unsigned cycle, boolean reorder)
{
  int		i;
  unsigned	earliest_cycle;
  unsigned	node;

  if (reorder) {
    for (i = 0; i < number_of_nodes; i++) {
      node = node_value [i].node;

      if (!node_processed [node]					&&
	  are_ordering_constraints_ok (node, &earliest_cycle)		&&
	  earliest_cycle <= cycle					&&
	  are_functional_units_available (node, cycle))
        return node;
      }
    }
  else {
    node = n_issued_nodes;
    if (!node_processed [node]						&&
        are_ordering_constraints_ok (node, &earliest_cycle)		&&
	earliest_cycle <= cycle						&&
	are_functional_units_available (node, cycle))
      return node;
    }
  return NO_NODE_FOUND;
}

/*
**	Return a node that can be dual issued with the last issued
**	instruction, in the given cycle that meets ordering
**	constraints and will not stall on issuing.
*/
static unsigned find_node_to_dual_issue (unsigned cycle, unsigned last_node, boolean reorder)
{
  int		i;
  unsigned	earliest_cycle;
  unsigned	node;

  if (reorder) {
    for (i = 0; i < number_of_nodes; i++) {
      node = node_value [i].node;

      if (!node_processed [node]					&&
	  are_ordering_constraints_ok (node, &earliest_cycle)		&&
	  earliest_cycle <= cycle					&&
          is_dual_issue_ok (node, last_node)				&&
	  are_functional_units_available (node, cycle))
        return node;
      }
    }
  else {
    node = n_issued_nodes;
    if (!node_processed [node]						&&
	are_ordering_constraints_ok (node, &earliest_cycle)		&&
	earliest_cycle <= cycle						&&
        is_dual_issue_ok (node, last_node)				&&
	are_functional_units_available (node, cycle))
      return node;
    }
  return NO_NODE_FOUND;
}

/*
**	Return true if the functional units used by 'node' are available
**	in the current cycle.
*/
static boolean are_functional_units_available (unsigned node, unsigned cycle)
{
  int i, j;

  if ((j = fu_count [node]) != 0)
    for (i = 0; i < j; i++)
      if (fu_available [(int) fu_usage [node][i].unit] > cycle)
        return false;
  return true;
}

/*
**	Return true iff the instruction ordering constraints for "node" have
**	been met. If true is returned, cycle also contains the earliest cycle
**	in which the node can be issued without stalling.
*/
static boolean are_ordering_constraints_ok (unsigned node, unsigned *cycle)
{
  int i;
  unsigned earliest_cycle;

  /*
  **	If this information was already computed and cached,
  **	return it.
  */
  if (cache [node].valid) {
    *cycle = cache [node].cycle;
    return cache [node].cons_ok;
    }

  /*
  **	1. Check for ordering constraints being met, and also compute
  **	   the earliest cycle for issue as long as the constraints are
  **	   being met.
  **	2. Cache the results.
  */
  earliest_cycle = 0;
  cache [node].valid = true;
  for (i = 0; i < number_of_nodes; i++) {
    if (constraints [i][node].valid) {
      if (!node_processed [i]) {
	cache [node].cons_ok = false;
	return false;
	}
      else
	earliest_cycle = MAX (earliest_cycle, constraints [i][node].latency + cycle_issue [i].cycle);
      }
    }
  cache [node].cycle = earliest_cycle;
  cache [node].cons_ok = true;
  *cycle = earliest_cycle;
  return true;
}

/*
**	Return TRUE if the instructions represented by their resource
**	classes are acceptable for dual issue.
**	It is needed to figure out the dual issue rules for EV-4.
*/
static boolean  is_dual_issue_ok (unsigned node1, unsigned node2)
{
  int flag1 = class_flags [(int) node_class (node1)];
  int flag2 = class_flags [(int) node_class (node2)];

  return ((((flag1 & is_in_ibus1) && (flag2 & is_in_ibus2))		||
           ((flag1 & is_in_ibus2) && (flag2 & is_in_ibus1)))		&&
	  !((flag1 & is_in_abox)  && (flag2 & is_in_abox))		&&
	  !((flag1 & is_in_bbox)  && (flag2 & is_in_bbox)));
}

/*
**	Called when node "node" has been picked to issue
**	in cycle "cycle". Updates some tracking structures
**	needed during code scheduling.
*/
static void issue_node (unsigned node, unsigned cycle)
{
  int				i, j;
  /*
  ** Invalidate cache, since some ordering/timing constraints
  ** may now have changed.
  */
  for (i = 0; i < number_of_nodes; i++)
    cache [i].valid = false;
  /*
  ** 1. Mark node as having been issued and the cycle in which
  **    it was issued.
  ** 2. Update the last node global.
  ** 3. Update the function unit availability information.
  */
  node_processed [node]             = true;
  node_issue [n_issued_nodes].node  = node;
  node_issue [n_issued_nodes].cycle = cycle;
  cycle_issue [node].cycle          = cycle;
  last_node                         = node;
  n_issued_nodes                    = n_issued_nodes + 1;
  if ((j = fu_count [node]) != 0)
    for (i = 0; i < j; i++)
      fu_available [(int) fu_usage [node][i].unit] = cycle + fu_usage [node][i].cycles;
}

/*
**	The resource class of the instruction that is
**	needed to figure out the dual issue rules for EV-4.
*/
static enum resource_class node_class (unsigned node)
{
  return instruction_class (instructions [node]);
}

/*
//	Add an ordering constraint between node1 and node2 with a timing constraint
//	of latency.
*/
static void add_arc (unsigned node1, unsigned node2, unsigned latency)
{
  if (node1 == node2)
    return;

  if (constraints [node1][node2].valid)
    constraints [node1][node2].latency = MAX (constraints [node1][node2].latency, latency);
  else {
    constraints [node1][node2].latency = latency;
    constraints [node1][node2].valid = true;
    }
}

/*
**	Mark all nodes that have no successors, by setting the
**	corresponding boolean in the node_processed array.
*/
static void mark_ending_nodes ()
{
  int i, j;
  for (i = 0; i < number_of_nodes; i++) {
    node_processed [i] = true;
    for (j = 0; j < number_of_nodes; j++) {
      if (constraints [i][j].valid) {
	node_processed [i] = false;
	break;
	}
      }
    }
}

/*
**	Function used as callback routine by the library routine qsort
**	to create a sort the array of node_value structures.
*/
static int sort_node_value (void *ent1, void *ent2)
{
  struct node_value_s *sort1, *sort2;

  sort1 = (struct node_value_s *) ent1;
  sort2 = (struct node_value_s *) ent2;
  return (sort2->value - sort1->value);
}

/*
**	Assign values to the nodes, computed as the distance from
**	the end (in cycles).
*/
static void compute_aftercycles ()
{
  int i, j, nodes_left;
  unsigned value, latency;

  /*
  ** 1. Number all the nodes in the node_value structure.
  ** 2. Assign a value of 0 to all nodes that have no successors.
  */
  mark_ending_nodes ();
  for (i = 0, nodes_left = number_of_nodes; i < number_of_nodes; i++) {
    node_value [i].node = i;
    if (node_processed [i]) {
      node_value [i].value = 0;
      --nodes_left;
      }
    }

  /*
  **	A node is can be assigned a value if all it's successors have
  **	values assigned to them. A node's value is the maximum over all
  **	the sucessors, of the sum of the arc value (latency) and the
  **	successor node value.
  **	The nodes are traversed in reverse order to minimize iterations.
  */
  if (nodes_left) {
    for (i = number_of_nodes-1; i >= 0; i--) {
      if (!node_processed [i]) {
        value = 0;
	for (j = i; j < number_of_nodes; j++) {
	  if (constraints [i][j].valid && i != j) {
	    latency = constraints [i][j].latency;
	    latency = latency ? SCALE_FACTOR * latency : 1;
	    value   = MAX (value, latency + node_value [j].value);
	    }
	  }
	node_value [i].value = value;
	}
      }
    }

  /*
  **	Sort the node values in descending order, i.e. longest distance from the end.
  */
  qsort ((void *) node_value, (size_t) number_of_nodes, sizeof (node_value [0]), sort_node_value);
}

/*
//	Add the resource as being ref/def'ed by instruction current_node_id.
//	For purposes of classification the resource class is included to determine
//	latencies from the producer/consumer table.
*/
static void addres (boolean def, unsigned resource, enum resource_class class)
{
  /*
  // Ignore the RAZ/IGN resources
  */
  if (resource == 31 || resource == 63)
    return;
  else if (resource == (unsigned) resource_mem) {
    memory_resource *memp;

    memp        = (memory_resource *) malloc (sizeof(memory_resource));
    memp->next  = 0;			/* Fill in structure information */
    memp->class = class;
    memp->node  = current_node_id;
    memory_defuse (instructions [current_node_id], memp);

    if (mem_tail) {			/* Add to tail of linked list */
      mem_tail->next = memp;
      mem_tail = memp;
      }
    else
      mem_head = mem_tail = memp;
    }
  else {
    plain_resource *plp;

    plp        = (plain_resource *) malloc (sizeof(plain_resource));
    plp->next  = 0;			/* Fill in structure information */
    plp->def   = def;
    plp->class = class;
    plp->node  = current_node_id;

    if (plain_tail [(unsigned) resource]) {	/* Add to tail of linked list */
      plain_tail [(unsigned) resource]->next = plp;
      plain_tail [(unsigned) resource] = plp;
      }
    else 
      plain_head [(unsigned) resource] = plain_tail [(unsigned) resource] = plp;
    }
}

/*
//	Walk through the resource utilization pattern for the specified
//	resource and determine all the ref/def, def/def and def/ref based
//	ordering constraints for the resource.
*/
static void resource_constraints (unsigned resource)
{
  plain_resource *entb, *ente, *next;
  
  entb = ente = plain_head [resource];

  if (!entb)
    return;

  while (ente) {
    if (entb->def == SETS) {			/* Here ente == entb, Found a def, go down list till a ref or end */
      ente = entb->next;
      while (ente && ente->def == SETS) {	/* def def constraints */
        add_arc (entb->node, ente->node, PPLatencies [entb->class][ente->class]);
	entb = ente;
	ente = ente->next;
	}
      if (ente) {				/* ente points to a use, and entb points to a set */
						/* Go down list till last use or end of list */

	do {					/* def use constraint */
	  add_arc (entb->node, ente->node, PCLatencies [entb->class][ente->class]);
	  ente = ente->next;
	  } while (ente && ente->def == USES);
	ente = entb->next;			/* Synchronize ente and entb to first ref */
	entb = ente;
	}
      }
    else {					/* Here ente == entb, Found a ref, go down list till a def or end */
      while (ente && ente->def == USES)
	ente = ente->next;
      if (ente) {				/* Found a def, add constraints between all refs and this def */
	do {					/* use def constraint */
	  add_arc (entb->node, ente->node, 0);
	  entb = entb->next;
	  } while (entb != ente);
        }
      }
    }
}

/*
**	Return TRUE if the memory regions represented by ent1 and ent2
//	could possibly overlap.
*/
static boolean memory_overlap_possible (memory_resource *ent1, memory_resource *ent2)
{
  if (ent1->base != ent2->base)
    return true;
  else
    return (ent1->offset <= ent2->offset				?
	    (ent1->offset + ent1->size - 1 >= ent2->offset)		:
	    (ent2->offset + ent2->size - 1 >= ent1->offset));
}

/*
**	Return TRUE if memory overlap is certain. EV4 has this stupid rule
**	that a store to a location followed by a load from the same location
**	is better off with 6 bubbles in between, else the hit is more than
**	6 cycles.
*/
static boolean memory_overlap_certain (memory_resource *ent1, memory_resource *ent2)
{
  return ((ent1->base == ent2->base)					&&
          (ent1->offset <= ent2->offset				?
	    (ent1->offset + ent1->size - 1 >= ent2->offset)	:
	    (ent2->offset + ent2->size - 1 >= ent1->offset)));

}

/*
**	Update the functional unit availability information to
**	accurately model the machine execution units. Function
**	unit usage is critical only for integer multiplies and
**	floating point divides on EV-4. This will change for other
**	implementations.
*/
static void updatefu (enum functional_units unit, unsigned cycles)
{
  int current_index = fu_count [current_node_id];
  exec_time [current_node_id] = MAX (exec_time [current_node_id], cycles);
  fu_usage  [current_node_id][current_index].unit   = unit;
  fu_usage  [current_node_id][current_index].cycles = cycles;
  fu_count  [current_node_id]++;
}

/*
**	Walk through the resource utilization pattern for the memory resource and
**	determine all the ref/def, def/def and def/ref based ordering constraints.
*/
static void memory_constraints ()
{
  memory_resource *entb, *ente;

  entb = mem_head;

  while (entb) {
    ente = entb->next;
    while (ente) {
      if (ente->def == SETS) {
        /* USES/SETS and SETS/SETS */
        if (memory_overlap_possible (entb, ente))
	  add_arc (entb->node, ente->node, 0);
        }
      else if (entb->def == SETS && ente->def == USES) {
        /* SETS/USES */
	if (memory_overlap_certain (entb, ente))
	  add_arc (entb->node, ente->node, 6);
        else if (memory_overlap_possible (entb, ente))
	  add_arc (entb->node, ente->node, 0);
        }
      else {
        }
      ente = ente->next;
      }
    entb = entb->next;
    }
}

static unsigned instruction_flags (alpha_inst inst)
{
  unsigned opcode, flag = 0;

  opcode = inst.o_format.opcode;
  switch (opcode) {
    case EVX_OPC_JSR:
    case EVX_OPC_BSR:
    case EVX_OPC_BR:
      flag |= FLAG_ENDBB;
      break;
    default:
      break;
    }
  return flag;
  }

/*
**	Determine the set of resources that are set and used by the alpha instruction.
*/
static void resource_defuse (alpha_inst inst)
{
  register unsigned ra, rb, rc, opcode, fa, fb, fc, func;

  ra = inst.o_format.ra;
  rb = inst.o_format.rb;
  rc = inst.o_format.rc;

  fa = ra + 32;
  fb = rb + 32;
  fc = rc + 32;

  opcode = inst.o_format.opcode;
  addres (USES, (unsigned) resource_pc, class_null);

  switch (opcode) {
  case EVX_OPC_LDL:
  case EVX_OPC_LDQ:
  case EVX_OPC_LDQU:
    /* Integer load instructions ==> opcode ra, disp(rb) */
    addres (USES, rb, class_load);
    addres (USES, (unsigned) resource_mem, class_null);
    addres (USES, (unsigned) resource_mb, class_null);
    addres (USES, (unsigned) resource_lock_flag, class_null);
    addres (SETS, ra, class_load);
    break;
  case EVX_OPC_LDLL:
  case EVX_OPC_LDQL:
    /* Integer load instructions ==> opcode ra, disp(rb) */
    addres (USES, rb, class_load);
    addres (USES, (unsigned) resource_mem, class_null);
    addres (USES, (unsigned) resource_mb, class_null);
    addres (SETS, (unsigned) resource_lock_flag, class_null);
    addres (SETS, ra, class_load);
    break;
  case EVX_OPC_STL:
  case EVX_OPC_STQ:
  case EVX_OPC_STQU:
    /* Integer store  instructions ==> opcode ra, disp(rb) */
    addres (USES, ra, class_store_data);
    addres (USES, rb, class_store_addr);
    addres (USES, (unsigned) resource_mem, class_null);
    addres (USES, (unsigned) resource_mb, class_null);
    addres (USES, (unsigned) resource_lock_flag, class_null);
    break;
  case EVX_OPC_STLC:
  case EVX_OPC_STQC:
    /* Integer store  instructions ==> opcode ra, disp(rb) */
    addres (USES, ra, class_store_data);
    addres (USES, rb, class_store_addr);
    addres (USES, (unsigned) resource_mem, class_null);
    addres (USES, (unsigned) resource_mb, class_null);
    addres (SETS, (unsigned) resource_lock_flag, class_null);
    break;
  case EVX_OPC_LDAH:
  case EVX_OPC_LDA:
    /* Address instructions ==> opcode ra, disp(rb) */
    addres (USES, rb, class_iarithmetic);
    addres (SETS, ra, class_iarithmetic);
    break;
  case EVX_OPC_LDF:
  case EVX_OPC_LDD:
  case EVX_OPC_LDS:
  case EVX_OPC_LDT:
    /* Floating load instructions ==> opcode fa, disp(rb) */
    addres (USES, rb, class_fload);
    addres (USES, (unsigned) resource_mem, class_null);
    addres (USES, (unsigned) resource_mb, class_null);
    addres (USES, (unsigned) resource_lock_flag, class_null);
    addres (SETS, fa, class_fload);
    break;
  case EVX_OPC_STF:
  case EVX_OPC_STD:
  case EVX_OPC_STS:
  case EVX_OPC_STT:
    /* Floating store instructions ==> opcode fa, disp(rb) */
    addres (USES, fa, class_fstore_data);
    addres (USES, rb, class_fstore_addr);
    addres (USES, (unsigned) resource_mem, class_null);
    addres (USES, (unsigned) resource_mb, class_null);
    addres (USES, (unsigned) resource_lock_flag, class_null);
    break;
  case EVX_OPC_SYNC: {
    func = inst.word & 32767u;
    switch(func) {
      case EVX_SYNC_MB:
        addres (SETS, (unsigned) resource_mb, class_null);
	break;
      case EVX_SYNC_RS:
      case EVX_SYNC_RC:
        addres (SETS, (unsigned) resource_pc, class_null);
        break;
      case EVX_SYNC_TRAPB:
        addres (SETS, (unsigned) resource_pc,   class_null);
	addres (SETS, (unsigned) resource_trap, class_null);
        break;
      case EVX_SYNC_FETCH_M:
        addres (USES, rb, class_null);
        addres (USES, (unsigned) resource_mb, class_null);
	break;
      case EVX_SYNC_RCC:
	addres (USES, ra, class_null);
        addres (SETS, (unsigned) resource_pc, class_null);
        break;
      default:
        addres (SETS, (unsigned) resource_pc, class_null);
        break;
      }
    }
    break;
  case EVX_OPC_JSR:
    /* Memory format jsr instructions ==> opcode ra, hints(rb) */
    addres (USES, rb, class_subroutine);
    addres (SETS, ra, class_subroutine);
    addres (SETS, (unsigned) resource_pc, class_null);
    break;
  case EVX_OPC_BR:
  case EVX_OPC_BSR:
    /* Branch format instructions ==> opcode ra, branch_displacement */
    addres (SETS, ra, class_subroutine);
    addres (SETS, (unsigned) resource_pc, class_null);
    break;
  case EVX_OPC_BLBC:
  case EVX_OPC_BLBS:
  case EVX_OPC_BEQ:
  case EVX_OPC_BNE:
  case EVX_OPC_BLT:
  case EVX_OPC_BLE:
  case EVX_OPC_BGT:
  case EVX_OPC_BGE:
    /* Branch format instructions ==> opcode ra, branch_displacement */
    addres (USES, ra, class_ibranch);
    addres (SETS, (unsigned) resource_pc, class_null);
    break;
  case EVX_OPC_FBEQ:
  case EVX_OPC_FBNE:
  case EVX_OPC_FBLT:
  case EVX_OPC_FBLE:
  case EVX_OPC_FBGT:
  case EVX_OPC_FBGE:
    /* Branch format instructions ==> opcode ra, branch_displacement */
    addres (USES, fa, class_fbranch);
    addres (SETS, (unsigned) resource_pc, class_null);
    break;
  case EVX_OPC_ADDI: {
    func = inst.o_format.function;
    switch(func) {
      case EVX_ADDI_ADDLV:
      case EVX_ADDI_ADDQV:
      case EVX_ADDI_SUBLV:
      case EVX_ADDI_SUBQV:
        addres (USES, (unsigned) resource_trap, class_null);
      case EVX_ADDI_ADDL:
      case EVX_ADDI_ADDQ:
      case EVX_ADDI_S4ADDL:
      case EVX_ADDI_S4ADDQ:
      case EVX_ADDI_S8ADDL:
      case EVX_ADDI_S8ADDQ:
      case EVX_ADDI_S4SUBL:
      case EVX_ADDI_S4SUBQ:
      case EVX_ADDI_S8SUBL:
      case EVX_ADDI_S8SUBQ:
      case EVX_ADDI_SUBL:
      case EVX_ADDI_SUBQ:
        if (inst.o_format.form) {
	  /* Integer/logical instruction in literal format */
          addres (USES, ra, class_iarithmetic);
          addres (SETS, rc, class_iarithmetic);
	  }
        else {
	  /* Integer/logical instruction not in literal format */
          addres (USES, ra, class_iarithmetic);
          addres (USES, rb, class_iarithmetic);
          addres (SETS, rc, class_iarithmetic);
          }
	break;
      case EVX_ADDI_CMPEQ:
      case EVX_ADDI_CMPLT:
      case EVX_ADDI_CMPLE:
      case EVX_ADDI_CMPULT:
      case EVX_ADDI_CMPULE:
      case EVX_ADDI_CMPBGE:
	if (inst.o_format.form) {
	  /* Integer/logical instruction in literal format */
          addres (USES, ra, class_icompare);
	  addres (SETS, rc, class_icompare);
          }
        else {
	  /* Integer/logical instruction not in literal format */
          addres (USES, ra, class_icompare);
	  addres (USES, rb, class_icompare);
          addres (SETS, rc, class_icompare);
          }
	break;
      default:
        addres (SETS, (unsigned) resource_pc, class_null);
        break;
      }
    }
    break;
  case EVX_OPC_LOGI: {
    func = inst.o_format.function;
    switch(func) {
      case EVX_LOGI_AND:
      case EVX_LOGI_BIC:
      case EVX_LOGI_BIS:
      case EVX_LOGI_ORNOT:
      case EVX_LOGI_XOR:
      case EVX_LOGI_EQV:
	if (inst.o_format.form) {
          /* Integer/logical instruction in literal format */
          addres (USES, ra, class_ilogical);
          addres (SETS, rc, class_ilogical);
          }
        else {
          /* Integer/logical instruction not in literal format */
          addres (USES, ra, class_ilogical);
          addres (USES, rb, class_ilogical);
          addres (SETS, rc, class_ilogical);
          }
        break;
      case EVX_LOGI_CMOVEQ:
      case EVX_LOGI_CMOVNE:
      case EVX_LOGI_CMOVLBS:
      case EVX_LOGI_CMOVLT:
      case EVX_LOGI_CMOVGE:
      case EVX_LOGI_CMOVLBC:
      case EVX_LOGI_CMOVLE:
      case EVX_LOGI_CMOVGT:
        if (inst.o_format.form) {
          /* Integer/logical instruction in literal format */
          addres (USES, ra, class_icondmove);
          addres (SETS, rc, class_icondmove);
          }
        else {
          /* Integer/logical instruction not in literal format */
          addres (USES, ra, class_icondmove);
          addres (USES, rb, class_icondmove);
          addres (SETS, rc, class_icondmove);
          }
        break;
      default:
	addres (SETS, (unsigned) resource_pc, class_null);
	break;
      }
    }
    break;
  case EVX_OPC_SHFI: {
    func = inst.o_format.function;
    switch(func) {
      case EVX_SHFI_SLL:
      case EVX_SHFI_SRA:
      case EVX_SHFI_SRL:
      case EVX_SHFI_EXTBL:
      case EVX_SHFI_EXTWL:
      case EVX_SHFI_EXTLL:
      case EVX_SHFI_EXTQL:
      case EVX_SHFI_EXTWH:
      case EVX_SHFI_EXTLH:
      case EVX_SHFI_EXTQH:
      case EVX_SHFI_INSBL:
      case EVX_SHFI_INSWL:
      case EVX_SHFI_INSLL:
      case EVX_SHFI_INSQL:
      case EVX_SHFI_INSWH:
      case EVX_SHFI_INSLH:
      case EVX_SHFI_INSQH:
      case EVX_SHFI_MSKBL:
      case EVX_SHFI_MSKWL:
      case EVX_SHFI_MSKLL:
      case EVX_SHFI_MSKQL:
      case EVX_SHFI_MSKWH:
      case EVX_SHFI_MSKLH:
      case EVX_SHFI_MSKQH:
      case EVX_SHFI_ZAP:
      case EVX_SHFI_ZAPNOT:
        if (inst.o_format.form) {
          /* Integer/logical instruction in literal format */
          addres (USES, ra, class_ishift);
          addres (SETS, rc, class_ishift);
	  }
	else {
          /* Integer/logical instruction not in literal format */
	  addres (USES, ra, class_ishift);
          addres (USES, rb, class_ishift);
          addres (SETS, rc, class_ishift);
	  }
        break;
      default:
        addres (SETS, (unsigned) resource_pc, class_null);
        break;
      }
    }
    break;
  case EVX_OPC_MULI: {
    func = inst.o_format.function;
    switch(func) {
      case EVX_MULI_MULQV:
        addres (USES, (unsigned) resource_trap, class_null);
      case EVX_MULI_UMULH:
      case EVX_MULI_MULQ:
        if (inst.o_format.form) {
	  /* Integer/logical instruction in literal format */
          addres (USES, ra, class_imultiplyq);
          addres (SETS, rc, class_imultiplyq);
	  }
        else {
          /* Integer/logical instruction not in literal format */
          addres (USES, ra, class_imultiplyq);
          addres (USES, rb, class_imultiplyq);
          addres (SETS, rc, class_imultiplyq);
          }
        break;
      case EVX_MULI_MULL:
      case EVX_MULI_MULLV:
        if (inst.o_format.form) {
          /* Integer/logical instruction in literal format */
          addres (USES, ra, class_imultiplyl);
          addres (SETS, rc, class_imultiplyl);
	  }
	else {
          /* Integer/logical instruction not in literal format */
	  addres (USES, ra, class_imultiplyl);
          addres (USES, rb, class_imultiplyl);
          addres (SETS, rc, class_imultiplyl);
	  }
        break;
      default:
        addres (SETS, (unsigned) resource_pc, class_null);
        break;
      }
    }
    break;
  case EVX_OPC_FPAUX: {
      func = inst.f_format.function;
    switch(func) {
      case EVX_FPAUX_CPYS:
      case EVX_FPAUX_CPYSN:
      case EVX_FPAUX_CPYSE:
	addres (USES, fa, class_fpop);
        addres (USES, fb, class_fpop);
	addres (SETS, fc, class_fpop);
        break;
      case EVX_FPAUX_CVTQLV:
      case EVX_FPAUX_CVTQLSV:
      case EVX_FPAUX_CVTLQ:
      case EVX_FPAUX_CVTQL:
        addres (USES, (unsigned) resource_trap, class_null);
        addres (USES, fb, class_fpop);
        addres (SETS, fc, class_fpop);
        break;
      case EVX_FPAUX_FCMOVEQ:
      case EVX_FPAUX_FCMOVNE:
      case EVX_FPAUX_FCMOVLE:
      case EVX_FPAUX_FCMOVLT:
      case EVX_FPAUX_FCMOVGE:
      case EVX_FPAUX_FCMOVGT:
        addres (USES, fa, class_fpop);
	addres (USES, fb, class_fpop);
	addres (SETS, fc, class_fpop);
        break;
      case EVX_FPAUX_WR_FPCTL:
        addres (USES, fa, class_fpop);
        addres (SETS, (unsigned) resource_pc, class_null);
	break;
      case EVX_FPAUX_RD_FPCTL:
        addres (SETS, fa, class_fpop);
        addres (SETS, (unsigned) resource_pc, class_null);
        break;
      default:
        addres (SETS, (unsigned) resource_pc, class_null);
        break;
      }
    }
    break;
  case EVX_OPC_FPDEC: {
    func = inst.f_format.function;
    switch(func) {
      case EVX_FPDEC_ADDDU:
      case EVX_FPDEC_ADDDUC:
      case EVX_FPDEC_ADDDSU:
      case EVX_FPDEC_ADDDSUC:
      case EVX_FPDEC_ADDFU:
      case EVX_FPDEC_ADDFUC:
      case EVX_FPDEC_ADDFSU:
      case EVX_FPDEC_ADDFSUC:
      case EVX_FPDEC_ADDGU:
      case EVX_FPDEC_ADDGUC:
      case EVX_FPDEC_ADDGSU:
      case EVX_FPDEC_ADDGSUC:
      case EVX_FPDEC_SUBDU:
      case EVX_FPDEC_SUBDUC:
      case EVX_FPDEC_SUBDSU:
      case EVX_FPDEC_SUBDSUC:
      case EVX_FPDEC_SUBFU:
      case EVX_FPDEC_SUBFUC:
      case EVX_FPDEC_SUBFSU:
      case EVX_FPDEC_SUBFSUC:
      case EVX_FPDEC_SUBGU:
      case EVX_FPDEC_SUBGUC:
      case EVX_FPDEC_SUBGSU:
      case EVX_FPDEC_SUBGSUC:
      case EVX_FPDEC_MULDU:
      case EVX_FPDEC_MULDUC:
      case EVX_FPDEC_MULDSU:
      case EVX_FPDEC_MULDSUC:
      case EVX_FPDEC_MULFU:
      case EVX_FPDEC_MULFUC:
      case EVX_FPDEC_MULFSU:
      case EVX_FPDEC_MULFSUC:
      case EVX_FPDEC_MULGU:
      case EVX_FPDEC_MULGUC:
      case EVX_FPDEC_MULGSU:
      case EVX_FPDEC_MULGSUC:
      case EVX_FPDEC_ADDD:
      case EVX_FPDEC_ADDDC:
      case EVX_FPDEC_ADDDS:
      case EVX_FPDEC_ADDDSC:
      case EVX_FPDEC_ADDF:
      case EVX_FPDEC_ADDFC:
      case EVX_FPDEC_ADDFS:
      case EVX_FPDEC_ADDFSC:
      case EVX_FPDEC_ADDG:
      case EVX_FPDEC_ADDGC:
      case EVX_FPDEC_ADDGS:
      case EVX_FPDEC_ADDGSC:
      case EVX_FPDEC_SUBD:
      case EVX_FPDEC_SUBDC:
      case EVX_FPDEC_SUBDS:
      case EVX_FPDEC_SUBDSC:
      case EVX_FPDEC_SUBF:
      case EVX_FPDEC_SUBFC:
      case EVX_FPDEC_SUBFS:
      case EVX_FPDEC_SUBFSC:
      case EVX_FPDEC_SUBG:
      case EVX_FPDEC_SUBGC:
      case EVX_FPDEC_SUBGS:
      case EVX_FPDEC_SUBGSC:
      case EVX_FPDEC_MULD:
      case EVX_FPDEC_MULDC:
      case EVX_FPDEC_MULDS:
      case EVX_FPDEC_MULDSC:
      case EVX_FPDEC_MULF:
      case EVX_FPDEC_MULFC:
      case EVX_FPDEC_MULFS:
      case EVX_FPDEC_MULFSC:
      case EVX_FPDEC_MULG:
      case EVX_FPDEC_MULGC:
      case EVX_FPDEC_MULGS:
      case EVX_FPDEC_MULGSC:
      case EVX_FPDEC_CMPDEQ:
      case EVX_FPDEC_CMPDLT:
      case EVX_FPDEC_CMPDLE:
      case EVX_FPDEC_CMPGEQ:
      case EVX_FPDEC_CMPGLT:
      case EVX_FPDEC_CMPGLE:
      case EVX_FPDEC_CMPDEQS:
      case EVX_FPDEC_CMPDLTS:
      case EVX_FPDEC_CMPDLES:
      case EVX_FPDEC_CMPGEQS:
      case EVX_FPDEC_CMPGLTS:
      case EVX_FPDEC_CMPGLES:
        addres (USES, fa, class_fpop);
        addres (USES, fb, class_fpop);
	addres (SETS, fc, class_fpop);
        addres (USES, (unsigned) resource_trap, class_null);
	break;
      case EVX_FPDEC_DIVF:
      case EVX_FPDEC_DIVFC:
      case EVX_FPDEC_DIVFS:
      case EVX_FPDEC_DIVFSC:
      case EVX_FPDEC_DIVFU:
      case EVX_FPDEC_DIVFUC:
      case EVX_FPDEC_DIVFSU:
      case EVX_FPDEC_DIVFSUC:
        addres (USES, fa, class_fdivs);
        addres (USES, fb, class_fdivs);
        addres (SETS, fc, class_fdivs);
	addres (USES, (unsigned) resource_trap, class_null);
	break;
      case EVX_FPDEC_DIVD:
      case EVX_FPDEC_DIVDC:
      case EVX_FPDEC_DIVDS:
      case EVX_FPDEC_DIVDSC:
      case EVX_FPDEC_DIVDU:
      case EVX_FPDEC_DIVDUC:
      case EVX_FPDEC_DIVDSU:
      case EVX_FPDEC_DIVDSUC:
      case EVX_FPDEC_DIVG:
      case EVX_FPDEC_DIVGC:
      case EVX_FPDEC_DIVGS:
      case EVX_FPDEC_DIVGSC:
      case EVX_FPDEC_DIVGU:
      case EVX_FPDEC_DIVGUC:
      case EVX_FPDEC_DIVGSU:
      case EVX_FPDEC_DIVGSUC:
        addres (USES, fa, class_fdivd);
        addres (USES, fb, class_fdivd);
	addres (SETS, fc, class_fdivd);
	addres (USES, (unsigned) resource_trap, class_null);
	break;
      case EVX_FPDEC_CVTDGU:
      case EVX_FPDEC_CVTDGUC:
      case EVX_FPDEC_CVTDGSU:
      case EVX_FPDEC_CVTDGSUC:
      case EVX_FPDEC_CVTGDU:
      case EVX_FPDEC_CVTGDUC:
      case EVX_FPDEC_CVTGDSU:
      case EVX_FPDEC_CVTGDSUC:
      case EVX_FPDEC_CVTGFU:
      case EVX_FPDEC_CVTGFUC:
      case EVX_FPDEC_CVTGFSU:
      case EVX_FPDEC_CVTGFSUC:
      case EVX_FPDEC_CVTDQV:
      case EVX_FPDEC_CVTDQVC:
      case EVX_FPDEC_CVTDQSV:
      case EVX_FPDEC_CVTDQSVC:
      case EVX_FPDEC_CVTGQV:
      case EVX_FPDEC_CVTGQVC:
      case EVX_FPDEC_CVTGQSV:
      case EVX_FPDEC_CVTGQSVC:
      case EVX_FPDEC_CVTDF:
      case EVX_FPDEC_CVTDFC:
      case EVX_FPDEC_CVTDFS:
      case EVX_FPDEC_CVTDFSC:
      case EVX_FPDEC_CVTDG:
      case EVX_FPDEC_CVTDGC:
      case EVX_FPDEC_CVTDGS:
      case EVX_FPDEC_CVTDGSC:
      case EVX_FPDEC_CVTGD:
      case EVX_FPDEC_CVTGDC:
      case EVX_FPDEC_CVTGDS:
      case EVX_FPDEC_CVTGDSC:
      case EVX_FPDEC_CVTFG:
      case EVX_FPDEC_CVTFGC:
      case EVX_FPDEC_CVTFGS:
      case EVX_FPDEC_CVTFGSC:
      case EVX_FPDEC_CVTGF:
      case EVX_FPDEC_CVTGFC:
      case EVX_FPDEC_CVTGFS:
      case EVX_FPDEC_CVTGFSC:
      case EVX_FPDEC_CVTQD:
      case EVX_FPDEC_CVTQDC:
      case EVX_FPDEC_CVTQDS:
      case EVX_FPDEC_CVTQDSC:
      case EVX_FPDEC_CVTQF:
      case EVX_FPDEC_CVTQFC:
      case EVX_FPDEC_CVTQFS:
      case EVX_FPDEC_CVTQFSC:
      case EVX_FPDEC_CVTQG:
      case EVX_FPDEC_CVTQGC:
      case EVX_FPDEC_CVTQGS:
      case EVX_FPDEC_CVTQGSC:
      case EVX_FPDEC_CVTDQ:
      case EVX_FPDEC_CVTDQC:
      case EVX_FPDEC_CVTDQS:
      case EVX_FPDEC_CVTDQSC:
      case EVX_FPDEC_CVTGQ:
      case EVX_FPDEC_CVTGQC:
      case EVX_FPDEC_CVTGQS:
      case EVX_FPDEC_CVTGQSC:
	addres (USES, fb, class_fpop);
        addres (SETS, fc, class_fpop);
        addres (USES, (unsigned) resource_trap, class_null);
	break;
      default:
        addres (SETS, (unsigned) resource_pc, class_null);
        break;
      }
    }
    break;
  case EVX_OPC_FPIEEE: {
    func = inst.f_format.function;
    switch(func) {
      case EVX_FPIEEE_CMPSEQSU:
      case EVX_FPIEEE_CMPSLTSU:
      case EVX_FPIEEE_CMPSLESU:
      case EVX_FPIEEE_CMPSUNSU:
      case EVX_FPIEEE_CMPTEQSU:
      case EVX_FPIEEE_CMPTLTSU:
      case EVX_FPIEEE_CMPTLESU:
      case EVX_FPIEEE_CMPTUNSU:
      case EVX_FPIEEE_CMPSEQ:
      case EVX_FPIEEE_CMPSLT:
      case EVX_FPIEEE_CMPSLE:
      case EVX_FPIEEE_CMPSUN:
      case EVX_FPIEEE_CMPTEQ:
      case EVX_FPIEEE_CMPTLT:
      case EVX_FPIEEE_CMPTLE:
      case EVX_FPIEEE_CMPTUN:
        addres (USES, fa, class_fpop);
        addres (USES, fb, class_fpop);
        addres (SETS, fc, class_fpop);
        addres (USES, (unsigned) resource_trap, class_null);
	break;
      case EVX_FPIEEE_CVTQSSUI:
      case EVX_FPIEEE_CVTQSSUIC:
      case EVX_FPIEEE_CVTQSSUIM:
      case EVX_FPIEEE_CVTQSSUIP:
      case EVX_FPIEEE_CVTQTSUI:
      case EVX_FPIEEE_CVTQTSUIC:
      case EVX_FPIEEE_CVTQTSUIM:
      case EVX_FPIEEE_CVTQTSUIP:
      case EVX_FPIEEE_CVTSTSUI:
      case EVX_FPIEEE_CVTSTSUIC:
      case EVX_FPIEEE_CVTSTSUIM:
      case EVX_FPIEEE_CVTSTSUIP:
      case EVX_FPIEEE_CVTTSU:
      case EVX_FPIEEE_CVTTSUC:
      case EVX_FPIEEE_CVTTSUM:
      case EVX_FPIEEE_CVTTSUP:
      case EVX_FPIEEE_CVTTSSU:
      case EVX_FPIEEE_CVTTSSUC:
      case EVX_FPIEEE_CVTTSSUM:
      case EVX_FPIEEE_CVTTSSUP:
      case EVX_FPIEEE_CVTTSSUI:
      case EVX_FPIEEE_CVTTSSUIC:
      case EVX_FPIEEE_CVTTSSUIM:
      case EVX_FPIEEE_CVTTSSUIP:
      case EVX_FPIEEE_CVTQS:
      case EVX_FPIEEE_CVTQSC:
      case EVX_FPIEEE_CVTQSM:
      case EVX_FPIEEE_CVTQSP:
      case EVX_FPIEEE_CVTQT:
      case EVX_FPIEEE_CVTQTC:
      case EVX_FPIEEE_CVTQTM:
      case EVX_FPIEEE_CVTQTP:
      case EVX_FPIEEE_CVTST:
      case EVX_FPIEEE_CVTSTC:
      case EVX_FPIEEE_CVTSTM:
      case EVX_FPIEEE_CVTSTP:
      case EVX_FPIEEE_CVTTS:
      case EVX_FPIEEE_CVTTSC:
      case EVX_FPIEEE_CVTTSM:
      case EVX_FPIEEE_CVTTSP:
        addres (USES, fb, class_fpop);
        addres (SETS, fc, class_fpop);
	addres (USES, (unsigned) resource_trap, class_null);
	break;
      case EVX_FPIEEE_MULSU:
      case EVX_FPIEEE_MULSUC:
      case EVX_FPIEEE_MULSUM:
      case EVX_FPIEEE_MULSUP:
      case EVX_FPIEEE_MULSSU:
      case EVX_FPIEEE_MULSSUC:
      case EVX_FPIEEE_MULSSUM:
      case EVX_FPIEEE_MULSSUP:
      case EVX_FPIEEE_MULSSUI:
      case EVX_FPIEEE_MULSSUIC:
      case EVX_FPIEEE_MULSSUIM:
      case EVX_FPIEEE_MULSSUIP:
      case EVX_FPIEEE_MULTU:
      case EVX_FPIEEE_MULTUC:
      case EVX_FPIEEE_MULTUM:
      case EVX_FPIEEE_MULTUP:
      case EVX_FPIEEE_MULTSU:
      case EVX_FPIEEE_MULTSUC:
      case EVX_FPIEEE_MULTSUM:
      case EVX_FPIEEE_MULTSUP:
      case EVX_FPIEEE_MULTSUI:
      case EVX_FPIEEE_MULTSUIC:
      case EVX_FPIEEE_MULTSUIM:
      case EVX_FPIEEE_MULTSUIP:
      case EVX_FPIEEE_SUBSU:
      case EVX_FPIEEE_SUBSUC:
      case EVX_FPIEEE_SUBSUM:
      case EVX_FPIEEE_SUBSUP:
      case EVX_FPIEEE_SUBSSU:
      case EVX_FPIEEE_SUBSSUC:
      case EVX_FPIEEE_SUBSSUM:
      case EVX_FPIEEE_SUBSSUP:
      case EVX_FPIEEE_SUBSSUI:
      case EVX_FPIEEE_SUBSSUIC:
      case EVX_FPIEEE_SUBSSUIM:
      case EVX_FPIEEE_SUBSSUIP:
      case EVX_FPIEEE_SUBTU:
      case EVX_FPIEEE_SUBTUC:
      case EVX_FPIEEE_SUBTUM:
      case EVX_FPIEEE_SUBTUP:
      case EVX_FPIEEE_SUBTSU:
      case EVX_FPIEEE_SUBTSUC:
      case EVX_FPIEEE_SUBTSUM:
      case EVX_FPIEEE_SUBTSUP:
      case EVX_FPIEEE_SUBTSUI:
      case EVX_FPIEEE_SUBTSUIC:
      case EVX_FPIEEE_SUBTSUIM:
      case EVX_FPIEEE_SUBTSUIP:
      case EVX_FPIEEE_ADDSU:
      case EVX_FPIEEE_ADDSUC:
      case EVX_FPIEEE_ADDSUM:
      case EVX_FPIEEE_ADDSUP:
      case EVX_FPIEEE_ADDSSU:
      case EVX_FPIEEE_ADDSSUC:
      case EVX_FPIEEE_ADDSSUM:
      case EVX_FPIEEE_ADDSSUP:
      case EVX_FPIEEE_ADDSSUI:
      case EVX_FPIEEE_ADDSSUIC:
      case EVX_FPIEEE_ADDSSUIM:
      case EVX_FPIEEE_ADDSSUIP:
      case EVX_FPIEEE_ADDTU:
      case EVX_FPIEEE_ADDTUC:
      case EVX_FPIEEE_ADDTUM:
      case EVX_FPIEEE_ADDTUP:
      case EVX_FPIEEE_ADDTSU:
      case EVX_FPIEEE_ADDTSUC:
      case EVX_FPIEEE_ADDTSUM:
      case EVX_FPIEEE_ADDTSUP:
      case EVX_FPIEEE_ADDTSUI:
      case EVX_FPIEEE_ADDTSUIC:
      case EVX_FPIEEE_ADDTSUIM:
      case EVX_FPIEEE_ADDTSUIP:
      case EVX_FPIEEE_MULS:
      case EVX_FPIEEE_MULSC:
      case EVX_FPIEEE_MULSM:
      case EVX_FPIEEE_MULSP:
      case EVX_FPIEEE_MULT:
      case EVX_FPIEEE_MULTC:
      case EVX_FPIEEE_MULTM:
      case EVX_FPIEEE_MULTP:
      case EVX_FPIEEE_SUBS:
      case EVX_FPIEEE_SUBSC:
      case EVX_FPIEEE_SUBSM:
      case EVX_FPIEEE_SUBSP:
      case EVX_FPIEEE_SUBT:
      case EVX_FPIEEE_SUBTC:
      case EVX_FPIEEE_SUBTM:
      case EVX_FPIEEE_SUBTP:
      case EVX_FPIEEE_ADDS:
      case EVX_FPIEEE_ADDSC:
      case EVX_FPIEEE_ADDSM:
      case EVX_FPIEEE_ADDSP:
      case EVX_FPIEEE_ADDT:
      case EVX_FPIEEE_ADDTC:
      case EVX_FPIEEE_ADDTM:
      case EVX_FPIEEE_ADDTP:
        addres (USES, fa, class_fpop);
        addres (USES, fb, class_fpop);
	addres (SETS, fc, class_fpop);
	addres (USES, (unsigned) resource_trap, class_null);
	break;
      case EVX_FPIEEE_DIVS:
      case EVX_FPIEEE_DIVSC:
      case EVX_FPIEEE_DIVSM:
      case EVX_FPIEEE_DIVSP:
      case EVX_FPIEEE_DIVSU:
      case EVX_FPIEEE_DIVSUC:
      case EVX_FPIEEE_DIVSUM:
      case EVX_FPIEEE_DIVSUP:
      case EVX_FPIEEE_DIVSSU:
      case EVX_FPIEEE_DIVSSUC:
      case EVX_FPIEEE_DIVSSUM:
      case EVX_FPIEEE_DIVSSUP:
      case EVX_FPIEEE_DIVSSUI:
      case EVX_FPIEEE_DIVSSUIC:
      case EVX_FPIEEE_DIVSSUIM:
      case EVX_FPIEEE_DIVSSUIP:
        addres (USES, fa, class_fdivs);
	addres (USES, fb, class_fdivs);
	addres (SETS, fc, class_fdivs);
        addres (USES, (unsigned) resource_trap, class_null);
	break;
      case EVX_FPIEEE_DIVT:
      case EVX_FPIEEE_DIVTC:
      case EVX_FPIEEE_DIVTM:
      case EVX_FPIEEE_DIVTP:
      case EVX_FPIEEE_DIVTU:
      case EVX_FPIEEE_DIVTUC:
      case EVX_FPIEEE_DIVTUM:
      case EVX_FPIEEE_DIVTUP:
      case EVX_FPIEEE_DIVTSU:
      case EVX_FPIEEE_DIVTSUC:
      case EVX_FPIEEE_DIVTSUM:
      case EVX_FPIEEE_DIVTSUP:
      case EVX_FPIEEE_DIVTSUI:
      case EVX_FPIEEE_DIVTSUIC:
      case EVX_FPIEEE_DIVTSUIM:
      case EVX_FPIEEE_DIVTSUIP:
	addres (USES, fa, class_fdivd);
	addres (USES, fb, class_fdivd);
        addres (SETS, fc, class_fdivd);
	addres (USES, (unsigned) resource_trap, class_null);
	break;
      case EVX_FPIEEE_CVTSQV:
      case EVX_FPIEEE_CVTSQVC:
      case EVX_FPIEEE_CVTSQSV:
      case EVX_FPIEEE_CVTSQSVC:
      case EVX_FPIEEE_CVTSQSVI:
      case EVX_FPIEEE_CVTSQSVIC:
      case EVX_FPIEEE_CVTTQV:
      case EVX_FPIEEE_CVTTQVC:
      case EVX_FPIEEE_CVTTQSV:
      case EVX_FPIEEE_CVTTQSVC:
      case EVX_FPIEEE_CVTTQSVI:
      case EVX_FPIEEE_CVTTQSVIC:
      case EVX_FPIEEE_CVTTQSVD:
      case EVX_FPIEEE_CVTTQSVID:
      case EVX_FPIEEE_CVTTQVD:
      case EVX_FPIEEE_CVTTQVM:
      case EVX_FPIEEE_CVTTQSVM:
      case EVX_FPIEEE_CVTTQSVIM:
      case EVX_FPIEEE_CVTSQ:
      case EVX_FPIEEE_CVTSQC:
      case EVX_FPIEEE_CVTTQ:
      case EVX_FPIEEE_CVTTQC:
      case EVX_FPIEEE_CVTTQD:
      case EVX_FPIEEE_CVTTQM:
        addres (USES, fb, class_fpop);
        addres (SETS, fc, class_fpop);
        addres (USES, (unsigned) resource_trap, class_fpop);
        break;
      default:
        addres (SETS, (unsigned) resource_pc, class_null);
        break;
      }
    }
    break;
  case EVX_OPC_PAL: {
    func = inst.pal_format.function;
    switch(func) {
      case EVX_PAL_BPT:
      case EVX_PAL_BUGCHK:
      case EVX_PAL_CHME:
      case EVX_PAL_CHMK:
      case EVX_PAL_CHMS:
      case EVX_PAL_CHMU:
      case EVX_PAL_IMB:
        break;
      case EVX_PAL_INSQHIL:
      case EVX_PAL_INSQTIL:
      case EVX_PAL_INSQHIQ:
      case EVX_PAL_INSQTIQ:
      case EVX_PAL_INSQUEL:
      case EVX_PAL_INSQUEQ:
      case EVX_PAL_INSQUELD:
      case EVX_PAL_INSQUEQD:
        addres (USES, 16, class_null);
	addres (USES, 17, class_null);
	addres (SETS,  0, class_null);
	break;
      case EVX_PAL_PROBER:
      case EVX_PAL_PROBEW:
        addres (USES, 16, class_null);
	addres (USES, 17, class_null);
	addres (USES, 18, class_null);
	addres (SETS,  0, class_null);
	break;
      case EVX_PAL_RD_PS:
	addres (SETS, 0, class_null);
	break;
      case EVX_PAL_REI:
        break;
      case EVX_PAL_REMQHIL:
      case EVX_PAL_REMQTIL:
      case EVX_PAL_REMQHIQ:
      case EVX_PAL_REMQTIQ:
      case EVX_PAL_REMQUEL:
      case EVX_PAL_REMQUEQ:
      case EVX_PAL_REMQUELD:
      case EVX_PAL_REMQUEQD:
        addres (USES, 16, class_null);
	addres (SETS,  0, class_null);
	addres (SETS,  1, class_null);
	break;
      case EVX_PAL_SWASTEN:
        addres (USES, 16, class_null);
	addres (SETS,  0, class_null);
	break;
      case EVX_PAL_WR_PS_SW:
        addres (USES, 16, class_null);
	break;
      case EVX_PAL_RSCC:
      case EVX_PAL_READ_UNQ:
        addres (SETS, 0, class_null);
	break;
      case EVX_PAL_WRITE_UNQ:
        addres (USES, 16, class_null);
	break;
      case EVX_PAL_CFLUSH:
	addres (USES, 16, class_null);
	break;
      case EVX_PAL_DRAINA:
      case EVX_PAL_HALT:
        break;
      case EVX_PAL_LDQP:
        addres (USES, 16, class_null);
        addres (SETS,  0, class_null);
	break;
      case EVX_PAL_STQP:
        addres (USES, 16, class_null);
        addres (USES, 17, class_null);
	break;
      case EVX_PAL_SWPCTX:
        addres (USES, 16, class_null);
	break;
      case EVX_PAL_MFPR_ASN:
      case EVX_PAL_MTPR_ASTEN:
      case EVX_PAL_MTPR_ASTSR:
      case EVX_PAL_MFPR_AT:
      case EVX_PAL_MTPR_AT:
      case EVX_PAL_MFPR_FEN:
      case EVX_PAL_MTPR_FEN:
      case EVX_PAL_MTPR_IPIR:
      case EVX_PAL_MFPR_IPL:
      case EVX_PAL_MTPR_IPL:
      case EVX_PAL_MFPR_MCES:
      case EVX_PAL_MTPR_MCES:
      case EVX_PAL_MFPR_PCBB:
      case EVX_PAL_MFPR_PRBR:
      case EVX_PAL_MTPR_PRBR:
      case EVX_PAL_MFPR_PTBR:
      case EVX_PAL_MFPR_SCBB:
      case EVX_PAL_MTPR_SCBB:
      case EVX_PAL_MTPR_SIRR:
      case EVX_PAL_MFPR_SISR:
      case EVX_PAL_MFPR_TBCHK:
      case EVX_PAL_MTPR_TBIA:
      case EVX_PAL_MTPR_TBIAP:
      case EVX_PAL_MTPR_TBIS:
      case EVX_PAL_MFPR_ESP:
      case EVX_PAL_MTPR_ESP:
      case EVX_PAL_MFPR_SSP:
      case EVX_PAL_MTPR_SSP:
      case EVX_PAL_MFPR_USP:
      case EVX_PAL_MTPR_USP:
      case EVX_PAL_MTPR_TBISD:
      case EVX_PAL_MTPR_TBISI:
      case EVX_PAL_MFPR_ASTEN:
      case EVX_PAL_MFPR_ASTSR:
      case EVX_PAL_MFPR_WHAMI:
        addres (USES, 16, class_null);
        addres (SETS,  0, class_null);
        addres (SETS,  1, class_null);
        addres (SETS, 16, class_null);
	break;
      case EVX_PAL_RESERVED_GETSS:
      case EVX_PAL_RESERVED_PUTSS:
      case EVX_PAL_RESERVED_RLS1:
      case EVX_PAL_RESERVED_RLS2:
      case EVX_PAL_RESERVED_DVT1:
      case EVX_PAL_NPHALT:
      case EVX_PAL_RESERVED_DVT:
      case EVX_PAL_RESERVED_SIM:
        break;
      default:
        break;
      }
    }
    addres (SETS, (unsigned) resource_pc, class_null);
    break;
  case EVX_OPC_PAL19:
  case EVX_OPC_PAL1B:
  case EVX_OPC_PAL1D:
  case EVX_OPC_PAL1E:
  case EVX_OPC_PAL1F:
    /* Opcodes reserved to pal code */
  case EVX_OPC_OPC01:
  case EVX_OPC_OPC02:
  case EVX_OPC_OPC03:
  case EVX_OPC_OPC04:
  case EVX_OPC_OPC05:
  case EVX_OPC_OPC06:
  case EVX_OPC_OPC07:
  case EVX_OPC_OPC0A:
  case EVX_OPC_OPC0C:
  case EVX_OPC_OPC0D:
  case EVX_OPC_OPC0E:
  case EVX_OPC_OPC14:
  case EVX_OPC_OPC1C:
    /* Opcodes reserved to Digital */
  default:
    addres (SETS, (unsigned) resource_pc, class_null);
    break;
  }
}

/*
**	Fill up the memdefuse structure with the memory access information
**	for 'inst'.
*/
static void memory_defuse (alpha_inst inst, memory_resource *mem)
{
  register unsigned opcode;

  opcode = inst.o_format.opcode;

  switch (opcode) {
  case EVX_OPC_LDL:
  case EVX_OPC_LDLL:
  case EVX_OPC_LDF:
  case EVX_OPC_LDS:
    mem->base   = inst.m_format.rb;
    mem->size   = 4;
    mem->offset = inst.m_format.memory_displacement;
    mem->def    = USES;
    break;
  case EVX_OPC_LDQ:
  case EVX_OPC_LDQL:
  case EVX_OPC_LDD:
  case EVX_OPC_LDT:
    mem->base   = inst.m_format.rb;
    mem->size   = 8;
    mem->offset = inst.m_format.memory_displacement;
    mem->def    = USES;
    break;
  case EVX_OPC_LDQU:
    /*
    //	The adjustment to size and offset are needed, because the hardware
    //	ignores the low order 3 address bits. Since there is not information
    //	about the contents of the base register, in order to determine overlapping
    //	regions of memory, the safe thing to do is to reduce the offset by 7
    //	and increase the size by 7.
    */
    mem->base   = inst.m_format.rb;
    mem->size   = 8 + UNALIGNED_ACCESS_SIZE_ADJUSTMENT;
    mem->offset = inst.m_format.memory_displacement + UNALIGNED_ACCESS_OFFSET_ADJUSTMENT;
    mem->def    = USES;
    break;
  case EVX_OPC_STL:
  case EVX_OPC_STLC:
  case EVX_OPC_STF:
  case EVX_OPC_STS:
    mem->base   = inst.m_format.rb;
    mem->size   = 4;
    mem->offset = inst.m_format.memory_displacement;
    mem->def    = SETS;
    break;
  case EVX_OPC_STQ:
  case EVX_OPC_STQC:
  case EVX_OPC_STD:
  case EVX_OPC_STT:
    mem->base   = inst.m_format.rb;
    mem->size   = 8;
    mem->offset = inst.m_format.memory_displacement;
    mem->def    = SETS;
    break;
  case EVX_OPC_STQU:
    /*
    //	The adjustment to size and offset are needed, because the hardware
    //	ignores the low order 3 address bits. Since there is not information
    //	about the contents of the base register, in order to determine overlapping
    //	regions of memory, the safe thing to do is to reduce the offset by 7
    //	and increase the size by 7.
    */
    mem->base   = inst.m_format.rb;
    mem->size   = 8 + UNALIGNED_ACCESS_SIZE_ADJUSTMENT;
    mem->offset = inst.m_format.memory_displacement + UNALIGNED_ACCESS_OFFSET_ADJUSTMENT;
    mem->def    = SETS;
    break;
  default:
    break;
  }
}

/*
**	Keep track of the functional units used by instructions along
**	with the number of cycles for which they are kept busy.
**
**	This function can use the instruction_class function to
**	determine the functional unit usage (later...)
*/
static void function_unit_use (alpha_inst inst, fu_func funcp)
{
  register unsigned opcode, func;

  opcode = inst.o_format.opcode;

  switch (opcode) {
  case EVX_OPC_LDL:
  case EVX_OPC_LDQ:
  case EVX_OPC_LDQU:
  case EVX_OPC_LDLL:
  case EVX_OPC_LDQL:
  case EVX_OPC_LDF:
  case EVX_OPC_LDD:
  case EVX_OPC_LDS:
  case EVX_OPC_LDT:
  case EVX_OPC_STL:
  case EVX_OPC_STQ:
  case EVX_OPC_STQU:
  case EVX_OPC_STLC:
  case EVX_OPC_STQC:
  case EVX_OPC_STF:
  case EVX_OPC_STD:
  case EVX_OPC_STS:
  case EVX_OPC_STT:
    (*funcp) (fu_memory_pipeline, 1);
    break;
  case EVX_OPC_LDAH:
  case EVX_OPC_LDA:
    (*funcp) (fu_integer_pipeline, 1);
    break;
  case EVX_OPC_SYNC:
    (*funcp) (fu_memory_pipeline, 1);
    break;
  case EVX_OPC_BR:
  case EVX_OPC_JSR:
  case EVX_OPC_BSR:
    (*funcp) (fu_branch_unit, 1);
    break;
  case EVX_OPC_BLBC:
  case EVX_OPC_BLBS:
  case EVX_OPC_BEQ:
  case EVX_OPC_BNE:
  case EVX_OPC_BLT:
  case EVX_OPC_BLE:
  case EVX_OPC_BGT:
  case EVX_OPC_BGE:
    (*funcp) (fu_branch_unit, 1);
    break;
  case EVX_OPC_FBEQ:
  case EVX_OPC_FBNE:
  case EVX_OPC_FBLT:
  case EVX_OPC_FBLE:
  case EVX_OPC_FBGT:
  case EVX_OPC_FBGE:
    (*funcp) (fu_branch_unit, 1);
    break;
  case EVX_OPC_ADDI:
  case EVX_OPC_LOGI:
  case EVX_OPC_SHFI:
    (*funcp) (fu_integer_pipeline, 1);
    break;
  case EVX_OPC_MULI: {
    func = inst.o_format.function;
    switch(func) {
      case EVX_MULI_MULQV:
      case EVX_MULI_UMULH:
      case EVX_MULI_MULQ:
        (*funcp) (fu_imultiplier, 21);
        break;
      case EVX_MULI_MULL:
      case EVX_MULI_MULLV:
        (*funcp) (fu_imultiplier, 19);
        break;
      default:
        break;
      }
    }
    break;
  case EVX_OPC_FPAUX:
    (*funcp) (fu_floating_pipeline, 1);
    break;
  case EVX_OPC_FPDEC:
    func = inst.f_format.function;
    switch(func) {
      case EVX_FPDEC_CVTDGU:
      case EVX_FPDEC_CVTDGUC:
      case EVX_FPDEC_CVTDGSU:
      case EVX_FPDEC_CVTDGSUC:
      case EVX_FPDEC_CVTGDU:
      case EVX_FPDEC_CVTGDUC:
      case EVX_FPDEC_CVTGDSU:
      case EVX_FPDEC_CVTGDSUC:
      case EVX_FPDEC_CVTGFU:
      case EVX_FPDEC_CVTGFUC:
      case EVX_FPDEC_CVTGFSU:
      case EVX_FPDEC_CVTGFSUC:
      case EVX_FPDEC_CVTDQV:
      case EVX_FPDEC_CVTDQVC:
      case EVX_FPDEC_CVTDQSV:
      case EVX_FPDEC_CVTDQSVC:
      case EVX_FPDEC_CVTGQV:
      case EVX_FPDEC_CVTGQVC:
      case EVX_FPDEC_CVTGQSV:
      case EVX_FPDEC_CVTGQSVC:
      case EVX_FPDEC_CVTDF:
      case EVX_FPDEC_CVTDFC:
      case EVX_FPDEC_CVTDFS:
      case EVX_FPDEC_CVTDFSC:
      case EVX_FPDEC_CVTDG:
      case EVX_FPDEC_CVTDGC:
      case EVX_FPDEC_CVTDGS:
      case EVX_FPDEC_CVTDGSC:
      case EVX_FPDEC_CVTGD:
      case EVX_FPDEC_CVTGDC:
      case EVX_FPDEC_CVTGDS:
      case EVX_FPDEC_CVTGDSC:
      case EVX_FPDEC_CVTFG:
      case EVX_FPDEC_CVTFGC:
      case EVX_FPDEC_CVTFGS:
      case EVX_FPDEC_CVTFGSC:
      case EVX_FPDEC_CVTGF:
      case EVX_FPDEC_CVTGFC:
      case EVX_FPDEC_CVTGFS:
      case EVX_FPDEC_CVTGFSC:
      case EVX_FPDEC_CVTQD:
      case EVX_FPDEC_CVTQDC:
      case EVX_FPDEC_CVTQDS:
      case EVX_FPDEC_CVTQDSC:
      case EVX_FPDEC_CVTQF:
      case EVX_FPDEC_CVTQFC:
      case EVX_FPDEC_CVTQFS:
      case EVX_FPDEC_CVTQFSC:
      case EVX_FPDEC_CVTQG:
      case EVX_FPDEC_CVTQGC:
      case EVX_FPDEC_CVTQGS:
      case EVX_FPDEC_CVTQGSC:
      case EVX_FPDEC_CVTDQ:
      case EVX_FPDEC_CVTDQC:
      case EVX_FPDEC_CVTDQS:
      case EVX_FPDEC_CVTDQSC:
      case EVX_FPDEC_CVTGQ:
      case EVX_FPDEC_CVTGQC:
      case EVX_FPDEC_CVTGQS:
      case EVX_FPDEC_CVTGQSC:
      case EVX_FPDEC_ADDDU:
      case EVX_FPDEC_ADDDUC:
      case EVX_FPDEC_ADDDSU:
      case EVX_FPDEC_ADDDSUC:
      case EVX_FPDEC_ADDFU:
      case EVX_FPDEC_ADDFUC:
      case EVX_FPDEC_ADDFSU:
      case EVX_FPDEC_ADDFSUC:
      case EVX_FPDEC_ADDGU:
      case EVX_FPDEC_ADDGUC:
      case EVX_FPDEC_ADDGSU:
      case EVX_FPDEC_ADDGSUC:
      case EVX_FPDEC_SUBDU:
      case EVX_FPDEC_SUBDUC:
      case EVX_FPDEC_SUBDSU:
      case EVX_FPDEC_SUBDSUC:
      case EVX_FPDEC_SUBFU:
      case EVX_FPDEC_SUBFUC:
      case EVX_FPDEC_SUBFSU:
      case EVX_FPDEC_SUBFSUC:
      case EVX_FPDEC_SUBGU:
      case EVX_FPDEC_SUBGUC:
      case EVX_FPDEC_SUBGSU:
      case EVX_FPDEC_SUBGSUC:
      case EVX_FPDEC_MULDU:
      case EVX_FPDEC_MULDUC:
      case EVX_FPDEC_MULDSU:
      case EVX_FPDEC_MULDSUC:
      case EVX_FPDEC_MULFU:
      case EVX_FPDEC_MULFUC:
      case EVX_FPDEC_MULFSU:
      case EVX_FPDEC_MULFSUC:
      case EVX_FPDEC_MULGU:
      case EVX_FPDEC_MULGUC:
      case EVX_FPDEC_MULGSU:
      case EVX_FPDEC_MULGSUC:
      case EVX_FPDEC_ADDD:
      case EVX_FPDEC_ADDDC:
      case EVX_FPDEC_ADDDS:
      case EVX_FPDEC_ADDDSC:
      case EVX_FPDEC_ADDF:
      case EVX_FPDEC_ADDFC:
      case EVX_FPDEC_ADDFS:
      case EVX_FPDEC_ADDFSC:
      case EVX_FPDEC_ADDG:
      case EVX_FPDEC_ADDGC:
      case EVX_FPDEC_ADDGS:
      case EVX_FPDEC_ADDGSC:
      case EVX_FPDEC_SUBD:
      case EVX_FPDEC_SUBDC:
      case EVX_FPDEC_SUBDS:
      case EVX_FPDEC_SUBDSC:
      case EVX_FPDEC_SUBF:
      case EVX_FPDEC_SUBFC:
      case EVX_FPDEC_SUBFS:
      case EVX_FPDEC_SUBFSC:
      case EVX_FPDEC_SUBG:
      case EVX_FPDEC_SUBGC:
      case EVX_FPDEC_SUBGS:
      case EVX_FPDEC_SUBGSC:
      case EVX_FPDEC_MULD:
      case EVX_FPDEC_MULDC:
      case EVX_FPDEC_MULDS:
      case EVX_FPDEC_MULDSC:
      case EVX_FPDEC_MULF:
      case EVX_FPDEC_MULFC:
      case EVX_FPDEC_MULFS:
      case EVX_FPDEC_MULFSC:
      case EVX_FPDEC_MULG:
      case EVX_FPDEC_MULGC:
      case EVX_FPDEC_MULGS:
      case EVX_FPDEC_MULGSC:
      case EVX_FPDEC_CMPDEQ:
      case EVX_FPDEC_CMPDLT:
      case EVX_FPDEC_CMPDLE:
      case EVX_FPDEC_CMPGEQ:
      case EVX_FPDEC_CMPGLT:
      case EVX_FPDEC_CMPGLE:
      case EVX_FPDEC_CMPDEQS:
      case EVX_FPDEC_CMPDLTS:
      case EVX_FPDEC_CMPDLES:
      case EVX_FPDEC_CMPGEQS:
      case EVX_FPDEC_CMPGLTS:
      case EVX_FPDEC_CMPGLES:
        (*funcp) (fu_floating_pipeline, 1);
	break;
      case EVX_FPDEC_DIVF:
      case EVX_FPDEC_DIVFC:
      case EVX_FPDEC_DIVFS:
      case EVX_FPDEC_DIVFSC:
      case EVX_FPDEC_DIVFU:
      case EVX_FPDEC_DIVFUC:
      case EVX_FPDEC_DIVFSU:
      case EVX_FPDEC_DIVFSUC:
        (*funcp) (fu_fpdivider, 30);
	break;
      case EVX_FPDEC_DIVD:
      case EVX_FPDEC_DIVDC:
      case EVX_FPDEC_DIVDS:
      case EVX_FPDEC_DIVDSC:
      case EVX_FPDEC_DIVDU:
      case EVX_FPDEC_DIVDUC:
      case EVX_FPDEC_DIVDSU:
      case EVX_FPDEC_DIVDSUC:
      case EVX_FPDEC_DIVG:
      case EVX_FPDEC_DIVGC:
      case EVX_FPDEC_DIVGS:
      case EVX_FPDEC_DIVGSC:
      case EVX_FPDEC_DIVGU:
      case EVX_FPDEC_DIVGUC:
      case EVX_FPDEC_DIVGSU:
      case EVX_FPDEC_DIVGSUC:
        (*funcp) (fu_fpdivider, 59);
	break;
      default:
        break;
      }
    break;
  case EVX_OPC_FPIEEE:
    func = inst.f_format.function;
    switch(func) {
      case EVX_FPIEEE_CMPSEQSU:
      case EVX_FPIEEE_CMPSLTSU:
      case EVX_FPIEEE_CMPSLESU:
      case EVX_FPIEEE_CMPSUNSU:
      case EVX_FPIEEE_CMPTEQSU:
      case EVX_FPIEEE_CMPTLTSU:
      case EVX_FPIEEE_CMPTLESU:
      case EVX_FPIEEE_CMPTUNSU:
      case EVX_FPIEEE_CMPSEQ:
      case EVX_FPIEEE_CMPSLT:
      case EVX_FPIEEE_CMPSLE:
      case EVX_FPIEEE_CMPSUN:
      case EVX_FPIEEE_CMPTEQ:
      case EVX_FPIEEE_CMPTLT:
      case EVX_FPIEEE_CMPTLE:
      case EVX_FPIEEE_CMPTUN:
      case EVX_FPIEEE_CVTQSSUI:
      case EVX_FPIEEE_CVTQSSUIC:
      case EVX_FPIEEE_CVTQSSUIM:
      case EVX_FPIEEE_CVTQSSUIP:
      case EVX_FPIEEE_CVTQTSUI:
      case EVX_FPIEEE_CVTQTSUIC:
      case EVX_FPIEEE_CVTQTSUIM:
      case EVX_FPIEEE_CVTQTSUIP:
      case EVX_FPIEEE_CVTSTSUI:
      case EVX_FPIEEE_CVTSTSUIC:
      case EVX_FPIEEE_CVTSTSUIM:
      case EVX_FPIEEE_CVTSTSUIP:
      case EVX_FPIEEE_CVTTSU:
      case EVX_FPIEEE_CVTTSUC:
      case EVX_FPIEEE_CVTTSUM:
      case EVX_FPIEEE_CVTTSUP:
      case EVX_FPIEEE_CVTTSSU:
      case EVX_FPIEEE_CVTTSSUC:
      case EVX_FPIEEE_CVTTSSUM:
      case EVX_FPIEEE_CVTTSSUP:
      case EVX_FPIEEE_CVTTSSUI:
      case EVX_FPIEEE_CVTTSSUIC:
      case EVX_FPIEEE_CVTTSSUIM:
      case EVX_FPIEEE_CVTTSSUIP:
      case EVX_FPIEEE_CVTQS:
      case EVX_FPIEEE_CVTQSC:
      case EVX_FPIEEE_CVTQSM:
      case EVX_FPIEEE_CVTQSP:
      case EVX_FPIEEE_CVTQT:
      case EVX_FPIEEE_CVTQTC:
      case EVX_FPIEEE_CVTQTM:
      case EVX_FPIEEE_CVTQTP:
      case EVX_FPIEEE_CVTST:
      case EVX_FPIEEE_CVTSTC:
      case EVX_FPIEEE_CVTSTM:
      case EVX_FPIEEE_CVTSTP:
      case EVX_FPIEEE_CVTTS:
      case EVX_FPIEEE_CVTTSC:
      case EVX_FPIEEE_CVTTSM:
      case EVX_FPIEEE_CVTTSP:
      case EVX_FPIEEE_MULSU:
      case EVX_FPIEEE_MULSUC:
      case EVX_FPIEEE_MULSUM:
      case EVX_FPIEEE_MULSUP:
      case EVX_FPIEEE_MULSSU:
      case EVX_FPIEEE_MULSSUC:
      case EVX_FPIEEE_MULSSUM:
      case EVX_FPIEEE_MULSSUP:
      case EVX_FPIEEE_MULSSUI:
      case EVX_FPIEEE_MULSSUIC:
      case EVX_FPIEEE_MULSSUIM:
      case EVX_FPIEEE_MULSSUIP:
      case EVX_FPIEEE_MULTU:
      case EVX_FPIEEE_MULTUC:
      case EVX_FPIEEE_MULTUM:
      case EVX_FPIEEE_MULTUP:
      case EVX_FPIEEE_MULTSU:
      case EVX_FPIEEE_MULTSUC:
      case EVX_FPIEEE_MULTSUM:
      case EVX_FPIEEE_MULTSUP:
      case EVX_FPIEEE_MULTSUI:
      case EVX_FPIEEE_MULTSUIC:
      case EVX_FPIEEE_MULTSUIM:
      case EVX_FPIEEE_MULTSUIP:
      case EVX_FPIEEE_SUBSU:
      case EVX_FPIEEE_SUBSUC:
      case EVX_FPIEEE_SUBSUM:
      case EVX_FPIEEE_SUBSUP:
      case EVX_FPIEEE_SUBSSU:
      case EVX_FPIEEE_SUBSSUC:
      case EVX_FPIEEE_SUBSSUM:
      case EVX_FPIEEE_SUBSSUP:
      case EVX_FPIEEE_SUBSSUI:
      case EVX_FPIEEE_SUBSSUIC:
      case EVX_FPIEEE_SUBSSUIM:
      case EVX_FPIEEE_SUBSSUIP:
      case EVX_FPIEEE_SUBTU:
      case EVX_FPIEEE_SUBTUC:
      case EVX_FPIEEE_SUBTUM:
      case EVX_FPIEEE_SUBTUP:
      case EVX_FPIEEE_SUBTSU:
      case EVX_FPIEEE_SUBTSUC:
      case EVX_FPIEEE_SUBTSUM:
      case EVX_FPIEEE_SUBTSUP:
      case EVX_FPIEEE_SUBTSUI:
      case EVX_FPIEEE_SUBTSUIC:
      case EVX_FPIEEE_SUBTSUIM:
      case EVX_FPIEEE_SUBTSUIP:
      case EVX_FPIEEE_ADDSU:
      case EVX_FPIEEE_ADDSUC:
      case EVX_FPIEEE_ADDSUM:
      case EVX_FPIEEE_ADDSUP:
      case EVX_FPIEEE_ADDSSU:
      case EVX_FPIEEE_ADDSSUC:
      case EVX_FPIEEE_ADDSSUM:
      case EVX_FPIEEE_ADDSSUP:
      case EVX_FPIEEE_ADDSSUI:
      case EVX_FPIEEE_ADDSSUIC:
      case EVX_FPIEEE_ADDSSUIM:
      case EVX_FPIEEE_ADDSSUIP:
      case EVX_FPIEEE_ADDTU:
      case EVX_FPIEEE_ADDTUC:
      case EVX_FPIEEE_ADDTUM:
      case EVX_FPIEEE_ADDTUP:
      case EVX_FPIEEE_ADDTSU:
      case EVX_FPIEEE_ADDTSUC:
      case EVX_FPIEEE_ADDTSUM:
      case EVX_FPIEEE_ADDTSUP:
      case EVX_FPIEEE_ADDTSUI:
      case EVX_FPIEEE_ADDTSUIC:
      case EVX_FPIEEE_ADDTSUIM:
      case EVX_FPIEEE_ADDTSUIP:
      case EVX_FPIEEE_MULS:
      case EVX_FPIEEE_MULSC:
      case EVX_FPIEEE_MULSM:
      case EVX_FPIEEE_MULSP:
      case EVX_FPIEEE_MULT:
      case EVX_FPIEEE_MULTC:
      case EVX_FPIEEE_MULTM:
      case EVX_FPIEEE_MULTP:
      case EVX_FPIEEE_SUBS:
      case EVX_FPIEEE_SUBSC:
      case EVX_FPIEEE_SUBSM:
      case EVX_FPIEEE_SUBSP:
      case EVX_FPIEEE_SUBT:
      case EVX_FPIEEE_SUBTC:
      case EVX_FPIEEE_SUBTM:
      case EVX_FPIEEE_SUBTP:
      case EVX_FPIEEE_ADDS:
      case EVX_FPIEEE_ADDSC:
      case EVX_FPIEEE_ADDSM:
      case EVX_FPIEEE_ADDSP:
      case EVX_FPIEEE_ADDT:
      case EVX_FPIEEE_ADDTC:
      case EVX_FPIEEE_ADDTM:
      case EVX_FPIEEE_ADDTP:
        (*funcp) (fu_floating_pipeline, 1);
	break;
      case EVX_FPIEEE_DIVS:
      case EVX_FPIEEE_DIVSC:
      case EVX_FPIEEE_DIVSM:
      case EVX_FPIEEE_DIVSP:
      case EVX_FPIEEE_DIVSU:
      case EVX_FPIEEE_DIVSUC:
      case EVX_FPIEEE_DIVSUM:
      case EVX_FPIEEE_DIVSUP:
      case EVX_FPIEEE_DIVSSU:
      case EVX_FPIEEE_DIVSSUC:
      case EVX_FPIEEE_DIVSSUM:
      case EVX_FPIEEE_DIVSSUP:
      case EVX_FPIEEE_DIVSSUI:
      case EVX_FPIEEE_DIVSSUIC:
      case EVX_FPIEEE_DIVSSUIM:
      case EVX_FPIEEE_DIVSSUIP:
        (*funcp) (fu_fpdivider, 30);
	break;
      case EVX_FPIEEE_DIVT:
      case EVX_FPIEEE_DIVTC:
      case EVX_FPIEEE_DIVTM:
      case EVX_FPIEEE_DIVTP:
      case EVX_FPIEEE_DIVTU:
      case EVX_FPIEEE_DIVTUC:
      case EVX_FPIEEE_DIVTUM:
      case EVX_FPIEEE_DIVTUP:
      case EVX_FPIEEE_DIVTSU:
      case EVX_FPIEEE_DIVTSUC:
      case EVX_FPIEEE_DIVTSUM:
      case EVX_FPIEEE_DIVTSUP:
      case EVX_FPIEEE_DIVTSUI:
      case EVX_FPIEEE_DIVTSUIC:
      case EVX_FPIEEE_DIVTSUIM:
      case EVX_FPIEEE_DIVTSUIP:
        (*funcp) (fu_fpdivider, 59);
	break;
      case EVX_FPIEEE_CVTSQV:
      case EVX_FPIEEE_CVTSQVC:
      case EVX_FPIEEE_CVTSQSV:
      case EVX_FPIEEE_CVTSQSVC:
      case EVX_FPIEEE_CVTSQSVI:
      case EVX_FPIEEE_CVTSQSVIC:
      case EVX_FPIEEE_CVTTQV:
      case EVX_FPIEEE_CVTTQVC:
      case EVX_FPIEEE_CVTTQSV:
      case EVX_FPIEEE_CVTTQSVC:
      case EVX_FPIEEE_CVTTQSVI:
      case EVX_FPIEEE_CVTTQSVIC:
      case EVX_FPIEEE_CVTTQSVD:
      case EVX_FPIEEE_CVTTQSVID:
      case EVX_FPIEEE_CVTTQVD:
      case EVX_FPIEEE_CVTTQVM:
      case EVX_FPIEEE_CVTTQSVM:
      case EVX_FPIEEE_CVTTQSVIM:
      case EVX_FPIEEE_CVTSQ:
      case EVX_FPIEEE_CVTSQC:
      case EVX_FPIEEE_CVTTQ:
      case EVX_FPIEEE_CVTTQC:
      case EVX_FPIEEE_CVTTQD:
      case EVX_FPIEEE_CVTTQM:
        (*funcp) (fu_floating_pipeline, 1);
        break;
      default:
        break;
      }
    break;
  case EVX_OPC_PAL:
    break;
  case EVX_OPC_PAL19:
  case EVX_OPC_PAL1B:
  case EVX_OPC_PAL1D:
  case EVX_OPC_PAL1E:
  case EVX_OPC_PAL1F:
    break;
  case EVX_OPC_OPC01:
  case EVX_OPC_OPC02:
  case EVX_OPC_OPC03:
  case EVX_OPC_OPC04:
  case EVX_OPC_OPC05:
  case EVX_OPC_OPC06:
  case EVX_OPC_OPC07:
  case EVX_OPC_OPC0A:
  case EVX_OPC_OPC0C:
  case EVX_OPC_OPC0D:
  case EVX_OPC_OPC0E:
  case EVX_OPC_OPC14:
  case EVX_OPC_OPC1C:
    break;
  default:
    break;
  }
}

/*
**	Return the resource class of the instruction in order
**	to determine dual issue compatability for EV-4.
*/
static enum resource_class instruction_class (alpha_inst inst)
{
  register unsigned opcode, func;

  opcode = inst.o_format.opcode;

  switch (opcode) {
  case EVX_OPC_LDL:
  case EVX_OPC_LDQ:
  case EVX_OPC_LDQU:
  case EVX_OPC_LDLL:
  case EVX_OPC_LDQL:
    return class_load;
    break;
  case EVX_OPC_STL:
  case EVX_OPC_STQ:
  case EVX_OPC_STQU:
  case EVX_OPC_STLC:
  case EVX_OPC_STQC:
    return class_store_data;
    break;
  case EVX_OPC_LDAH:
  case EVX_OPC_LDA:
    return class_iarithmetic;
    break;
  case EVX_OPC_LDF:
  case EVX_OPC_LDD:
  case EVX_OPC_LDS:
  case EVX_OPC_LDT:
    return class_fload;
    break;
  case EVX_OPC_STF:
  case EVX_OPC_STD:
  case EVX_OPC_STS:
  case EVX_OPC_STT:
    return class_fstore_data;
    break;
  case EVX_OPC_SYNC:
    return class_load;
    break;
  case EVX_OPC_JSR:
  case EVX_OPC_BR:
  case EVX_OPC_BSR:
    return class_subroutine;
    break;
  case EVX_OPC_BLBC:
  case EVX_OPC_BLBS:
  case EVX_OPC_BEQ:
  case EVX_OPC_BNE:
  case EVX_OPC_BLT:
  case EVX_OPC_BLE:
  case EVX_OPC_BGT:
  case EVX_OPC_BGE:
    return class_ibranch;
    break;
  case EVX_OPC_FBEQ:
  case EVX_OPC_FBNE:
  case EVX_OPC_FBLT:
  case EVX_OPC_FBLE:
  case EVX_OPC_FBGT:
  case EVX_OPC_FBGE:
    return class_fbranch;
    break;
  case EVX_OPC_ADDI: {
    func = inst.o_format.function;
    switch(func) {
      case EVX_ADDI_ADDLV:
      case EVX_ADDI_ADDQV:
      case EVX_ADDI_SUBLV:
      case EVX_ADDI_SUBQV:
      case EVX_ADDI_ADDL:
      case EVX_ADDI_ADDQ:
      case EVX_ADDI_S4ADDL:
      case EVX_ADDI_S4ADDQ:
      case EVX_ADDI_S8ADDL:
      case EVX_ADDI_S8ADDQ:
      case EVX_ADDI_S4SUBL:
      case EVX_ADDI_S4SUBQ:
      case EVX_ADDI_S8SUBL:
      case EVX_ADDI_S8SUBQ:
      case EVX_ADDI_SUBL:
      case EVX_ADDI_SUBQ:
        return class_iarithmetic;
	break;
      case EVX_ADDI_CMPEQ:
      case EVX_ADDI_CMPLT:
      case EVX_ADDI_CMPLE:
      case EVX_ADDI_CMPULT:
      case EVX_ADDI_CMPULE:
      case EVX_ADDI_CMPBGE:
        return class_icompare;
	break;
      default:
        return class_iarithmetic;
        break;
      }
    }
    break;
  case EVX_OPC_LOGI: {
    func = inst.o_format.function;
    switch(func) {
      case EVX_LOGI_AND:
      case EVX_LOGI_BIC:
      case EVX_LOGI_BIS:
      case EVX_LOGI_ORNOT:
      case EVX_LOGI_XOR:
      case EVX_LOGI_EQV:
        return class_ilogical;
        break;
      case EVX_LOGI_CMOVEQ:
      case EVX_LOGI_CMOVNE:
      case EVX_LOGI_CMOVLBS:
      case EVX_LOGI_CMOVLT:
      case EVX_LOGI_CMOVGE:
      case EVX_LOGI_CMOVLBC:
      case EVX_LOGI_CMOVLE:
      case EVX_LOGI_CMOVGT:
        return class_icondmove;
        break;
      default:
        return class_ilogical;
	break;
      }
    }
    break;
  case EVX_OPC_SHFI: {
    func = inst.o_format.function;
    switch(func) {
      case EVX_SHFI_SLL:
      case EVX_SHFI_SRA:
      case EVX_SHFI_SRL:
      case EVX_SHFI_EXTBL:
      case EVX_SHFI_EXTWL:
      case EVX_SHFI_EXTLL:
      case EVX_SHFI_EXTQL:
      case EVX_SHFI_EXTWH:
      case EVX_SHFI_EXTLH:
      case EVX_SHFI_EXTQH:
      case EVX_SHFI_INSBL:
      case EVX_SHFI_INSWL:
      case EVX_SHFI_INSLL:
      case EVX_SHFI_INSQL:
      case EVX_SHFI_INSWH:
      case EVX_SHFI_INSLH:
      case EVX_SHFI_INSQH:
      case EVX_SHFI_MSKBL:
      case EVX_SHFI_MSKWL:
      case EVX_SHFI_MSKLL:
      case EVX_SHFI_MSKQL:
      case EVX_SHFI_MSKWH:
      case EVX_SHFI_MSKLH:
      case EVX_SHFI_MSKQH:
      case EVX_SHFI_ZAP:
      case EVX_SHFI_ZAPNOT:
        return class_ishift;
        break;
      default:
        return class_ishift;
        break;
      }
    }
    break;
  case EVX_OPC_MULI: {
    func = inst.o_format.function;
    switch(func) {
      case EVX_MULI_MULQV:
      case EVX_MULI_UMULH:
      case EVX_MULI_MULQ:
        return class_imultiplyq;
        break;
      case EVX_MULI_MULL:
      case EVX_MULI_MULLV:
        return class_imultiplyl;
        break;
      default:
        return class_imultiplyl;
        break;
      }
    }
    break;
  case EVX_OPC_FPAUX:
    return class_fpop;
    break;
  case EVX_OPC_FPDEC: {
    func = inst.f_format.function;
    switch(func) {
      case EVX_FPDEC_ADDDU:
      case EVX_FPDEC_ADDDUC:
      case EVX_FPDEC_ADDDSU:
      case EVX_FPDEC_ADDDSUC:
      case EVX_FPDEC_ADDFU:
      case EVX_FPDEC_ADDFUC:
      case EVX_FPDEC_ADDFSU:
      case EVX_FPDEC_ADDFSUC:
      case EVX_FPDEC_ADDGU:
      case EVX_FPDEC_ADDGUC:
      case EVX_FPDEC_ADDGSU:
      case EVX_FPDEC_ADDGSUC:
      case EVX_FPDEC_SUBDU:
      case EVX_FPDEC_SUBDUC:
      case EVX_FPDEC_SUBDSU:
      case EVX_FPDEC_SUBDSUC:
      case EVX_FPDEC_SUBFU:
      case EVX_FPDEC_SUBFUC:
      case EVX_FPDEC_SUBFSU:
      case EVX_FPDEC_SUBFSUC:
      case EVX_FPDEC_SUBGU:
      case EVX_FPDEC_SUBGUC:
      case EVX_FPDEC_SUBGSU:
      case EVX_FPDEC_SUBGSUC:
      case EVX_FPDEC_MULDU:
      case EVX_FPDEC_MULDUC:
      case EVX_FPDEC_MULDSU:
      case EVX_FPDEC_MULDSUC:
      case EVX_FPDEC_MULFU:
      case EVX_FPDEC_MULFUC:
      case EVX_FPDEC_MULFSU:
      case EVX_FPDEC_MULFSUC:
      case EVX_FPDEC_MULGU:
      case EVX_FPDEC_MULGUC:
      case EVX_FPDEC_MULGSU:
      case EVX_FPDEC_MULGSUC:
      case EVX_FPDEC_ADDD:
      case EVX_FPDEC_ADDDC:
      case EVX_FPDEC_ADDDS:
      case EVX_FPDEC_ADDDSC:
      case EVX_FPDEC_ADDF:
      case EVX_FPDEC_ADDFC:
      case EVX_FPDEC_ADDFS:
      case EVX_FPDEC_ADDFSC:
      case EVX_FPDEC_ADDG:
      case EVX_FPDEC_ADDGC:
      case EVX_FPDEC_ADDGS:
      case EVX_FPDEC_ADDGSC:
      case EVX_FPDEC_SUBD:
      case EVX_FPDEC_SUBDC:
      case EVX_FPDEC_SUBDS:
      case EVX_FPDEC_SUBDSC:
      case EVX_FPDEC_SUBF:
      case EVX_FPDEC_SUBFC:
      case EVX_FPDEC_SUBFS:
      case EVX_FPDEC_SUBFSC:
      case EVX_FPDEC_SUBG:
      case EVX_FPDEC_SUBGC:
      case EVX_FPDEC_SUBGS:
      case EVX_FPDEC_SUBGSC:
      case EVX_FPDEC_MULD:
      case EVX_FPDEC_MULDC:
      case EVX_FPDEC_MULDS:
      case EVX_FPDEC_MULDSC:
      case EVX_FPDEC_MULF:
      case EVX_FPDEC_MULFC:
      case EVX_FPDEC_MULFS:
      case EVX_FPDEC_MULFSC:
      case EVX_FPDEC_MULG:
      case EVX_FPDEC_MULGC:
      case EVX_FPDEC_MULGS:
      case EVX_FPDEC_MULGSC:
      case EVX_FPDEC_CMPDEQ:
      case EVX_FPDEC_CMPDLT:
      case EVX_FPDEC_CMPDLE:
      case EVX_FPDEC_CMPGEQ:
      case EVX_FPDEC_CMPGLT:
      case EVX_FPDEC_CMPGLE:
      case EVX_FPDEC_CMPDEQS:
      case EVX_FPDEC_CMPDLTS:
      case EVX_FPDEC_CMPDLES:
      case EVX_FPDEC_CMPGEQS:
      case EVX_FPDEC_CMPGLTS:
      case EVX_FPDEC_CMPGLES:
        return class_fpop;
	break;
      case EVX_FPDEC_DIVF:
      case EVX_FPDEC_DIVFC:
      case EVX_FPDEC_DIVFS:
      case EVX_FPDEC_DIVFSC:
      case EVX_FPDEC_DIVFU:
      case EVX_FPDEC_DIVFUC:
      case EVX_FPDEC_DIVFSU:
      case EVX_FPDEC_DIVFSUC:
        return class_fdivs;
	break;
      case EVX_FPDEC_DIVD:
      case EVX_FPDEC_DIVDC:
      case EVX_FPDEC_DIVDS:
      case EVX_FPDEC_DIVDSC:
      case EVX_FPDEC_DIVDU:
      case EVX_FPDEC_DIVDUC:
      case EVX_FPDEC_DIVDSU:
      case EVX_FPDEC_DIVDSUC:
      case EVX_FPDEC_DIVG:
      case EVX_FPDEC_DIVGC:
      case EVX_FPDEC_DIVGS:
      case EVX_FPDEC_DIVGSC:
      case EVX_FPDEC_DIVGU:
      case EVX_FPDEC_DIVGUC:
      case EVX_FPDEC_DIVGSU:
      case EVX_FPDEC_DIVGSUC:
        return class_fdivd;
	break;
      case EVX_FPDEC_CVTDGU:
      case EVX_FPDEC_CVTDGUC:
      case EVX_FPDEC_CVTDGSU:
      case EVX_FPDEC_CVTDGSUC:
      case EVX_FPDEC_CVTGDU:
      case EVX_FPDEC_CVTGDUC:
      case EVX_FPDEC_CVTGDSU:
      case EVX_FPDEC_CVTGDSUC:
      case EVX_FPDEC_CVTGFU:
      case EVX_FPDEC_CVTGFUC:
      case EVX_FPDEC_CVTGFSU:
      case EVX_FPDEC_CVTGFSUC:
      case EVX_FPDEC_CVTDQV:
      case EVX_FPDEC_CVTDQVC:
      case EVX_FPDEC_CVTDQSV:
      case EVX_FPDEC_CVTDQSVC:
      case EVX_FPDEC_CVTGQV:
      case EVX_FPDEC_CVTGQVC:
      case EVX_FPDEC_CVTGQSV:
      case EVX_FPDEC_CVTGQSVC:
      case EVX_FPDEC_CVTDF:
      case EVX_FPDEC_CVTDFC:
      case EVX_FPDEC_CVTDFS:
      case EVX_FPDEC_CVTDFSC:
      case EVX_FPDEC_CVTDG:
      case EVX_FPDEC_CVTDGC:
      case EVX_FPDEC_CVTDGS:
      case EVX_FPDEC_CVTDGSC:
      case EVX_FPDEC_CVTGD:
      case EVX_FPDEC_CVTGDC:
      case EVX_FPDEC_CVTGDS:
      case EVX_FPDEC_CVTGDSC:
      case EVX_FPDEC_CVTFG:
      case EVX_FPDEC_CVTFGC:
      case EVX_FPDEC_CVTFGS:
      case EVX_FPDEC_CVTFGSC:
      case EVX_FPDEC_CVTGF:
      case EVX_FPDEC_CVTGFC:
      case EVX_FPDEC_CVTGFS:
      case EVX_FPDEC_CVTGFSC:
      case EVX_FPDEC_CVTQD:
      case EVX_FPDEC_CVTQDC:
      case EVX_FPDEC_CVTQDS:
      case EVX_FPDEC_CVTQDSC:
      case EVX_FPDEC_CVTQF:
      case EVX_FPDEC_CVTQFC:
      case EVX_FPDEC_CVTQFS:
      case EVX_FPDEC_CVTQFSC:
      case EVX_FPDEC_CVTQG:
      case EVX_FPDEC_CVTQGC:
      case EVX_FPDEC_CVTQGS:
      case EVX_FPDEC_CVTQGSC:
      case EVX_FPDEC_CVTDQ:
      case EVX_FPDEC_CVTDQC:
      case EVX_FPDEC_CVTDQS:
      case EVX_FPDEC_CVTDQSC:
      case EVX_FPDEC_CVTGQ:
      case EVX_FPDEC_CVTGQC:
      case EVX_FPDEC_CVTGQS:
      case EVX_FPDEC_CVTGQSC:
        return class_fpop;
	break;
      default:
        return class_fpop;
        break;
      }
    }
    break;
  case EVX_OPC_FPIEEE: {
    func = inst.f_format.function;
    switch(func) {
      case EVX_FPIEEE_CMPSEQSU:
      case EVX_FPIEEE_CMPSLTSU:
      case EVX_FPIEEE_CMPSLESU:
      case EVX_FPIEEE_CMPSUNSU:
      case EVX_FPIEEE_CMPTEQSU:
      case EVX_FPIEEE_CMPTLTSU:
      case EVX_FPIEEE_CMPTLESU:
      case EVX_FPIEEE_CMPTUNSU:
      case EVX_FPIEEE_CMPSEQ:
      case EVX_FPIEEE_CMPSLT:
      case EVX_FPIEEE_CMPSLE:
      case EVX_FPIEEE_CMPSUN:
      case EVX_FPIEEE_CMPTEQ:
      case EVX_FPIEEE_CMPTLT:
      case EVX_FPIEEE_CMPTLE:
      case EVX_FPIEEE_CMPTUN:
        return class_fpop;
	break;
      case EVX_FPIEEE_CVTQSSUI:
      case EVX_FPIEEE_CVTQSSUIC:
      case EVX_FPIEEE_CVTQSSUIM:
      case EVX_FPIEEE_CVTQSSUIP:
      case EVX_FPIEEE_CVTQTSUI:
      case EVX_FPIEEE_CVTQTSUIC:
      case EVX_FPIEEE_CVTQTSUIM:
      case EVX_FPIEEE_CVTQTSUIP:
      case EVX_FPIEEE_CVTSTSUI:
      case EVX_FPIEEE_CVTSTSUIC:
      case EVX_FPIEEE_CVTSTSUIM:
      case EVX_FPIEEE_CVTSTSUIP:
      case EVX_FPIEEE_CVTTSU:
      case EVX_FPIEEE_CVTTSUC:
      case EVX_FPIEEE_CVTTSUM:
      case EVX_FPIEEE_CVTTSUP:
      case EVX_FPIEEE_CVTTSSU:
      case EVX_FPIEEE_CVTTSSUC:
      case EVX_FPIEEE_CVTTSSUM:
      case EVX_FPIEEE_CVTTSSUP:
      case EVX_FPIEEE_CVTTSSUI:
      case EVX_FPIEEE_CVTTSSUIC:
      case EVX_FPIEEE_CVTTSSUIM:
      case EVX_FPIEEE_CVTTSSUIP:
      case EVX_FPIEEE_CVTQS:
      case EVX_FPIEEE_CVTQSC:
      case EVX_FPIEEE_CVTQSM:
      case EVX_FPIEEE_CVTQSP:
      case EVX_FPIEEE_CVTQT:
      case EVX_FPIEEE_CVTQTC:
      case EVX_FPIEEE_CVTQTM:
      case EVX_FPIEEE_CVTQTP:
      case EVX_FPIEEE_CVTST:
      case EVX_FPIEEE_CVTSTC:
      case EVX_FPIEEE_CVTSTM:
      case EVX_FPIEEE_CVTSTP:
      case EVX_FPIEEE_CVTTS:
      case EVX_FPIEEE_CVTTSC:
      case EVX_FPIEEE_CVTTSM:
      case EVX_FPIEEE_CVTTSP:
        return class_fpop;
	break;
      case EVX_FPIEEE_MULSU:
      case EVX_FPIEEE_MULSUC:
      case EVX_FPIEEE_MULSUM:
      case EVX_FPIEEE_MULSUP:
      case EVX_FPIEEE_MULSSU:
      case EVX_FPIEEE_MULSSUC:
      case EVX_FPIEEE_MULSSUM:
      case EVX_FPIEEE_MULSSUP:
      case EVX_FPIEEE_MULSSUI:
      case EVX_FPIEEE_MULSSUIC:
      case EVX_FPIEEE_MULSSUIM:
      case EVX_FPIEEE_MULSSUIP:
      case EVX_FPIEEE_MULTU:
      case EVX_FPIEEE_MULTUC:
      case EVX_FPIEEE_MULTUM:
      case EVX_FPIEEE_MULTUP:
      case EVX_FPIEEE_MULTSU:
      case EVX_FPIEEE_MULTSUC:
      case EVX_FPIEEE_MULTSUM:
      case EVX_FPIEEE_MULTSUP:
      case EVX_FPIEEE_MULTSUI:
      case EVX_FPIEEE_MULTSUIC:
      case EVX_FPIEEE_MULTSUIM:
      case EVX_FPIEEE_MULTSUIP:
      case EVX_FPIEEE_SUBSU:
      case EVX_FPIEEE_SUBSUC:
      case EVX_FPIEEE_SUBSUM:
      case EVX_FPIEEE_SUBSUP:
      case EVX_FPIEEE_SUBSSU:
      case EVX_FPIEEE_SUBSSUC:
      case EVX_FPIEEE_SUBSSUM:
      case EVX_FPIEEE_SUBSSUP:
      case EVX_FPIEEE_SUBSSUI:
      case EVX_FPIEEE_SUBSSUIC:
      case EVX_FPIEEE_SUBSSUIM:
      case EVX_FPIEEE_SUBSSUIP:
      case EVX_FPIEEE_SUBTU:
      case EVX_FPIEEE_SUBTUC:
      case EVX_FPIEEE_SUBTUM:
      case EVX_FPIEEE_SUBTUP:
      case EVX_FPIEEE_SUBTSU:
      case EVX_FPIEEE_SUBTSUC:
      case EVX_FPIEEE_SUBTSUM:
      case EVX_FPIEEE_SUBTSUP:
      case EVX_FPIEEE_SUBTSUI:
      case EVX_FPIEEE_SUBTSUIC:
      case EVX_FPIEEE_SUBTSUIM:
      case EVX_FPIEEE_SUBTSUIP:
      case EVX_FPIEEE_ADDSU:
      case EVX_FPIEEE_ADDSUC:
      case EVX_FPIEEE_ADDSUM:
      case EVX_FPIEEE_ADDSUP:
      case EVX_FPIEEE_ADDSSU:
      case EVX_FPIEEE_ADDSSUC:
      case EVX_FPIEEE_ADDSSUM:
      case EVX_FPIEEE_ADDSSUP:
      case EVX_FPIEEE_ADDSSUI:
      case EVX_FPIEEE_ADDSSUIC:
      case EVX_FPIEEE_ADDSSUIM:
      case EVX_FPIEEE_ADDSSUIP:
      case EVX_FPIEEE_ADDTU:
      case EVX_FPIEEE_ADDTUC:
      case EVX_FPIEEE_ADDTUM:
      case EVX_FPIEEE_ADDTUP:
      case EVX_FPIEEE_ADDTSU:
      case EVX_FPIEEE_ADDTSUC:
      case EVX_FPIEEE_ADDTSUM:
      case EVX_FPIEEE_ADDTSUP:
      case EVX_FPIEEE_ADDTSUI:
      case EVX_FPIEEE_ADDTSUIC:
      case EVX_FPIEEE_ADDTSUIM:
      case EVX_FPIEEE_ADDTSUIP:
      case EVX_FPIEEE_MULS:
      case EVX_FPIEEE_MULSC:
      case EVX_FPIEEE_MULSM:
      case EVX_FPIEEE_MULSP:
      case EVX_FPIEEE_MULT:
      case EVX_FPIEEE_MULTC:
      case EVX_FPIEEE_MULTM:
      case EVX_FPIEEE_MULTP:
      case EVX_FPIEEE_SUBS:
      case EVX_FPIEEE_SUBSC:
      case EVX_FPIEEE_SUBSM:
      case EVX_FPIEEE_SUBSP:
      case EVX_FPIEEE_SUBT:
      case EVX_FPIEEE_SUBTC:
      case EVX_FPIEEE_SUBTM:
      case EVX_FPIEEE_SUBTP:
      case EVX_FPIEEE_ADDS:
      case EVX_FPIEEE_ADDSC:
      case EVX_FPIEEE_ADDSM:
      case EVX_FPIEEE_ADDSP:
      case EVX_FPIEEE_ADDT:
      case EVX_FPIEEE_ADDTC:
      case EVX_FPIEEE_ADDTM:
      case EVX_FPIEEE_ADDTP:
        return class_fpop;
	break;
      case EVX_FPIEEE_DIVS:
      case EVX_FPIEEE_DIVSC:
      case EVX_FPIEEE_DIVSM:
      case EVX_FPIEEE_DIVSP:
      case EVX_FPIEEE_DIVSU:
      case EVX_FPIEEE_DIVSUC:
      case EVX_FPIEEE_DIVSUM:
      case EVX_FPIEEE_DIVSUP:
      case EVX_FPIEEE_DIVSSU:
      case EVX_FPIEEE_DIVSSUC:
      case EVX_FPIEEE_DIVSSUM:
      case EVX_FPIEEE_DIVSSUP:
      case EVX_FPIEEE_DIVSSUI:
      case EVX_FPIEEE_DIVSSUIC:
      case EVX_FPIEEE_DIVSSUIM:
      case EVX_FPIEEE_DIVSSUIP:
        return class_fdivs;
	break;
      case EVX_FPIEEE_DIVT:
      case EVX_FPIEEE_DIVTC:
      case EVX_FPIEEE_DIVTM:
      case EVX_FPIEEE_DIVTP:
      case EVX_FPIEEE_DIVTU:
      case EVX_FPIEEE_DIVTUC:
      case EVX_FPIEEE_DIVTUM:
      case EVX_FPIEEE_DIVTUP:
      case EVX_FPIEEE_DIVTSU:
      case EVX_FPIEEE_DIVTSUC:
      case EVX_FPIEEE_DIVTSUM:
      case EVX_FPIEEE_DIVTSUP:
      case EVX_FPIEEE_DIVTSUI:
      case EVX_FPIEEE_DIVTSUIC:
      case EVX_FPIEEE_DIVTSUIM:
      case EVX_FPIEEE_DIVTSUIP:
        return class_fdivd;
	break;
      case EVX_FPIEEE_CVTSQV:
      case EVX_FPIEEE_CVTSQVC:
      case EVX_FPIEEE_CVTSQSV:
      case EVX_FPIEEE_CVTSQSVC:
      case EVX_FPIEEE_CVTSQSVI:
      case EVX_FPIEEE_CVTSQSVIC:
      case EVX_FPIEEE_CVTTQV:
      case EVX_FPIEEE_CVTTQVC:
      case EVX_FPIEEE_CVTTQSV:
      case EVX_FPIEEE_CVTTQSVC:
      case EVX_FPIEEE_CVTTQSVI:
      case EVX_FPIEEE_CVTTQSVIC:
      case EVX_FPIEEE_CVTTQSVD:
      case EVX_FPIEEE_CVTTQSVID:
      case EVX_FPIEEE_CVTTQVD:
      case EVX_FPIEEE_CVTTQVM:
      case EVX_FPIEEE_CVTTQSVM:
      case EVX_FPIEEE_CVTTQSVIM:
      case EVX_FPIEEE_CVTSQ:
      case EVX_FPIEEE_CVTSQC:
      case EVX_FPIEEE_CVTTQ:
      case EVX_FPIEEE_CVTTQC:
      case EVX_FPIEEE_CVTTQD:
      case EVX_FPIEEE_CVTTQM:
        return class_fpop;
        break;
      default:
        return class_fpop;
        break;
      }
    }
    break;
  case EVX_OPC_PAL:
  case EVX_OPC_PAL19:
  case EVX_OPC_PAL1B:
  case EVX_OPC_PAL1D:
  case EVX_OPC_PAL1E:
  case EVX_OPC_PAL1F:
  case EVX_OPC_OPC01:
  case EVX_OPC_OPC02:
  case EVX_OPC_OPC03:
  case EVX_OPC_OPC04:
  case EVX_OPC_OPC05:
  case EVX_OPC_OPC06:
  case EVX_OPC_OPC07:
  case EVX_OPC_OPC0A:
  case EVX_OPC_OPC0C:
  case EVX_OPC_OPC0D:
  case EVX_OPC_OPC0E:
  case EVX_OPC_OPC14:
  case EVX_OPC_OPC1C:
  default:
    return class_null;
    break;
  }
}

#define PAL_SIM_VAX_RTL	  0x00FF
#define PAL_SIM_ALPHA_RTL 0x00FD
#define PAL_SIM_LOC_BPT	  0x00FC
#define PAL_SIM_REM_BPT	  0x00FB

#define RA(x)                (((x) >> 21) & 0x1f)
#define RB(x)                (((x) >> 16) & 0x1f)
#define RC(x)                ((x) & 0x1F)
#define OPCODE(x)            (((x) >> 26) & 0x3f)
#define MEMORY_DISP_FIELD(x) ((x) & 0xFFFF)
#define FUNCTION(x)          (((x) >> 5) & 0x7f)
#define LITERAL_FLAG(x)      ((x) & 0x1000)
#define LITERAL(x)           (((x) >> 13) & 0xFF)
#define FLOAT_FUNCTION(x)    (((x) >> 5) & 0x7FF)
#define PALCODE_FUNCTION(x)  ((x) & 0x3FFFFFF)

static int BRANCH_DISP_FIELD(unsigned x)
{
  unsigned plain = (x & 0x1FFFFF) << 11;
  int retval = ((int) plain) >> 9;
  return retval;
  }

static void evax_inst_decode(unsigned inst, unsigned va, char *string)
{
  int ra = RA(inst);
  int rb = RB(inst);
  int rc = RC(inst);
  int md = MEMORY_DISP_FIELD(inst);
  int bd = BRANCH_DISP_FIELD(inst) + va + 4;

  char rbs[10];
  if (LITERAL_FLAG(inst))
    sprintf(rbs, "#%X", LITERAL(inst));
  else sprintf(rbs, "R%d", RB(inst));

  /*
  //
  // Alpha srm v4 display section (and non-conflicting v3)
  //
  */
  switch (OPCODE(inst)) {
    case EVX_OPC_LDA:
      if (rb == 31)		sprintf(string, "MOV      %X,R%d", md, ra);
      else			sprintf(string, "LDA      R%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_LDAH:		sprintf(string, "LDAH     R%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_LDL:		sprintf(string, "LDL      R%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_LDQ:	 	sprintf(string, "LDQ      R%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_LDLL:		sprintf(string, "LDL_L    R%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_LDQL:		sprintf(string, "LDQ_L    R%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_LDQU:		sprintf(string, "LDQ_U    R%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_STL:		sprintf(string, "STL      R%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_STQ:		sprintf(string, "STQ      R%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_STLC:		sprintf(string, "STL_C    R%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_STQC:		sprintf(string, "STQ_C    R%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_STQU:		sprintf(string, "STQ_U    R%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_LDF:		sprintf(string, "LDF      F%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_LDD:		sprintf(string, "LDD      F%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_LDS:		sprintf(string, "LDS      F%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_LDT:		sprintf(string, "LDT      F%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_STF:		sprintf(string, "STF      F%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_STD:		sprintf(string, "STD      F%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_STS:		sprintf(string, "STS      F%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_STT:		sprintf(string, "STT      F%d,%X(R%d)", ra, md, rb);	break;
    case EVX_OPC_SYNC:		
      switch (md) {
        case EVX_SYNC_TRAPB:	sprintf(string, "TRAPB");						break;
        case EVX_SYNC_MB:	sprintf(string, "MB");						break;
        case EVX_SYNC_RCC:	sprintf(string, "RPCC     R%d", ra);				break;
        case EVX_SYNC_FETCH:	sprintf(string, "FETCH    (%s)", rbs);				break;
        case EVX_SYNC_FETCH_M:	sprintf(string, "FETCH_M  (%s)", rbs);				break;
        case EVX_SYNC_RC:	sprintf(string, "RC       R%d", ra);				break;
        case EVX_SYNC_RS:	sprintf(string, "RS       R%d", ra);				break;
        default:		sprintf(string, "Illegal Opcode");					break;
      }											break;
    case EVX_OPC_JSR:
      switch (md & 0xC000) {
        case 0x0000:		sprintf(string, "JMP      R%d,(R%d)", ra, rb);			break;
        case 0x4000:		sprintf(string, "JSR      R%d,(R%d)", ra, rb);			break;
        case 0x8000:		sprintf(string, "RET      R%d,(R%d)", ra, rb);			break;
        case 0xC000:		sprintf(string, "JSR_COROUTINE R%d,(R%d)", ra, rb);		break;
      }											break;
    case EVX_OPC_BEQ:		sprintf(string, "BEQ      R%d,%X", ra, bd);		break;
    case EVX_OPC_BNE:		sprintf(string, "BNE      R%d,%X", ra, bd);		break;
    case EVX_OPC_BLT:		sprintf(string, "BLT      R%d,%X", ra, bd);		break;
    case EVX_OPC_BLE:		sprintf(string, "BLE      R%d,%X", ra, bd);		break;
    case EVX_OPC_BGT:		sprintf(string, "BGT      R%d,%X", ra, bd);		break;
    case EVX_OPC_BGE:		sprintf(string, "BGE      R%d,%X", ra, bd);		break;
    case EVX_OPC_BLBC:		sprintf(string, "BLBC     R%d,%X", ra, bd);		break;
    case EVX_OPC_BLBS:		sprintf(string, "BLBS     R%d,%X", ra, bd);		break;
    case EVX_OPC_BR:
      switch (ra) {
        case 31: sprintf(string, "BR       %X", bd);						break;
        default: sprintf(string, "BR       R%d,%X", ra, bd);					break;
        }										break;
    case EVX_OPC_BSR:		sprintf(string, "BSR      R%d,%X", ra, bd);		break;
    case EVX_OPC_FBEQ:		sprintf(string, "FBEQ     F%d,%X", ra, bd);		break;
    case EVX_OPC_FBNE:		sprintf(string, "FBNE     F%d,%X", ra, bd);		break;
    case EVX_OPC_FBLT:		sprintf(string, "FBLT     F%d,%X", ra, bd);		break;
    case EVX_OPC_FBLE:		sprintf(string, "FBLE     F%d,%X", ra, bd);		break;
    case EVX_OPC_FBGT:		sprintf(string, "FBGT     F%d,%X", ra, bd);		break;
    case EVX_OPC_FBGE:		sprintf(string, "FBGE     F%d,%X", ra, bd);		break;
    case EVX_OPC_ADDI:
      switch (FUNCTION(inst)) {
        case EVX_ADDI_ADDL:
          if (ra == 31)		sprintf(string, "SEXTL    %s,R%d", rbs, rc);
	  else			sprintf(string, "ADDL     R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_ADDI_ADDLV:	sprintf(string, "ADDL/V   R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_ADDI_ADDQ:	sprintf(string, "ADDQ     R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_ADDI_ADDQV:	sprintf(string, "ADDQ/V   R%d,%s,R%d", ra, rbs, rc);		break;
	case EVX_ADDI_S4ADDL:	sprintf(string, "S4ADDL   R%d,%s,R%d", ra, rbs, rc);		break;
	case EVX_ADDI_S4ADDQ:	sprintf(string, "S4ADDQ   R%d,%s,R%d", ra, rbs, rc);		break;
	case EVX_ADDI_S8ADDL:	sprintf(string, "S8ADDL   R%d,%s,R%d", ra, rbs, rc);		break;
	case EVX_ADDI_S8ADDQ:	sprintf(string, "S8ADDQ   R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_ADDI_SUBL:
          if (ra == 31)		sprintf(string, "NEGL     %s,R%d", rbs, rc);
          else			sprintf(string, "SUBL     R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_ADDI_SUBLV:
          if (ra == 31)		sprintf(string, "NEGL/V   %s,R%d", rbs, rc);
          else			sprintf(string, "SUBL/V   R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_ADDI_SUBQ:
          if (ra == 31)		sprintf(string, "NEGQ     %s,R%d", rbs, rc);
          else			sprintf(string, "SUBQ     R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_ADDI_SUBQV:
          if (ra == 31)		sprintf(string, "NEGQ/V   %s,R%d", rbs, rc);
          else			sprintf(string, "SUBQ/V   R%d,%s,R%d", ra, rbs, rc);		break;
	case EVX_ADDI_S4SUBL:	sprintf(string, "S4SUBL   R%d,%s,R%d", ra, rbs, rc);		break;
	case EVX_ADDI_S4SUBQ:	sprintf(string, "S4SUBQ   R%d,%s,R%d", ra, rbs, rc);		break;
	case EVX_ADDI_S8SUBL:	sprintf(string, "S8SUBL   R%d,%s,R%d", ra, rbs, rc);		break;
	case EVX_ADDI_S8SUBQ:	sprintf(string, "S8SUBQ   R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_ADDI_CMPEQ:	sprintf(string, "CMPEQ    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_ADDI_CMPLT:	sprintf(string, "CMPLT    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_ADDI_CMPLE:	sprintf(string, "CMPLE    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_ADDI_CMPULT:	sprintf(string, "CMPULT   R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_ADDI_CMPULE:	sprintf(string, "CMPULE   R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_ADDI_CMPBGE:	sprintf(string, "CMPBGE   R%d,%s,R%d", ra, rbs, rc);		break;
        default:		sprintf(string, "Illegal Opcode");					break;
      }											break;
    case EVX_OPC_LOGI:
      switch (FUNCTION(inst)) {
        case EVX_LOGI_AND:	sprintf(string, "AND      R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_LOGI_BIS:
	  if (ra == 31 && rb == 31 && rc == 31)
				sprintf(string, "NOP");
	  else if (ra == 31 && rb == 31)
				sprintf(string, "CLR      R%d", rc);
	  else if (ra == rb)	sprintf(string, "MOV      R%d,R%d", rb, rc);
	  else			sprintf(string, "BIS      R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_LOGI_XOR:	sprintf(string, "XOR      R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_LOGI_BIC:	sprintf(string, "BIC      R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_LOGI_ORNOT:
          if (ra == 31)		sprintf(string, "NOT      %s,R%d", rbs, rc);
          else			sprintf(string, "ORNOT    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_LOGI_EQV:	sprintf(string, "EQV      R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_LOGI_CMOVEQ:	sprintf(string, "CMOVEQ   R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_LOGI_CMOVNE:	sprintf(string, "CMOVNE   R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_LOGI_CMOVLT:	sprintf(string, "CMOVLT   R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_LOGI_CMOVLE:	sprintf(string, "CMOVLE   R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_LOGI_CMOVGT:	sprintf(string, "CMOVGT   R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_LOGI_CMOVGE:	sprintf(string, "CMOVGE   R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_LOGI_CMOVLBS:	sprintf(string, "CMOVLBS  R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_LOGI_CMOVLBC:	sprintf(string, "CMOVLBC  R%d,%s,R%d", ra, rbs, rc);		break;
        default:		sprintf(string, "Illegal Opcode");					break;
      }
											break;
    case EVX_OPC_SHFI:
      switch (FUNCTION(inst)) {
        case EVX_SHFI_SLL:	sprintf(string, "SLL      R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_SRL:	sprintf(string, "SRL      R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_SRA:	sprintf(string, "SRA      R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_EXTBL:	sprintf(string, "EXTBL    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_EXTWL:	sprintf(string, "EXTWL    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_EXTLL:	sprintf(string, "EXTLL    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_EXTQL:	sprintf(string, "EXTQL    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_EXTWH:	sprintf(string, "EXTWH    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_EXTLH:	sprintf(string, "EXTLH    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_EXTQH:	sprintf(string, "EXTQH    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_INSBL:	sprintf(string, "INSBL    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_INSWL:	sprintf(string, "INSWL    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_INSLL:	sprintf(string, "INSLL    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_INSQL:	sprintf(string, "INSQL    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_INSWH:	sprintf(string, "INSWH    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_INSLH:	sprintf(string, "INSLH    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_INSQH:	sprintf(string, "INSQH    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_MSKBL:	sprintf(string, "MSKBL    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_MSKWL:	sprintf(string, "MSKWL    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_MSKLL:	sprintf(string, "MSKLL    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_MSKQL:	sprintf(string, "MSKQL    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_MSKWH:	sprintf(string, "MSKWH    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_MSKLH:	sprintf(string, "MSKLH    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_MSKQH:	sprintf(string, "MSKQH    R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_ZAP:	sprintf(string, "ZAP      R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_SHFI_ZAPNOT:	sprintf(string, "ZAPNOT   R%d,%s,R%d", ra, rbs, rc);		break;
        default:		sprintf(string, "Illegal Opcode");					break;
      }											break;
    case EVX_OPC_MULI:
      switch (FUNCTION(inst)) {
        case EVX_MULI_MULL:	sprintf(string, "MULL     R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_MULI_MULLV:	sprintf(string, "MULL/V   R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_MULI_MULQ:	sprintf(string, "MULQ     R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_MULI_MULQV:	sprintf(string, "MULQ/V   R%d,%s,R%d", ra, rbs, rc);		break;
        case EVX_MULI_UMULH:	sprintf(string, "UMULH    R%d,%s,R%d", ra, rbs, rc);		break;
        default:		sprintf(string, "Illegal Opcode");					break;
      }											break;
    case EVX_OPC_FPAUX:
      switch (FLOAT_FUNCTION(inst)) {
        case EVX_FPAUX_CPYS:
          if (ra == 31 && rb == 31 && rc == 31)
				sprintf(string, "FNOP");
          else if (ra == 31 && rb == 31)
				sprintf(string, "FCLR     F%d", rc);
          else if (ra == rb)	sprintf(string, "FMOV     F%d,F%d", rb, rc);
          else if (ra == 31)	sprintf(string, "FABS     F%d,F%d", rb, rc);
          else			sprintf(string, "CPYS     F%d,F%d,F%d", ra, rb, rc);		break;
        case EVX_FPAUX_CPYSN:
          if (ra == rb)			sprintf(string, "FNEG     F%d,F%d", rb, rc);
          else				sprintf(string, "CPYSN    F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPAUX_CPYSE:		sprintf(string, "CPYSE    F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPAUX_CVTLQ:		sprintf(string, "CVTLQ    F%d,F%d", rb, rc);		break;
        case EVX_FPAUX_CVTQL:		sprintf(string, "CVTQL    F%d,F%d", rb, rc);		break;
        case EVX_FPAUX_CVTQLV:		sprintf(string, "CVTQL/V  F%d,F%d", rb, rc);		break;
        case EVX_FPAUX_CVTQLSV:		sprintf(string, "CVTQL/SV F%d,F%d", rb, rc);		break;
        case EVX_FPAUX_FCMOVEQ:		sprintf(string, "FCMOVEQ  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPAUX_FCMOVNE:		sprintf(string, "FCMOVNE  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPAUX_FCMOVLT:		sprintf(string, "FCMOVLT  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPAUX_FCMOVLE:		sprintf(string, "FCMOVLE  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPAUX_FCMOVGT:		sprintf(string, "FCMOVGT  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPAUX_FCMOVGE:		sprintf(string, "FCMOVGE  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPAUX_WR_FPCTL:
	  if ((ra == rb) && (rb == rc))	sprintf(string, "MT_FPCR  F%d", ra);
	  else				sprintf(string, "MT_FPCR  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPAUX_RD_FPCTL:
	  if ((ra == rb) && (rb == rc))	sprintf(string, "MF_FPCR  F%d", ra);
	  else				sprintf(string, "MF_FPCR  F%d,F%d,F%d", ra, rb, rc);	break;
        default:			sprintf(string, "Illegal Opcode");				break;
      }											break;
    case EVX_OPC_FPIEEE:
      switch (FLOAT_FUNCTION(inst)) {
        case EVX_FPIEEE_CMPTEQ:		sprintf(string, "CMPTEQ   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_CMPTLT:		sprintf(string, "CMPTLT   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_CMPTLE:		sprintf(string, "CMPTLE   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_CMPTUN:		sprintf(string, "CMPTUN   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_CMPTEQSU:	sprintf(string, "CMPTEQ/SU F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_CMPTLTSU:	sprintf(string, "CMPTLT/SU F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_CMPTLESU:	sprintf(string, "CMPTLE/SU F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_CMPTUNSU:	sprintf(string, "CMPTUN/SU F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_CVTQS:		sprintf(string, "CVTQS    F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTQSC:		sprintf(string, "CVTQS/C  F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTQSM:		sprintf(string, "CVTQS/M  F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTQSP:		sprintf(string, "CVTQS/P  F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTQSSUI:	sprintf(string, "CVTQS/SUI F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTQSSUIC:	sprintf(string, "CVTQS/SUIC F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTQSSUIM:	sprintf(string, "CVTQS/SUIM F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTQSSUIP:	sprintf(string, "CVTQS/SUIP F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTQT:		sprintf(string, "CVTQT    F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTQTC:		sprintf(string, "CVTQT/C  F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTQTM:		sprintf(string, "CVTQT/M  F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTQTP:		sprintf(string, "CVTQT/P  F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTQTSUI:	sprintf(string, "CVTQT/SUI F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTQTSUIC:	sprintf(string, "CVTQT/SUIC F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTQTSUIM:	sprintf(string, "CVTQT/SUIM F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTQTSUIP:	sprintf(string, "CVTQT/SUIP F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTS:		sprintf(string, "CVTTS    F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTSC:		sprintf(string, "CVTTS/C  F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTSM:		sprintf(string, "CVTTS/M  F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTSP:		sprintf(string, "CVTTS/P  F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTSU:		sprintf(string, "CVTTS/U  F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTSUC:	sprintf(string, "CVTTS/UC F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTSUM:	sprintf(string, "CVTTS/UM F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTSUP:	sprintf(string, "CVTTS/UP F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTSSU:	sprintf(string, "CVTTS/SU F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTSSUC:	sprintf(string, "CVTTS/SUC F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTSSUM:	sprintf(string, "CVTTS/SUM F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTSSUP:	sprintf(string, "CVTTS/SUP F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTSSUI:	sprintf(string, "CVTTS/SUI F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTSSUIC:	sprintf(string, "CVTTS/SUIC F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTSSUIM:	sprintf(string, "CVTTS/SUIM F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTSSUIP:	sprintf(string, "CVTTS/SUIP F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_DIVS:		sprintf(string, "DIVS     F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVSC:		sprintf(string, "DIVS/C   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVSM:		sprintf(string, "DIVS/M   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVSP:		sprintf(string, "DIVS/P   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVSU:		sprintf(string, "DIVS/U   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVSUC:		sprintf(string, "DIVS/UC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVSUM:		sprintf(string, "DIVS/UM  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVSUP:		sprintf(string, "DIVS/UP  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVSSU:		sprintf(string, "DIVS/SU  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVSSUC:	sprintf(string, "DIVS/SUC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVSSUM:	sprintf(string, "DIVS/SUM F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVSSUP:	sprintf(string, "DIVS/SUP F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVSSUI:	sprintf(string, "DIVS/SUI F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVSSUIC:	sprintf(string, "DIVS/SUIC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVSSUIM:	sprintf(string, "DIVS/SUIM F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVSSUIP:	sprintf(string, "DIVS/SUIP F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVT:		sprintf(string, "DIVT     F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVTC:		sprintf(string, "DIVT/C   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVTM:		sprintf(string, "DIVT/M   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVTP:		sprintf(string, "DIVT/P   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVTU:		sprintf(string, "DIVT/U   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVTUC:		sprintf(string, "DIVT/UC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVTUM:		sprintf(string, "DIVT/UM  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVTUP:		sprintf(string, "DIVT/UP  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVTSU:		sprintf(string, "DIVT/SU  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVTSUC:	sprintf(string, "DIVT/SUC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVTSUM:	sprintf(string, "DIVT/SUM F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVTSUP:	sprintf(string, "DIVT/SUP F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVTSUI:	sprintf(string, "DIVT/SUI F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVTSUIC:	sprintf(string, "DIVT/SUIC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVTSUIM:	sprintf(string, "DIVT/SUIM F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_DIVTSUIP:	sprintf(string, "DIVT/SUIP F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULS:		sprintf(string, "MULS     F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULSC:		sprintf(string, "MULS/C   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULSM:		sprintf(string, "MULS/M   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULSP:		sprintf(string, "MULS/P   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULSU:		sprintf(string, "MULS/U   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULSUC:		sprintf(string, "MULS/UC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULSUM:		sprintf(string, "MULS/UM  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULSUP:		sprintf(string, "MULS/UP  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULSSU:		sprintf(string, "MULS/SU  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULSSUC:	sprintf(string, "MULS/SUC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULSSUM:	sprintf(string, "MULS/SUM F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULSSUP:	sprintf(string, "MULS/SUP F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULSSUI:	sprintf(string, "MULS/SUI F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULSSUIC:	sprintf(string, "MULS/SUIC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULSSUIM:	sprintf(string, "MULS/SUIM F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULSSUIP:	sprintf(string, "MULS/SUIP F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULT:		sprintf(string, "MULT     F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULTC:		sprintf(string, "MULT/C   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULTM:		sprintf(string, "MULT/M   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULTP:		sprintf(string, "MULT/P   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULTU:		sprintf(string, "MULT/U   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULTUC:		sprintf(string, "MULT/UC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULTUM:		sprintf(string, "MULT/UM  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULTUP:		sprintf(string, "MULT/UP  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULTSU:		sprintf(string, "MULT/SU  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULTSUC:	sprintf(string, "MULT/SUC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULTSUM:	sprintf(string, "MULT/SUM F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULTSUP:	sprintf(string, "MULT/SUP F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULTSUI:	sprintf(string, "MULT/SUI F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULTSUIC:	sprintf(string, "MULT/SUIC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULTSUIM:	sprintf(string, "MULT/SUIM F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_MULTSUIP:	sprintf(string, "MULT/SUIP F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBS:
          if (ra == 31)			sprintf(string, "NEGS     F%d,F%d", rb, rc);
          else				sprintf(string, "SUBS     F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBSC:		sprintf(string, "SUBS/C   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBSM:		sprintf(string, "SUBS/M   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBSP:		sprintf(string, "SUBS/P   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBSU:		sprintf(string, "SUBS/U   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBSUC:		sprintf(string, "SUBS/UC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBSUM:		sprintf(string, "SUBS/UM  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBSUP:		sprintf(string, "SUBS/UP  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBSSU:		sprintf(string, "SUBS/SU  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBSSUC:	sprintf(string, "SUBS/SUC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBSSUM:	sprintf(string, "SUBS/SUM F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBSSUP:	sprintf(string, "SUBS/SUP F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBSSUI:	sprintf(string, "SUBS/SUI F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBSSUIC:	sprintf(string, "SUBS/SUIC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBSSUIM:	sprintf(string, "SUBS/SUIM F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBSSUIP:	sprintf(string, "SUBS/SUIP F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBT:
          if (ra == 31)			sprintf(string, "NEGT     F%d,F%d", rb, rc);
          else				sprintf(string, "SUBT     F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBTC:		sprintf(string, "SUBT/C   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBTM:		sprintf(string, "SUBT/M   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBTP:		sprintf(string, "SUBT/P   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBTU:		sprintf(string, "SUBT/U   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBTUC:		sprintf(string, "SUBT/UC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBTUM:		sprintf(string, "SUBT/UM  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBTUP:		sprintf(string, "SUBT/UP  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBTSU:		sprintf(string, "SUBT/SU  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBTSUC:	sprintf(string, "SUBT/SUC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBTSUM:	sprintf(string, "SUBT/SUM F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBTSUP:	sprintf(string, "SUBT/SUP F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBTSUI:	sprintf(string, "SUBT/SUI F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBTSUIC:	sprintf(string, "SUBT/SUIC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBTSUIM:	sprintf(string, "SUBT/SUIM F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_SUBTSUIP:	sprintf(string, "SUBT/SUIP F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDS:		sprintf(string, "ADDS     F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDSC:		sprintf(string, "ADDS/C   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDSM:		sprintf(string, "ADDS/M   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDSP:		sprintf(string, "ADDS/P   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDSU:		sprintf(string, "ADDS/U   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDSUC:		sprintf(string, "ADDS/UC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDSUM:		sprintf(string, "ADDS/UM  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDSUP:		sprintf(string, "ADDS/UP  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDSSU:		sprintf(string, "ADDS/SU  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDSSUC:	sprintf(string, "ADDS/SUC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDSSUM:	sprintf(string, "ADDS/SUM F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDSSUP:	sprintf(string, "ADDS/SUP F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDSSUI:	sprintf(string, "ADDS/SUI F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDSSUIC:	sprintf(string, "ADDS/SUIC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDSSUIM:	sprintf(string, "ADDS/SUIM F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDSSUIP:	sprintf(string, "ADDS/SUIP F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDT:		sprintf(string, "ADDT     F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDTC:		sprintf(string, "ADDT/C   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDTM:		sprintf(string, "ADDT/M   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDTP:		sprintf(string, "ADDT/P   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDTU:		sprintf(string, "ADDT/U   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDTUC:		sprintf(string, "ADDT/UC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDTUM:		sprintf(string, "ADDT/UM  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDTUP:		sprintf(string, "ADDT/UP  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDTSU:		sprintf(string, "ADDT/SU  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDTSUC:	sprintf(string, "ADDT/SUC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDTSUM:	sprintf(string, "ADDT/SUM F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDTSUP:	sprintf(string, "ADDT/SUP F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDTSUI:	sprintf(string, "ADDT/SUI F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDTSUIC:	sprintf(string, "ADDT/SUIC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDTSUIM:	sprintf(string, "ADDT/SUIM F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_ADDTSUIP:	sprintf(string, "ADDT/SUIP F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPIEEE_CVTTQ:		sprintf(string, "CVTTQ    F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTQV:		sprintf(string, "CVTTQ/V  F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTQSV:	sprintf(string, "CVTTQ/SV F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTQSVI:	sprintf(string, "CVTTQ/SVI F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTQC:		sprintf(string, "CVTTQ/C  F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTQVC:	sprintf(string, "CVTTQ/VC F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTQSVC:	sprintf(string, "CVTTQ/SVC F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTQSVIC:	sprintf(string, "CVTTQ/SVIC F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTQD:		sprintf(string, "CVTTQ/D  F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTQVD:	sprintf(string, "CVTTQ/VD F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTQSVD:	sprintf(string, "CVTTQ/SVD F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTQSVID:	sprintf(string, "CVTTQ/SVID F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTQM:		sprintf(string, "CVTTQ/M  F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTQVM:	sprintf(string, "CVTTQ/VM F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTQSVM:	sprintf(string, "CVTTQ/SVM F%d,F%d", rb, rc);		break;
        case EVX_FPIEEE_CVTTQSVIM:	sprintf(string, "CVTTQ/SVIM F%d,F%d", rb, rc);		break;
        default:			sprintf(string, "Illegal Opcode");				break;
      }											break;
    case EVX_OPC_FPDEC:
      switch (FLOAT_FUNCTION(inst)) {
        case EVX_FPDEC_ADDF:		sprintf(string, "ADDF     F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_ADDFC:		sprintf(string, "ADDF/C   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_ADDFU:		sprintf(string, "ADDF/U   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_ADDFUC:		sprintf(string, "ADDF/UC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_ADDFS:		sprintf(string, "ADDF/S   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_ADDFSC:		sprintf(string, "ADDF/SC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_ADDFSU:		sprintf(string, "ADDF/SU  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_ADDFSUC:		sprintf(string, "ADDF/SUC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_ADDG:		sprintf(string, "ADDG     F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_ADDGC:		sprintf(string, "ADDG/C   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_ADDGU:		sprintf(string, "ADDG/U   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_ADDGUC:		sprintf(string, "ADDG/UC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_ADDGS:		sprintf(string, "ADDG/S   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_ADDGSC:		sprintf(string, "ADDG/SC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_ADDGSU:		sprintf(string, "ADDG/SU  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_ADDGSUC:		sprintf(string, "ADDG/SUC F%d,F%d,F%d", ra, rb, rc);	break;

        case EVX_FPDEC_SUBF:
          if (ra == 31)			sprintf(string, "NEGF     F%d,F%d", rb, rc);
          else				sprintf(string, "SUBF     F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_SUBFC:		sprintf(string, "SUBF/C   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_SUBFU:		sprintf(string, "SUBF/U   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_SUBFUC:		sprintf(string, "SUBF/UC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_SUBFS:
          if (ra == 31)			sprintf(string, "NEGF/S   F%d,F%d", rb, rc);
          else				sprintf(string, "SUBF/S   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_SUBFSC:		sprintf(string, "SUBF/SC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_SUBFSU:		sprintf(string, "SUBF/SU  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_SUBFSUC:		sprintf(string, "SUBF/SUC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_SUBG:
          if (ra == 31)			sprintf(string, "NEGG     F%d,F%d", rb, rc);
          else				sprintf(string, "SUBG     F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_SUBGC:		sprintf(string, "SUBG/C   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_SUBGU:		sprintf(string, "SUBG/U   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_SUBGUC:		sprintf(string, "SUBG/UC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_SUBGS:
          if (ra == 31)			sprintf(string, "NEGG/S   F%d,F%d", rb, rc);
          else				sprintf(string, "SUBG/S   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_SUBGSC:		sprintf(string, "SUBG/SC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_SUBGSU:		sprintf(string, "SUBG/SU  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_SUBGSUC:		sprintf(string, "SUBG/SUC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_MULF:		sprintf(string, "MULF     F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_MULFC:		sprintf(string, "MULF/C   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_MULFU:		sprintf(string, "MULF/U   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_MULFUC:		sprintf(string, "MULF/UC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_MULFS:		sprintf(string, "MULF/S   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_MULFSC:		sprintf(string, "MULF/SC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_MULFSU:		sprintf(string, "MULF/SU  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_MULFSUC:		sprintf(string, "MULF/SUC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_MULG:		sprintf(string, "MULG     F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_MULGC:		sprintf(string, "MULG/C   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_MULGU:		sprintf(string, "MULG/U   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_MULGUC:		sprintf(string, "MULG/UC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_MULGS:		sprintf(string, "MULG/S   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_MULGSC:		sprintf(string, "MULG/SC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_MULGSU:		sprintf(string, "MULG/SU  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_MULGSUC:		sprintf(string, "MULG/SUC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_DIVF:		sprintf(string, "DIVF     F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_DIVFC:		sprintf(string, "DIVF/C   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_DIVFU:		sprintf(string, "DIVF/U   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_DIVFUC:		sprintf(string, "DIVF/UC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_DIVFS:		sprintf(string, "DIVF/S   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_DIVFSC:		sprintf(string, "DIVF/SC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_DIVFSU:		sprintf(string, "DIVF/SU  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_DIVFSUC:		sprintf(string, "DIVF/SUC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_DIVG:		sprintf(string, "DIVG     F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_DIVGC:		sprintf(string, "DIVG/C   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_DIVGU:		sprintf(string, "DIVG/U   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_DIVGUC:		sprintf(string, "DIVG/UC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_DIVGS:		sprintf(string, "DIVG/S   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_DIVGSC:		sprintf(string, "DIVG/SC  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_DIVGSU:		sprintf(string, "DIVG/SU  F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_DIVGSUC:		sprintf(string, "DIVG/SUC F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_CMPDEQ:		sprintf(string, "CMPDEQ   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_CMPDLT:		sprintf(string, "CMPDLT   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_CMPDLE:		sprintf(string, "CMPDLE   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_CMPGEQ:		sprintf(string, "CMPGEQ   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_CMPGLT:		sprintf(string, "CMPGLT   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_CMPGLE:		sprintf(string, "CMPGLE   F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_CMPDEQS:		sprintf(string, "CMPDEQ/S F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_CMPDLTS:		sprintf(string, "CMPDLT/S F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_CMPDLES:		sprintf(string, "CMPDLE/S F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_CMPGEQS:		sprintf(string, "CMPGEQ/S F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_CMPGLTS:		sprintf(string, "CMPGLT/S F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_CMPGLES:		sprintf(string, "CMPGLE/S F%d,F%d,F%d", ra, rb, rc);	break;
        case EVX_FPDEC_CVTDF:		sprintf(string, "CVTDF    F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTDFC:		sprintf(string, "CVTDF/C  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTDFS:		sprintf(string, "CVTDF/S  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTDFSC:		sprintf(string, "CVTDF/SC F%d,F%d", rb, rc);		break;

	case EVX_FPDEC_CVTDG:		sprintf(string, "CVTDG    F%d,F%d", rb, rc);		break;
	case EVX_FPDEC_CVTDGC:		sprintf(string, "CVTDG/C  F%d,F%d", rb, rc);		break;
	case EVX_FPDEC_CVTDGU:		sprintf(string, "CVTDG/U  F%d,F%d", rb, rc);		break;
	case EVX_FPDEC_CVTDGUC:		sprintf(string, "CVTDG/UC F%d,F%d", rb, rc);		break;
	case EVX_FPDEC_CVTDGS:		sprintf(string, "CVTDG/S  F%d,F%d", rb, rc);		break;
	case EVX_FPDEC_CVTDGSC:		sprintf(string, "CVTDG/SC F%d,F%d", rb, rc);		break;
	case EVX_FPDEC_CVTDGSU:		sprintf(string, "CVTDG/SU F%d,F%d", rb, rc);		break;
	case EVX_FPDEC_CVTDGSUC:	sprintf(string, "CVTDG/SUC F%d,F%d", rb, rc);		break;

        case EVX_FPDEC_CVTGD:		sprintf(string, "CVTGD    F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGDC:		sprintf(string, "CVTGD/C  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGDU:		sprintf(string, "CVTGD/U  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGDUC:		sprintf(string, "CVTGD/UC F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGDS:		sprintf(string, "CVTGD/S  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGDSC:		sprintf(string, "CVTGD/SC F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGDSU:		sprintf(string, "CVTGD/SU F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGDSUC:	sprintf(string, "CVTGD/SUC F%d,F%d", rb, rc);		break;

        case EVX_FPDEC_CVTGF:		sprintf(string, "CVTGF    F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGFC:		sprintf(string, "CVTGF/C  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGFU:		sprintf(string, "CVTGF/U  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGFUC:		sprintf(string, "CVTGF/UC F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGFS:		sprintf(string, "CVTGF/S  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGFSC:		sprintf(string, "CVTGF/SC F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGFSU:		sprintf(string, "CVTGF/SU F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGFSUC:	sprintf(string, "CVTGF/SUC F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTQD:		sprintf(string, "CVTQD    F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTQDC:		sprintf(string, "CVTQD/C  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTQDS:		sprintf(string, "CVTQD/S  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTQDSC:		sprintf(string, "CVTQD/SC F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTQF:		sprintf(string, "CVTQF    F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTQFC:		sprintf(string, "CVTQF/C  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTQFS:		sprintf(string, "CVTQF/S  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTQFSC:		sprintf(string, "CVTQF/SC F%d,F%d", rb, rc);	 	break;
        case EVX_FPDEC_CVTQG:		sprintf(string, "CVTQG    F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTQGC:		sprintf(string, "CVTQG/C  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTQGS:		sprintf(string, "CVTQG/S  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTQGSC:		sprintf(string, "CVTQG/SC F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTDQ:		sprintf(string, "CVTDQ    F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTDQC:		sprintf(string, "CVTDQ/C  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTDQV:		sprintf(string, "CVTDQ/V  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTDQVC:		sprintf(string, "CVTDQ/VC F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTDQS:		sprintf(string, "CVTDQ/S  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTDQSC:		sprintf(string, "CVTDQ/SC F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTDQSV:		sprintf(string, "CVTDQ/SV F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTDQSVC:	sprintf(string, "CVTDQ/SVC F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGQ:		sprintf(string, "CVTGQ    F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGQC:		sprintf(string, "CVTGQ/C  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGQV:		sprintf(string, "CVTGQ/V  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGQVC:		sprintf(string, "CVTGQ/VC F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGQS:		sprintf(string, "CVTGQ/S  F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGQSC:		sprintf(string, "CVTGQ/SC F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGQSV:		sprintf(string, "CVTGQ/SV F%d,F%d", rb, rc);		break;
        case EVX_FPDEC_CVTGQSVC:	sprintf(string, "CVTGQ/SVC F%d,F%d", rb, rc);		break;
        default:			sprintf(string, "Illegal Opcode");				break;
      }											break;
    case EVX_OPC_PAL:
      switch (PALCODE_FUNCTION(inst)) {
	case EVX_PAL_BPT:	sprintf(string, "CALL_PAL BPT");					break;
        case EVX_PAL_BUGCHK:	sprintf(string, "CALL_PAL BUGCHK");				break;
        case EVX_PAL_CHME:	sprintf(string, "CALL_PAL CHME");					break;
        case EVX_PAL_CHMK:	sprintf(string, "CALL_PAL CHMK");					break;
        case EVX_PAL_CHMS:	sprintf(string, "CALL_PAL CHMS");					break;
        case EVX_PAL_CHMU:	sprintf(string, "CALL_PAL CHMU");					break;
        case EVX_PAL_IMB:	sprintf(string, "CALL_PAL IMB");					break;
        case EVX_PAL_INSQHIL:	sprintf(string, "CALL_PAL INSQHIL");				break;
        case EVX_PAL_INSQTIL:	sprintf(string, "CALL_PAL INSQTIL");				break;
        case EVX_PAL_INSQHIQ:	sprintf(string, "CALL_PAL INSQHIQ");				break;
        case EVX_PAL_INSQTIQ:	sprintf(string, "CALL_PAL INSQTIQ");				break;
        case EVX_PAL_INSQUEL:	sprintf(string, "CALL_PAL INSQUEL");				break;
        case EVX_PAL_INSQUEQ:	sprintf(string, "CALL_PAL INSQUEQ");				break;
        case EVX_PAL_INSQUELD:	sprintf(string, "CALL_PAL INSQUELD");				break;
        case EVX_PAL_INSQUEQD:	sprintf(string, "CALL_PAL INSQUEQD");				break;
        case EVX_PAL_PROBER:	sprintf(string, "CALL_PAL PROBER");				break;
        case EVX_PAL_PROBEW:	sprintf(string, "CALL_PAL PROBEW");				break;
        case EVX_PAL_RD_PS:	sprintf(string, "CALL_PAL RD_PS");					break;
        case EVX_PAL_REI:	sprintf(string, "CALL_PAL REI");					break;
        case EVX_PAL_REMQHIL:	sprintf(string, "CALL_PAL REMQHIL");				break;
        case EVX_PAL_REMQTIL:	sprintf(string, "CALL_PAL REMQTIL");				break;
        case EVX_PAL_REMQHIQ:	sprintf(string, "CALL_PAL REMQHIQ");				break;
        case EVX_PAL_REMQTIQ:	sprintf(string, "CALL_PAL REMQTIQ");				break;
        case EVX_PAL_REMQUEL:	sprintf(string, "CALL_PAL REMQUEL");				break;
        case EVX_PAL_REMQUEQ:	sprintf(string, "CALL_PAL REMQUEQ");				break;
        case EVX_PAL_REMQUELD:	sprintf(string, "CALL_PAL REMQUELD");				break;
        case EVX_PAL_REMQUEQD:	sprintf(string, "CALL_PAL REMQUEQD");				break;
        case EVX_PAL_SWASTEN:	sprintf(string, "CALL_PAL SWASTEN");				break;
        case EVX_PAL_WR_PS_SW:	sprintf(string, "CALL_PAL WR_PS_SW");				break;
        case EVX_PAL_RSCC:	sprintf(string, "CALL_PAL RSCC");					break;
        case EVX_PAL_READ_UNQ:	sprintf(string, "CALL_PAL READ_UNQ");			break;
        case EVX_PAL_WRITE_UNQ:	sprintf(string, "CALL_PAL WRITE_UNQ");			break;
        case EVX_PAL_CFLUSH:	sprintf(string, "CALL_PAL CFLUSH");				break;
        case EVX_PAL_DRAINA:	sprintf(string, "CALL_PAL DRAINA");				break;
        case EVX_PAL_HALT:	sprintf(string, "CALL_PAL HALT");					break;
        case EVX_PAL_LDQP:	sprintf(string, "CALL_PAL LDQP");					break;
        case EVX_PAL_STQP:	sprintf(string, "CALL_PAL STQP");					break;
        case EVX_PAL_SWPCTX:	sprintf(string, "CALL_PAL SWPCTX");				break;
        case EVX_PAL_MFPR_ASN:	sprintf(string, "CALL_PAL MFPR_ASN");				break;
        case EVX_PAL_MTPR_ASTEN:sprintf(string, "CALL_PAL MTPR_ASTEN");				break;
        case EVX_PAL_MTPR_ASTSR:sprintf(string, "CALL_PAL MTPR_ASTSR");				break;
        case EVX_PAL_MFPR_AT:	sprintf(string, "CALL_PAL MFPR_AT");				break;
        case EVX_PAL_MTPR_AT:	sprintf(string, "CALL_PAL MTPR_AT");				break;
        case EVX_PAL_MFPR_FEN:	sprintf(string, "CALL_PAL MFPR_FEN");				break;
        case EVX_PAL_MTPR_FEN:	sprintf(string, "CALL_PAL MTPR_FEN");				break;
        case EVX_PAL_MTPR_IPIR:	sprintf(string, "CALL_PAL MTPR_IPIR");				break;
        case EVX_PAL_MFPR_IPL:	sprintf(string, "CALL_PAL MFPR_IPL");				break;
        case EVX_PAL_MTPR_IPL:	sprintf(string, "CALL_PAL MTPR_IPL");				break;
        case EVX_PAL_MFPR_MCES:	sprintf(string, "CALL_PAL MFPR_MCES");				break;
        case EVX_PAL_MTPR_MCES:	sprintf(string, "CALL_PAL MTPR_MCES");				break;
        case EVX_PAL_MFPR_PCBB:	sprintf(string, "CALL_PAL MFPR_PCBB");				break;
        case EVX_PAL_MFPR_PRBR:	sprintf(string, "CALL_PAL MFPR_PRBR");				break;
        case EVX_PAL_MTPR_PRBR:	sprintf(string, "CALL_PAL MTPR_PRBR");				break;
        case EVX_PAL_MFPR_PTBR:	sprintf(string, "CALL_PAL MFPR_PTBR");				break;
        case EVX_PAL_MFPR_SCBB:	sprintf(string, "CALL_PAL MFPR_SCBB");				break;
        case EVX_PAL_MTPR_SCBB:	sprintf(string, "CALL_PAL MTPR_SCBB");				break;
        case EVX_PAL_MTPR_SIRR:	sprintf(string, "CALL_PAL MTPR_SIRR");				break;
        case EVX_PAL_MFPR_SISR:	sprintf(string, "CALL_PAL MFPR_SISR");				break;
        case EVX_PAL_MFPR_TBCHK:sprintf(string, "CALL_PAL MFPR_TBCHK");				break;
        case EVX_PAL_MTPR_TBIA:	sprintf(string, "CALL_PAL MTPR_TBIA");				break;
        case EVX_PAL_MTPR_TBIAP:sprintf(string, "CALL_PAL MTPR_TBIAP");				break;
        case EVX_PAL_MTPR_TBIS:	sprintf(string, "CALL_PAL MTPR_TBIS");				break;
        case EVX_PAL_MFPR_ESP:	sprintf(string, "CALL_PAL MFPR_ESP");				break;
        case EVX_PAL_MTPR_ESP:	sprintf(string, "CALL_PAL MTPR_ESP");				break;
        case EVX_PAL_MFPR_SSP:	sprintf(string, "CALL_PAL MFPR_SSP");				break;
        case EVX_PAL_MTPR_SSP:	sprintf(string, "CALL_PAL MTPR_SSP");				break;
        case EVX_PAL_MFPR_USP:	sprintf(string, "CALL_PAL MFPR_USP");				break;
        case EVX_PAL_MTPR_USP:	sprintf(string, "CALL_PAL MTPR_USP");				break;
        case EVX_PAL_MTPR_TBISD:sprintf(string, "CALL_PAL MTPR_TBISD");				break;
        case EVX_PAL_MTPR_TBISI:sprintf(string, "CALL_PAL MTPR_TBISI");				break;
        case EVX_PAL_MFPR_ASTEN:sprintf(string, "CALL_PAL MFPR_ASTEN");				break;
        case EVX_PAL_MFPR_ASTSR:sprintf(string, "CALL_PAL MFPR_ASTSR");				break;
        case EVX_PAL_MFPR_WHAMI:	sprintf(string, "CALL_PAL MFPR_WHAMI");			break;

	case EVX_PAL_RESERVED_GETSS:	sprintf(string, "CALL_PAL RESERVED_GETSS");		break;
	case EVX_PAL_RESERVED_PUTSS:	sprintf(string, "CALL_PAL RESERVED_PUTSS");		break;
	case EVX_PAL_RESERVED_RLS1:	sprintf(string, "CALL_PAL RESERVED_RLS1");			break;
	case EVX_PAL_RESERVED_RLS2:	sprintf(string, "CALL_PAL RESERVED_RLS2");			break;
	case EVX_PAL_RESERVED_DVT1:	sprintf(string, "CALL_PAL RESERVED_DVT1");			break;
	case EVX_PAL_NPHALT: 		sprintf(string, "CALL_PAL NPHALT");			break;
        default:		sprintf(string, "CALL_PAL %X", PALCODE_FUNCTION(inst));		break;
      }											break;
    case EVX_OPC_HW_MFPR:
      sprintf(string, "HW_MFPR R%d,%s%s%s%d", ra, (md & 0x80) ? "P" : "", (md & 0x40) ? "A" : "", (md & 0x20) ? "I" : "", md & 0x1F);
											break;
    case EVX_OPC_HW_MTPR:
      sprintf(string, "HW_MTPR R%d,%s%s%s%d", ra, (md & 0x80) ? "P" : "", (md & 0x40) ? "A" : "", (md & 0x20) ? "I" : "", md & 0x1F);
											break;
    case EVX_OPC_HW_LD:
      sprintf(string, "HW_LD/%s%s%s%s R%d,%X(R%d)", (md & 0x1000) ? "Q" : "L", (md & 0x8000) ? "P" : "V",
	      (md & 0x4000) ? "A" : "", (md & 0x2000) ? "W" : "", ra, md & 0x0FFF, rb);
											break;
    case EVX_OPC_HW_ST:
      sprintf(string, "HW_ST/%s%s%s%s R%d,%X(R%d)", (md & 0x1000) ? "Q" : "L", (md & 0x8000) ? "P" : "V",
	      (md & 0x4000) ? "A" : "", (md & 0x2000) ? "W" : "", ra, md & 0x0FFF, rb);
											break;
    case EVX_OPC_HW_REI:	sprintf(string, "HW_REI");					break;
    case EVX_OPC_OPC01:
    case EVX_OPC_OPC02:
    case EVX_OPC_OPC03:
    case EVX_OPC_OPC04:
    case EVX_OPC_OPC05:
    case EVX_OPC_OPC06:
    case EVX_OPC_OPC07:
    case EVX_OPC_OPC0A:
    case EVX_OPC_OPC0C:
    case EVX_OPC_OPC0D:
    case EVX_OPC_OPC0E:
    case EVX_OPC_OPC14:
    case EVX_OPC_OPC1C:		sprintf(string, "Reserved Opcode");			break;
    default:			sprintf(string, "Illegal Opcode");				break;
  }
}

#define PIPE_LENGTH 20		/* The number of cycles displayed before
				** the diagram wraps.
				*/

static void print_ev4_pipeline_diagram (unsigned va)
{
  int 		i, j, node;
  char 		disassembly [132];
  unsigned 	issue_slot, exec_tim, cur_slot = 0;
  char 		boundary_box [PIPE_LENGTH+3];
  char 		exec_profile [PIPE_LENGTH+3];

  boundary_box [0] = '|';
  for (i = 0; i < PIPE_LENGTH; i++)
    boundary_box [i+1] = ((char) i%10) + '0';
  boundary_box [PIPE_LENGTH+1] = '|';
  boundary_box [PIPE_LENGTH+2] = '\0';

  exec_profile [0] = '|';
  exec_profile [PIPE_LENGTH+1] = '|';
  exec_profile [PIPE_LENGTH+2] = '\0';

  /*
  ** Print block header.
  */
  printf("%s Cycle   Offset Number Instruction\n", boundary_box);

  for (i = 0; i < number_of_nodes; i++, va += 4) {
    memset (exec_profile+1, ' ', PIPE_LENGTH);
    node       = node_issue [i].node;
    issue_slot = node_issue [i].cycle;
    exec_tim   = exec_time [node];
    evax_inst_decode (instructions [node].word, va, disassembly);

    if (exec_tim > 1)
      /*
      ** Mark execution time in the diagram.
      */
      for (j = issue_slot+1; j < exec_tim+issue_slot; j++)
        exec_profile [(j%PIPE_LENGTH)+1] = 'x';
    if (issue_slot != cur_slot)
      /*
      ** Mark stall time in the diagram.
      */
      for (j = cur_slot; j < issue_slot; j++)
        exec_profile [(j%PIPE_LENGTH)+1] = '.';
    exec_profile [(issue_slot%PIPE_LENGTH)+1] = 'I';
    
    /*
    ** Update the ideal slot in which the next instruction should
    ** should issue, this is a fucntion of the pc.
    */
    if (va % 8)
      cur_slot = issue_slot + 1;
    else
      cur_slot = issue_slot;
    /*
    ** Output the information for this instruction.
    */
    printf ("%s %5d %08X    %3d %s\n", exec_profile, issue_slot, va, node, disassembly);
    }
  printf ("\n");
}

/*
**	Free all the allocated resource information. Is called
**	once per call. from within reorganize_bb.
*/
static void reorganize_reset()
{
  unsigned i;
  plain_resource *pme, *pnext;
  memory_resource *mme, *mnext;
  
  /*
  **	Free all allocated storage, right now the only heap
  **	allocated storage is for the linked list elements.
  */
  for (i = 0; i < (unsigned) resource_end; i++) {
    pme = plain_head [i];
    while (pme) {
      pnext = pme->next;
      free ( (char *) pme);
      pme = pnext;
      }
    plain_head [i] = plain_tail [i] = 0;
    }

  mme = mem_head;
  while (mme) {
    mnext = mme->next;
    free ( (char *) mme);
    mme = mnext;
    }
  mem_head = mem_tail = 0;

  /*
  **	1. Invalidate the constraint information structure.
  **	2. Mark all functional units as available.
  */
  bzero ((void*) constraints, sizeof(constraints));
  bzero ((void*) fu_available, sizeof(fu_available));
  bzero ((void*) fu_count, sizeof(fu_count));
  bzero ((void*) exec_time, sizeof(exec_time));
}

main (int argc, char *argv[])
{
  alpha_inst insts [14];

  /* Create sample basic block */

  insts [0].m_format.ra = 0;			/* LDD	F0,0(R16) */
  insts [0].m_format.rb = 16;
  insts [0].m_format.memory_displacement = 0;
  insts [0].m_format.opcode = EVX_OPC_LDD;

  insts [1].m_format.ra = 8;			/* LDD	F8,0(R0) */
  insts [1].m_format.rb = 0;
  insts [1].m_format.memory_displacement = 0;
  insts [1].m_format.opcode = EVX_OPC_LDD;

  insts [2].m_format.ra = 1;			/* LDD	F1,8(R16) */
  insts [2].m_format.rb = 16;
  insts [2].m_format.memory_displacement = 8;
  insts [2].m_format.opcode = EVX_OPC_LDD;

  insts [3].m_format.ra = 9;			/* LDD	F9,8(R0) */
  insts [3].m_format.rb = 0;
  insts [3].m_format.memory_displacement = 8;
  insts [3].m_format.opcode = EVX_OPC_LDD;

  insts [4].f_format.fa = 0;			/* MULG F0,F5,F0 */
  insts [4].f_format.fb = 5;
  insts [4].f_format.fc = 0;
  insts [4].f_format.opcode = EVX_OPC_FPDEC;
  insts [4].f_format.function = EVX_FPDEC_MULG;

  insts [5].f_format.fa = 8;			/* ADDG F8,F0,F8 */
  insts [5].f_format.fb = 0;
  insts [5].f_format.fc = 8;
  insts [5].f_format.opcode = EVX_OPC_FPDEC;
  insts [5].f_format.function = EVX_FPDEC_ADDG;

  insts [6].f_format.fa = 1;			/* MULG F1,F5,F1 */
  insts [6].f_format.fb = 5;
  insts [6].f_format.fc = 1;
  insts [6].f_format.opcode = EVX_OPC_FPDEC;
  insts [6].f_format.function = EVX_FPDEC_MULG;

  insts [7].f_format.fa = 9;			/* ADDG F9,F1,F9 */
  insts [7].f_format.fb = 1;
  insts [7].f_format.fc = 9;
  insts [7].f_format.opcode = EVX_OPC_FPDEC;
  insts [7].f_format.function = EVX_FPDEC_ADDG;

  insts [8].m_format.ra = 16;			/* LDA R16,16(R16) */
  insts [8].m_format.rb = 16;
  insts [8].m_format.memory_displacement = 16;
  insts [8].m_format.opcode = EVX_OPC_LDA;

  insts [9].o_format.ra = 16;			/* CMPLT R16,R10,R11 */
  insts [9].o_format.rb = 10;
  insts [9].o_format.rc = 11;
  insts [9].o_format.form = 0;
  insts [9].o_format.function = EVX_ADDI_CMPLT;
  insts [9].o_format.opcode = EVX_OPC_ADDI;

  insts [10].m_format.ra = 0;			/* LDA R0,16(R0) */
  insts [10].m_format.rb = 0;
  insts [10].m_format.memory_displacement = 16;
  insts [10].m_format.opcode = EVX_OPC_LDA;

  insts [11].m_format.ra = 8;			/* STD	F8,0(R0) */
  insts [11].m_format.rb = 0;
  insts [11].m_format.memory_displacement = 0;
  insts [11].m_format.opcode = EVX_OPC_STD;

  insts [12].m_format.ra = 9;			/* STD	F9,8(R0) */
  insts [12].m_format.rb = 0;
  insts [12].m_format.memory_displacement = 8;
  insts [12].m_format.opcode = EVX_OPC_STD;

  insts [13].b_format.ra = 11;			/* BNE R11, */
  insts [13].b_format.opcode = EVX_OPC_BNE;
  insts [13].b_format.branch_displacement = -14;

  printf ("Basic block before code scheduling\n");
  schedule_bb_ev4 (insts, 14, 0, false);
  printf ("Basic block after code scheduling\n");
  schedule_bb_ev4 (insts, 14, 0, true);
  }

